//
//  STKAudioPlayer.swift
//  STKAudioPlayer
//
//  Created by Viktor on 14/05/2019.
//  Copyright © 2019 Thong Nguyen. All rights reserved.
//

import Foundation
import AudioToolbox
import Cocoa

protocol STKAudioPlayerDelegate {
    func audioPlayer(_ audioPlayer: STKAudioPlayer, stateChanged state: STKAudioPlayerState, previousState: STKAudioPlayerState)
    func audioPlayer(_ audioPlayer: STKAudioPlayer, didUpdateMetaData metaDataDictionary: NSDictionary)
}

let kOutputBus = 0
let kInputBus = 1

let STK_DEFAULT_PCM_BUFFER_SIZE_IN_SECONDS: Float32 = 10.0
let STK_DEFAULT_SECONDS_REQUIRED_TO_START_PLAYING: Float32 = 1.0
let STK_DEFAULT_SECONDS_REQUIRED_TO_START_PLAYING_AFTER_BUFFER_UNDERRUN: Float32 = 7.5
let STK_MAX_COMPRESSED_PACKETS_FOR_BITRATE_CALCULATION: UInt32 = 4096
let STK_DEFAULT_READ_BUFFER_SIZE: UInt32 = 64 * 1024
let STK_DEFAULT_PACKET_BUFFER_SIZE: UInt32 = 2048
let STK_DEFAULT_GRACE_PERIOD_AFTER_SEEK_SECONDS: Float32 = 0.5

let audioFileStreamPropertyListenerProc: AudioFileStream_PropertyListenerProc = { (clientData, audioFileStream, propertyId, flags) in
    let sself = unsafeBitCast(clientData, to: STKAudioPlayer.self)
    sself.handlePropertyChangeForFileStream(audioFileStreamIn: audioFileStream, fileStreamPropertyID: propertyId, ioFlags: flags)
}

let audioFileStreamPacketsProc: AudioFileStream_PacketsProc = { (clientData, numberBytes, numberPackets, inputData, packetDescriptions) in
    let sself = unsafeBitCast(clientData, to: STKAudioPlayer.self)
    sself.handleAudioPackets(inputData: inputData, numberBytes: numberBytes, numberPackets: numberPackets, packetDescriptions: packetDescriptions)
}

enum STKAudioPlayerInternalState: NSInteger {
    case initialised = 0
    case running = 1
    case playing = 3
    case rebuffering = 5
    case startingThread = 9
    case waitingForData = 16
    case waitingForDataAfterSeek = 33
    case paused = 64
    case stopped = 512
    case pendingNext = 1024
    case disposed = 1073741824
    case error = 2147483648
}

enum STKAudioPlayerState: NSInteger {
    case ready = 0
    case running = 1
    case playing = 3
    case buffering = 5
    case paused = 9
    case stopped = 16
    case error = 32
    case disposed = 64
}

enum STKAudioPlayerStopReason: NSInteger {
    case None = 0
    case eof
    case userAction
    case pendingNext
    case disposed
    case error = 0xffff
}

enum STKAudioPlayerErrorCode: NSInteger {
    case stkAudioPlayerErrorNone = 0
    case stkAudioPlayerErrorDataSource
    case stkAudioPlayerErrorStreamParseBytesFailed
    case stkAudioPlayerErrorAudioSystemError
    case cError
    case stkAudioPlayerErrorDataNotFound
    case stkAudioPlayerErrorOther = 0xffff
}

typealias STKFrameFilter = (UInt32, UInt32, UInt32, UnsafeMutableRawPointer?) -> Void

class STKFrameFilterEntry {
    init(filter filterIn: @escaping STKFrameFilter, andName nameIn: String?) {
        filter = filterIn
        name = nameIn
    }
    var name: String?
    var filter: STKFrameFilter?
}

class STKAudioPlayer: NSObject, STKHTTPDataSourceDelegate {
    
    var maxFramesPerSlice: UInt32 = 4096
    
    // Get the mixerUnit unit
    var mixerUnit: AudioComponentInstance?
    var mixerDesc = AudioComponentDescription(componentType: kAudioUnitType_Mixer, componentSubType: kAudioUnitSubType_MultiChannelMixer, componentManufacturer: kAudioUnitManufacturer_Apple, componentFlags: 0, componentFlagsMask: 0)
    
    var converterUnit: AudioComponentInstance?
    var converterDesc = AudioComponentDescription(componentType: kAudioUnitType_FormatConverter, componentSubType: kAudioUnitSubType_AUConverter, componentManufacturer: kAudioUnitManufacturer_Apple, componentFlags: 0, componentFlagsMask: 0)
    
    var canonicalAudioStreamBasicDescription: AudioStreamBasicDescription = {
        var bytesPerSample = UInt32(MemoryLayout<Int16>.size)
        let flags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagsNativeEndian | kAudioFormatFlagIsPacked
        let component = AudioStreamBasicDescription(mSampleRate: 44100, mFormatID: kAudioFormatLinearPCM, mFormatFlags: flags, mBytesPerPacket: bytesPerSample * 2, mFramesPerPacket: 1, mBytesPerFrame: bytesPerSample * 2, mChannelsPerFrame: 2, mBitsPerChannel: 8 * bytesPerSample, mReserved: 0)
        return component
    }()
    
    var recordAudioStreamBasicDescription = AudioStreamBasicDescription()
    
    var readBuffer: UnsafeMutablePointer<UInt8>?
    var readBufferSize: Int = 0
    
    var audioGraph: AUGraph?
    
    var mixerNode: AUNode = 0
    var outputNode: AUNode = 0
    var converterNode: AUNode = 0
    
    var waitingForDataAfterSeekFrameCount: Int32 = 0
    var framesRequiredToStartPlaying: UInt32 = 0
    var framesRequiredToPlayAfterRebuffering: UInt32 = 0
    var framesRequiredBeforeWaitingForDataAfterSeekBecomesPlaying: UInt32 = 0
    
    var currentlyPlayingEntry: STKHTTPDataSource!
    var currentlyReadingEntry: STKHTTPDataSource!
    var upcomingQueue: NSMutableArray = []
    var bufferingQueue: NSMutableArray = []
    var pcmBufferSpinLock = os_unfair_lock()
    var internalStateLock = os_unfair_lock()
    var pcmBufferTotalFrameCount: UInt32 = 0
    var pcmBufferFrameStartIndex: UInt32 = 0
    var pcmBufferUsedFrameCount: UInt32 = 0
    var pcmBufferFrameSizeInBytes: UInt32 = 0
    
    var pcmAudioBufferList: AudioBufferList = AudioBufferList(mNumberBuffers: 2, mBuffers: AudioBuffer())
    var pcmAudioBuffer: AudioBuffer {
        get { return pcmAudioBufferList.mBuffers }
        set { pcmAudioBufferList.mBuffers = newValue }
    }
    var audioConverterRef: AudioConverterRef?
    var audioConverterAudioStreamBasicDescription: AudioStreamBasicDescription = {
        var bytesPerSample = UInt32(MemoryLayout<Int16>.size)
        let flags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagsNativeEndian | kAudioFormatFlagIsPacked
        let component = AudioStreamBasicDescription(mSampleRate: 44100, mFormatID: kAudioFormatLinearPCM, mFormatFlags: flags, mBytesPerPacket: bytesPerSample * 2, mFramesPerPacket: 1, mBytesPerFrame: bytesPerSample * 2, mChannelsPerFrame: 2, mBitsPerChannel: 8 * bytesPerSample, mReserved: 0)
        return component
    }()
    
    var deallocating: Bool = false
    var discontinuous: Bool = false
    
    var playbackThread: Thread?
    var playbackThreadRunLoop: RunLoop?
    var audioFileStream: AudioFileStreamID?
    var threadStartedLock: NSConditionLock?
    var threadFinishedCondLock: NSConditionLock?
    
    var recordAudioFileId: AudioFileID?
    var recordFilePacketPosition: UInt32 = 0
    var recordAudioConverterRef: AudioConverterRef?
    var recordOutputBufferSize: UInt32 = 0
    var recordOutputBuffer: UnsafeMutablePointer<UInt8>?
    var recordPacketsPerBuffer: UInt32 = 0
    var recordPacketSize: UInt32 = 0
    var recordPacketDescriptions: AudioStreamPacketDescription! = AudioStreamPacketDescription()
    var stopBackBackgroundTaskBlock: (() -> Void)?
    var seekVersion: Atomic<Int>?
    var seekLock = os_unfair_lock()
    var currentEntryReferencesLock = os_unfair_lock()
    
    var playerMutex = pthread_mutex_t()
    var playerThreadReadyCondition = pthread_cond_t()
    var mainThreadSyncCallMutex = pthread_mutex_t()
    var mainThreadSyncCallReadyCondition = pthread_cond_t()
    var waiting: Bool = false
    var requestedSeekTime: Double = 0.0
    var disposeWasRequested: Bool = false
    var seekToTimeWasRequested: Bool = false
    
    /// Returns an array of STKFrameFilterEntry objects representing the filters currently in use
    var frameFilters: NSArray = []
    
    var internalState: STKAudioPlayerInternalState = .stopped
    var stateBeforePaused: STKAudioPlayerInternalState = .stopped
    /// Gets the current state of the player
    var state: STKAudioPlayerState = .ready
    /// Gets the reason why the player is stopped (if any)
    var stopReason: STKAudioPlayerStopReason = .None
    /// Gets and sets the delegate used for receiving events from the STKAudioPlayer
    var delegate: STKAudioPlayerDelegate!
    
    override init() {
        super.init()
    }
    
    convenience init(graph auGraph: AUGraph, _ convertNode: AUNode) {
        self.init()
        audioGraph = auGraph
        converterNode = convertNode
        initialize()
    }
    
    func initialize() {
        framesRequiredToStartPlaying = UInt32(Float32(canonicalAudioStreamBasicDescription.mSampleRate) * min(STK_DEFAULT_SECONDS_REQUIRED_TO_START_PLAYING, STK_DEFAULT_PCM_BUFFER_SIZE_IN_SECONDS))
        
        framesRequiredToPlayAfterRebuffering = UInt32(Float32(canonicalAudioStreamBasicDescription.mSampleRate) * min(STK_DEFAULT_SECONDS_REQUIRED_TO_START_PLAYING_AFTER_BUFFER_UNDERRUN, STK_DEFAULT_PCM_BUFFER_SIZE_IN_SECONDS))
        framesRequiredBeforeWaitingForDataAfterSeekBecomesPlaying = UInt32(Float32(canonicalAudioStreamBasicDescription.mSampleRate) * min(STK_DEFAULT_GRACE_PERIOD_AFTER_SEEK_SECONDS, STK_DEFAULT_PCM_BUFFER_SIZE_IN_SECONDS))
        
        pcmAudioBufferList.mNumberBuffers = 1
        pcmAudioBufferList.mBuffers.mDataByteSize = UInt32((Float32(canonicalAudioStreamBasicDescription.mSampleRate) * STK_DEFAULT_PCM_BUFFER_SIZE_IN_SECONDS) * Float32(canonicalAudioStreamBasicDescription.mBytesPerFrame))
        pcmAudioBufferList.mBuffers.mData = calloc(Int(pcmAudioBuffer.mDataByteSize), 1)
        pcmAudioBufferList.mBuffers.mNumberChannels = 2
        pcmBufferFrameSizeInBytes = canonicalAudioStreamBasicDescription.mBytesPerFrame
        pcmBufferTotalFrameCount = pcmAudioBuffer.mDataByteSize / pcmBufferFrameSizeInBytes
        readBufferSize = Int(STK_DEFAULT_READ_BUFFER_SIZE)
        readBuffer = malloc(Int(readBufferSize))!.assumingMemoryBound(to: UInt8.self)
        
        var attr = pthread_mutexattr_t()
        pthread_mutexattr_init(&attr)
        pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE)
        pthread_mutex_init(&playerMutex, &attr)
        pthread_mutex_init(&mainThreadSyncCallMutex, nil)
        pthread_cond_init(&playerThreadReadyCondition, nil)
        pthread_cond_init(&mainThreadSyncCallReadyCondition, nil)
        threadStartedLock = NSConditionLock(condition: 0)
        threadFinishedCondLock = NSConditionLock(condition: 0)
        
        self.internalState = .initialised
        
        upcomingQueue = []
        bufferingQueue = []
        resetPcmBuffers()
        
        
        //createPlaybackThread()
    }    
    
    func setInternalState(_ value: STKAudioPlayerInternalState) {
        var newState: STKAudioPlayerState?
        switch value {
        case .initialised :
            newState = .ready
            stopReason = .None
            break
        case .running, .startingThread, .playing, .waitingForDataAfterSeek  :
            newState = .playing
            stopReason = .None
            break
        case .pendingNext, .rebuffering, .waitingForData :
            newState = .buffering
            stopReason = .None
            break
        case .stopped :
            newState = .stopped
            break
        case .paused :
            newState = .paused
            stopReason = .None
            break
        case .disposed :
            newState = .disposed
            stopReason = .userAction
            break
        case .error :
            newState = .error
            stopReason = .error
            break
        }
        os_unfair_lock_lock(&internalStateLock)
        waitingForDataAfterSeekFrameCount = 0
        if value == internalState {
            os_unfair_lock_unlock(&internalStateLock)
            return
        }
        internalState = value
        let previousState = state
        if newState != previousState && !deallocating {
            state = newState!
            os_unfair_lock_unlock(&internalStateLock)
            DispatchQueue.main.async{
                self.delegate.audioPlayer(self, stateChanged: self.state, previousState: previousState)
            }
        }
        else {
            os_unfair_lock_unlock(&internalStateLock)
        }
    }
    
    func destroyAudioResources() {
        if currentlyReadingEntry != nil {
            currentlyReadingEntry.delegate = nil
            currentlyReadingEntry.unregisterForEvents()
            currentlyReadingEntry = nil
        }
        if currentlyPlayingEntry != nil {
            currentlyPlayingEntry.delegate = nil
            currentlyPlayingEntry.unregisterForEvents()
            os_unfair_lock_lock(&currentEntryReferencesLock)
            currentlyPlayingEntry = nil
            os_unfair_lock_unlock(&currentEntryReferencesLock)
        }
        closeRecordAudioFile()
        stopAudioUnit(with: .disposed)
        clearQueue()
        if audioFileStream != nil {
            AudioFileStreamClose(audioFileStream!)
            audioFileStream = nil
        }
        if audioConverterRef != nil {
            AudioConverterDispose(audioConverterRef!)
            audioConverterRef = nil
        }
        //        if audioGraph != nil {
        //            AUGraphUninitialize(audioGraph!)
        //            AUGraphClose(audioGraph!)
        //            DisposeAUGraph(audioGraph!)
        //            audioGraph = nil
        //        }
    }
    
    deinit {
        print("STKAudioPlayer dealloc")
        deallocating = true
        destroyAudioResources()
        pthread_mutex_destroy(&playerMutex)
        pthread_mutex_destroy(&mainThreadSyncCallMutex)
        pthread_cond_destroy(&playerThreadReadyCondition)
        pthread_cond_destroy(&mainThreadSyncCallReadyCondition)
        free(readBuffer)
        free(pcmAudioBufferList.mBuffers.mData)
    }
    
    class func dataSource(from url: URL?) -> STKHTTPDataSource? {
        var retval: STKHTTPDataSource? = nil
        
        if url?.scheme?.caseInsensitiveCompare("http") == .orderedSame || url?.scheme?.caseInsensitiveCompare("https") == .orderedSame {
            if let url = url {
                retval = STKHTTPDataSource.init(urlIn: url)
            }
        }
        return retval
    }
    
    func dataSource(_ dataSource: STKHTTPDataSource, didUpdateMetaData metaDataDictionary: NSDictionary) {
        delegate.audioPlayer(self, didUpdateMetaData: metaDataDictionary)
    }
    
    func clearQueue() {
        pthread_mutex_lock(&playerMutex)
        do {
            upcomingQueue.removeAllObjects()
            bufferingQueue.removeAllObjects()
        }
        pthread_mutex_unlock(&playerMutex)
    }
    
    func STKplay(_ url: URL) {
        STKplay(url, withQueueItemID: url as NSObject)
    }
    
    func STKplay(_ url: URL, withQueueItemID queueItemId: NSObject) {
        setDataSource(STKAudioPlayer.dataSource(from: url)!, withQueueItemId: queueItemId)
    }
    
    func STKplayDS(_ dataSource: STKHTTPDataSource) {
        STKplayDSWQ(dataSource, withQueueItemID: dataSource)
    }
    
    func STKplayDSWQ(_ dataSource: STKHTTPDataSource, withQueueItemID queueItemId: NSObject) {
        setDataSource(dataSource, withQueueItemId: queueItemId)
    }
    
    func setDataSource(_ dataSourceIn: STKHTTPDataSource, withQueueItemId queueItemId: NSObject) {
        pthread_mutex_lock(&playerMutex)
        do {
            clearQueue()
            upcomingQueue.enqueue(dataSourceIn)
            self.internalState = .pendingNext
        }
        pthread_mutex_unlock(&playerMutex)
        wakeupPlaybackThread()
    }
    
    func handlePropertyChangeForFileStream(audioFileStreamIn: AudioFileStreamID, fileStreamPropertyID propertyID: AudioFileStreamPropertyID, ioFlags: UnsafeMutablePointer<AudioFileStreamPropertyFlags>) {
        if currentlyReadingEntry == nil {
            return
        }
        switch propertyID {
        case kAudioFileStreamProperty_DataOffset: dataOffsetChanged()
        case kAudioFileStreamProperty_FileFormat: fileFormatChanged()
        case kAudioFileStreamProperty_DataFormat: dataFormatChanged()
        case kAudioFileStreamProperty_AudioDataByteCount: audioDataByteCountChanged()
        case kAudioFileStreamProperty_ReadyToProducePackets: readyToProducePacketsChanged()
        case kAudioFileStreamProperty_FormatList: formatListChanged()
        default: break
        }
    }
    
    func dataOffsetChanged() {
        guard let stream = audioFileStream else { return }
        var offset: UInt64 = 0
        var offsetSize = UInt32(MemoryLayout.size(ofValue: offset))
        AudioFileStreamGetProperty(stream, kAudioFileStreamProperty_DataOffset, &offsetSize, &offset)
        currentlyReadingEntry.parsedHeader = true
        currentlyReadingEntry.audioDataOffset = UInt64(offset )
    }
    
    func fileFormatChanged() {
        guard let stream = audioFileStream else { return }
        var fileFormat: [UInt8] = Array(repeating: 0, count: 4)
        var fileFormatSize = UInt32(4)
        AudioFileStreamGetProperty(stream, kAudioFileStreamProperty_FileFormat, &fileFormatSize, &fileFormat)
    }
    
    func dataFormatChanged() {
        var error: OSStatus
        var newBasicDescription: AudioStreamBasicDescription = {
            let bytesPerSample = UInt32(MemoryLayout<Int16>.size)            
            let flags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagsNativeEndian | kAudioFormatFlagIsPacked
            let component = AudioStreamBasicDescription(mSampleRate: 44100, mFormatID: kAudioFormatLinearPCM, mFormatFlags: flags, mBytesPerPacket: bytesPerSample * 2, mFramesPerPacket: 1, mBytesPerFrame: bytesPerSample * 2, mChannelsPerFrame: 2, mBitsPerChannel: 8 * bytesPerSample, mReserved: 0)
            return component
        }()
        let entryToUpdate = currentlyReadingEntry
        if !currentlyReadingEntry.parsedHeader {
            var size = UInt32(MemoryLayout.size(ofValue: newBasicDescription))
            AudioFileStreamGetProperty(audioFileStream!, kAudioFileStreamProperty_DataFormat, &size, &newBasicDescription)
            pthread_mutex_lock(&playerMutex)
            if entryToUpdate!.audioStreamBasicDescription.mFormatID == 0 {
                entryToUpdate!.audioStreamBasicDescription = newBasicDescription
            }
            entryToUpdate!.sampleRate = entryToUpdate!.audioStreamBasicDescription.mSampleRate
            entryToUpdate!.packetDuration = Double(entryToUpdate!.audioStreamBasicDescription.mFramesPerPacket) / entryToUpdate!.sampleRate
            var packetBufferSize: UInt32 = 0
            var sizeOfPacketBufferSize = UInt32(MemoryLayout.size(ofValue: packetBufferSize))
            error = AudioFileStreamGetProperty(audioFileStream!, kAudioFileStreamProperty_PacketSizeUpperBound, &sizeOfPacketBufferSize, &packetBufferSize)
            if error != noErr || packetBufferSize == 0 {
                error = AudioFileStreamGetProperty(audioFileStream!, kAudioFileStreamProperty_MaximumPacketSize, &sizeOfPacketBufferSize, &packetBufferSize)
                if error != noErr || packetBufferSize == 0 {
                    entryToUpdate!.packetBufferSize = UInt32(STK_DEFAULT_PACKET_BUFFER_SIZE)
                } else {
                    entryToUpdate!.packetBufferSize = packetBufferSize
                }
            } else {
                entryToUpdate!.packetBufferSize = packetBufferSize
            }
            createAudioConverter(&currentlyReadingEntry.audioStreamBasicDescription)
            pthread_mutex_unlock(&playerMutex)
        }
    }
    
    func audioDataByteCountChanged() {
        guard let stream = audioFileStream else { return }
        var audioDataByteCount: UInt64 = 0
        var byteCountSize = UInt32(MemoryLayout.size(ofValue: audioDataByteCount))
        AudioFileStreamGetProperty(stream, kAudioFileStreamProperty_AudioDataByteCount, &byteCountSize, &audioDataByteCount)
        currentlyReadingEntry.audioDataByteCount = audioDataByteCount
    }
    
    func readyToProducePacketsChanged() {
        if audioConverterAudioStreamBasicDescription.mFormatID != kAudioFormatLinearPCM {
            discontinuous = true
        }
    }
    
    func formatListChanged() {
        guard let stream = audioFileStream else { return }
        var outWriteable: DarwinBoolean = false
        var formatListSize: UInt32 = 0
        guard AudioFileStreamGetPropertyInfo(stream, kAudioFileStreamProperty_FormatList, &formatListSize, &outWriteable) == noErr else { return }
        var formatList: [AudioFormatListItem] = Array.init(repeating: AudioFormatListItem(), count: Int(formatListSize))
        guard AudioFileStreamGetProperty(stream, kAudioFileStreamProperty_FormatList, &formatListSize, &formatList) == noErr else { return }
        let increaseMent = MemoryLayout<AudioFormatListItem>.size
        let end = Int(formatListSize)
        var i = 0
        while i * increaseMent < end {
            let pasbd = formatList[i].mASBD
            if pasbd.mFormatID == kAudioFormatMPEG4AAC_HE || pasbd.mFormatID  == kAudioFormatMPEG4AAC_HE_V2 {
                currentlyReadingEntry!.audioStreamBasicDescription = pasbd;
                break
            }
            i += increaseMent
        }
    }
    
    func currentTimeInFrames() -> Float64 {
        if audioGraph == nil {
            return Float64(0)
        }
        return Float64(0)
    }
    
    func unexpectedError(_ errorCodeIn: STKAudioPlayerErrorCode) {
        internalState = .error
    }
    
    func duration() -> Double {
        if internalState == .pendingNext {
            return 0
        }
        os_unfair_lock_lock(&currentEntryReferencesLock)
        let entry = currentlyPlayingEntry
        os_unfair_lock_unlock(&currentEntryReferencesLock)
        if entry == nil {
            return 0
        }
        let retval = entry!.duration()
        let progress = self.progress()
        if retval < progress && retval > 0 {
            return progress
        }
        return retval
    }
    
    func progress() -> Double {
        if seekToTimeWasRequested {
            return requestedSeekTime
        }
        if internalState == .pendingNext {
            return 0
        }
        os_unfair_lock_lock(&currentEntryReferencesLock)
        let entry = currentlyPlayingEntry
        os_unfair_lock_unlock(&currentEntryReferencesLock)
        if entry == nil {
            return 0
        }
        os_unfair_lock_lock(&entry!.spinLock!)
        let retval: Double = entry!.seekTime + (Float64(entry!.framesPlayed) / canonicalAudioStreamBasicDescription.mSampleRate)
        os_unfair_lock_unlock(&entry!.spinLock!)
        return retval
    }
    
    
    func invoke(onPlaybackThread block: @escaping () -> Void) -> Bool {
        let runLoop = playbackThreadRunLoop
        if runLoop != nil {
            CFRunLoopPerformBlock(runLoop!.getCFRunLoop(), RunLoop.Mode.common as CFTypeRef, block)
            CFRunLoopWakeUp(runLoop!.getCFRunLoop())
            return true
        }
        return false
    }
    
    func wakeupPlaybackThread() {
        _ = invoke(onPlaybackThread: {
            _ = self.processRunloop()
        })
        pthread_mutex_lock(&playerMutex)
        if waiting {
            pthread_cond_signal(&playerThreadReadyCondition)
        }
        pthread_mutex_unlock(&playerMutex)
    }
    
    func seek(toTime value: Double) {
        if currentlyPlayingEntry == nil {
            return
        }
        os_unfair_lock_lock(&seekLock)
        let seekAlreadyRequested = seekToTimeWasRequested
        seekToTimeWasRequested = true
        requestedSeekTime = value
        if !seekAlreadyRequested {
            seekVersion!.mutate { $0 += 1 }
            os_unfair_lock_unlock(&seekLock)
            wakeupPlaybackThread()
            return
        }
        os_unfair_lock_unlock(&seekLock)
    }
    
    func createPlaybackThread() {
        playbackThread = Thread(target: self, selector: #selector(self.startInternal), object: nil)
        playbackThread!.start()
        threadStartedLock!.lock(whenCondition: 1)
        threadStartedLock!.unlock(withCondition: 0)
        assert(playbackThreadRunLoop != nil, "playbackThreadRunLoop != nil")
    }
    
    func audioQueueFinishedPlaying(_ entry: STKHTTPDataSource) {
        let next = bufferingQueue.dequeue() as? STKHTTPDataSource
        processFinishPlayingIfAnyAndPlayingNext(entry, withNext: next)
        _ = processRunloop()
    }
    
    func setCurrentlyReading(_ entry: STKHTTPDataSource, andStartPlaying startPlaying: Bool) {
        setCurrentlyReading(entry, andStartPlaying: startPlaying, clearQueue: true)
    }
    
    func setCurrentlyReading(_ entry: STKHTTPDataSource?, andStartPlaying startPlaying: Bool, clearQueue: Bool) {
        if startPlaying {
            memset(UnsafeMutableRawPointer(pcmAudioBuffer.mData), 0,  Int(pcmBufferTotalFrameCount * pcmBufferFrameSizeInBytes))
        }
        if audioFileStream != nil {
            AudioFileStreamClose(audioFileStream!)
            audioFileStream = nil
        }
        if currentlyReadingEntry != nil {
            currentlyReadingEntry.delegate = nil
            currentlyReadingEntry.close()
        }
        os_unfair_lock_lock(&currentEntryReferencesLock)
        currentlyReadingEntry = entry
        os_unfair_lock_unlock(&currentEntryReferencesLock)
        currentlyReadingEntry.delegate = self
        _ = currentlyReadingEntry.register(forEvents: RunLoop.current)
        currentlyReadingEntry.seek(toOffset: 0)
        closeRecordAudioFile()
        if startPlaying {
            if clearQueue {
                self.clearQueue()
            }
            processFinishPlayingIfAnyAndPlayingNext(currentlyPlayingEntry, withNext: entry)
            _ = startAudioGraph()
        }
        else {
            bufferingQueue.enqueue(entry!)
        }
    }
    
    func processFinishPlayingIfAnyAndPlayingNext(_ entry: STKHTTPDataSource?, withNext next: STKHTTPDataSource?) {
        if entry != currentlyPlayingEntry {
            return
        }
        let isPlayingSameItemProbablySeek: Bool = currentlyPlayingEntry == next
        if next != nil {
            if !isPlayingSameItemProbablySeek {
                os_unfair_lock_lock(&next!.spinLock!)
                next!.seekTime = 0
                os_unfair_lock_unlock(&next!.spinLock!)
                os_unfair_lock_lock(&seekLock)
                seekToTimeWasRequested = false
                os_unfair_lock_unlock(&seekLock)
            }
            os_unfair_lock_lock(&currentEntryReferencesLock)
            currentlyPlayingEntry = next
            os_unfair_lock_unlock(&currentEntryReferencesLock)
            if !isPlayingSameItemProbablySeek {
                self.setInternalState(.waitingForData)
            }
        }
        else {
            os_unfair_lock_lock(&currentEntryReferencesLock)
            currentlyPlayingEntry = nil
            os_unfair_lock_unlock(&currentEntryReferencesLock)
        }
        wakeupPlaybackThread()
    }
    
    func dispatchSync(onMainThread block: @escaping () -> Void) {
        var finished = false
        if disposeWasRequested {
            return
        }
        DispatchQueue.main.async{
            if !self.disposeWasRequested {
                block()
            }
            pthread_mutex_lock(&self.mainThreadSyncCallMutex)
            finished = true
            pthread_cond_signal(&self.mainThreadSyncCallReadyCondition)
            pthread_mutex_unlock(&self.mainThreadSyncCallMutex)
        }
        pthread_mutex_lock(&mainThreadSyncCallMutex)
        while true {
            if disposeWasRequested {
                break
            }
            if finished {
                break
            }
            pthread_cond_wait(&mainThreadSyncCallReadyCondition, &mainThreadSyncCallMutex)
        }
        pthread_mutex_unlock(&mainThreadSyncCallMutex)
    }
    
    func playbackThreadQueueMainThreadSyncBlock(_ block: @escaping () -> Void) {
        _ = invoke(onPlaybackThread: {
            if self.disposeWasRequested {
                return
            }
            self.dispatchSync(onMainThread: block)
        })
    }
    
    func requeueBufferingEntries() {
        if bufferingQueue.count > 0 {
            for queueEntry in bufferingQueue {
                (queueEntry as! STKHTTPDataSource).parsedHeader = false
                (queueEntry as! STKHTTPDataSource).reset()
            }
            upcomingQueue.skipQueue(withQueue: bufferingQueue)
            bufferingQueue.removeAllObjects()
        }
    }
    
    func processRunloop() -> Bool {
        pthread_mutex_lock(&playerMutex)
        if disposeWasRequested {
            pthread_mutex_unlock(&playerMutex)
            return false
        }
        else if self.internalState == .paused {
            pthread_mutex_unlock(&playerMutex)
            return true
        }
        else if self.internalState == .pendingNext {
            let entry = upcomingQueue.dequeue() as? STKHTTPDataSource
            self.internalState = .waitingForData
            setCurrentlyReading(entry!, andStartPlaying: true)
            resetPcmBuffers()
        }
        else if seekToTimeWasRequested && currentlyPlayingEntry != nil && currentlyPlayingEntry != currentlyReadingEntry {
            currentlyPlayingEntry.parsedHeader = false
            currentlyPlayingEntry.reset()
            if currentlyReadingEntry != nil {
                currentlyReadingEntry.delegate = nil
                currentlyReadingEntry.unregisterForEvents()
            }
            self.internalState = .waitingForDataAfterSeek
            setCurrentlyReading(currentlyPlayingEntry, andStartPlaying: true, clearQueue: true)
        }
        else if currentlyReadingEntry == nil {
            if upcomingQueue.count > 0 {
                let entry = upcomingQueue.dequeue() as? STKHTTPDataSource
                let startPlaying: Bool = currentlyPlayingEntry == nil
                self.internalState = .waitingForData
                setCurrentlyReading(entry!, andStartPlaying: startPlaying)
            }
            else if currentlyPlayingEntry == nil {
                if self.internalState != .stopped {
                    stopAudioUnit(with: .eof)
                }
            }
        }
        if currentlyPlayingEntry != nil && currentlyPlayingEntry.parsedHeader && currentlyPlayingEntry.calculatedBitRate() > 0.0 {
            var originalSeekVersion: Atomic<Int>?
            var originalSeekToTimeRequested = false
            os_unfair_lock_lock(&seekLock)
            originalSeekVersion = seekVersion
            originalSeekToTimeRequested = seekToTimeWasRequested
            os_unfair_lock_unlock(&seekLock)
            if originalSeekToTimeRequested && currentlyReadingEntry == currentlyPlayingEntry {
                processSeekToTime()
                os_unfair_lock_lock(&seekLock)
                if originalSeekVersion?.value == seekVersion?.value {
                    seekToTimeWasRequested = false
                }
                os_unfair_lock_unlock(&seekLock)
            }
        }
        else if currentlyPlayingEntry == nil && seekToTimeWasRequested {
            seekToTimeWasRequested = false
        }
        pthread_mutex_unlock(&playerMutex)
        return true
    }
    
    @objc func startInternal() {
        autoreleasepool {
            playbackThreadRunLoop = RunLoop.current
            Thread.current.threadPriority = 1
            threadStartedLock!.lock(whenCondition: 0)
            threadStartedLock!.unlock(withCondition: 1)
            playbackThreadRunLoop!.add(NSMachPort(), forMode: .default)
            var looped: Bool = true
            while looped == true  {
                autoreleasepool {
                    if !self.processRunloop() {
                        looped = false
                    }
                }
                let date = Date(timeIntervalSinceNow: 10)
                playbackThreadRunLoop!.run(mode: .default, before: date)
            }
            disposeWasRequested = false
            seekToTimeWasRequested = false
            if currentlyReadingEntry != nil { currentlyReadingEntry.unregisterForEvents() }
            if currentlyPlayingEntry != nil { currentlyPlayingEntry.unregisterForEvents() }
            
            if currentlyReadingEntry != nil { currentlyReadingEntry.delegate = nil }
            if currentlyPlayingEntry != nil { currentlyPlayingEntry.delegate = nil }
            pthread_mutex_lock(&playerMutex)
            os_unfair_lock_lock(&currentEntryReferencesLock)
            currentlyPlayingEntry = nil
            currentlyReadingEntry = nil
            os_unfair_lock_unlock(&currentEntryReferencesLock)
            pthread_mutex_unlock(&playerMutex)
            self.internalState = .disposed
            playbackThreadRunLoop = nil
            destroyAudioResources()
            threadFinishedCondLock!.lock()
            threadFinishedCondLock!.unlock(withCondition: 1)
        }
    }
    
    func processSeekToTime() {
        var error: OSStatus = 0
        let currentEntry = currentlyReadingEntry
        assert(currentEntry == currentlyPlayingEntry, "playing and reading must be the same")
        if currentEntry == nil || (currentEntry!.calculatedBitRate() == 0.0 || currentlyPlayingEntry.length()! <= 0) {
            return
        }
        var seekByteOffset = Int((Double(currentEntry!.audioDataOffset) + (requestedSeekTime / duration()) * Double((currentlyReadingEntry.audioDataLengthInBytes()))))
        if Int(seekByteOffset) > currentEntry!.length()! - Int((2 * currentEntry!.packetBufferSize)) {
            seekByteOffset = currentEntry!.length()! - Int(2 * currentEntry!.packetBufferSize)
        }
        os_unfair_lock_lock(&currentEntry!.spinLock!)
        currentEntry!.seekTime = requestedSeekTime
        os_unfair_lock_unlock(&currentEntry!.spinLock!)
        let calculatedBitRate: Double = currentEntry!.calculatedBitRate()
        if currentEntry!.packetDuration > 0 && calculatedBitRate > 0 {
            var ioFlags = AudioFileStreamSeekFlags(rawValue: 0)
            var packetAlignedByteOffset: Int64 = 0
            let seekPacket = Int64(floor(requestedSeekTime / currentEntry!.packetDuration))
            error = AudioFileStreamSeek(audioFileStream!, seekPacket, &packetAlignedByteOffset, &ioFlags)
            if error == noErr && (ioFlags.rawValue & AudioFileStreamSeekFlags.offsetIsEstimated.rawValue) == 0 {
                var delta: Double? = nil
                if let audioDataOffset: UInt64 = currentEntry?.audioDataOffset {
                    delta = Double(((seekByteOffset - Int(audioDataOffset)) - Int(packetAlignedByteOffset)) / Int(calculatedBitRate)) * 8
                }
                os_unfair_lock_lock(&currentEntry!.spinLock!)
                currentEntry!.seekTime -= delta!
                os_unfair_lock_unlock(&currentEntry!.spinLock!)
                if let audioDataOffset: UInt64 = currentEntry?.audioDataOffset {
                    seekByteOffset = Int(packetAlignedByteOffset) + Int(audioDataOffset)
                }
            }
        }
        if audioConverterRef != nil {
            AudioConverterReset(audioConverterRef!)
        }
        if recordAudioConverterRef != nil {
            AudioConverterReset(recordAudioConverterRef!)
        }
        currentEntry!.reset()
        currentEntry!.seek(toOffset: seekByteOffset)
        waitingForDataAfterSeekFrameCount = 0
        internalState = .waitingForDataAfterSeek
        if audioGraph != nil {
            resetPcmBuffers()
        }
    }
    
    func dataSourceDataAvailable(_ dataSource: STKHTTPDataSource) {
        var error: OSStatus = 0
        if currentlyReadingEntry != dataSource {
            return
        }
        if !currentlyReadingEntry.hasBytesAvailable {
            return
        }
        let read = currentlyReadingEntry.read(intoBuffer: readBuffer, withSize: Int32(readBufferSize))
        if read == 0 {
            return
        }
        if audioFileStream == nil {
            error = AudioFileStreamOpen(UnsafeMutableRawPointer(Unmanaged<STKAudioPlayer>.passUnretained(self).toOpaque()), audioFileStreamPropertyListenerProc, audioFileStreamPacketsProc, dataSource.audioFileTypeHint!, &audioFileStream)
            if error != noErr {
                unexpectedError(.stkAudioPlayerErrorAudioSystemError)
                return
            }
        }
        if read < 0 {
            // iOS will shutdown network connections if the app is backgrounded (i.e. device is locked when player is paused)
            // We try to reopen -- should probably add a back-off protocol in the future
            let position = currentlyReadingEntry.position()
            currentlyReadingEntry.seek(toOffset: Int(position))
            return
        }
        let flags = discontinuous ? AudioFileStreamParseFlags.discontinuity :  AudioFileStreamParseFlags(rawValue: 0)
        
        if audioFileStream != nil {
            error = AudioFileStreamParseBytes(audioFileStream!, UInt32(read), readBuffer, flags)
            if error != noErr {
                if dataSource == currentlyPlayingEntry {
                    unexpectedError(.stkAudioPlayerErrorStreamParseBytesFailed)
                }
                return
            }
            os_unfair_lock_lock(&currentEntryReferencesLock)
            if currentlyReadingEntry == nil {
                dataSource.unregisterForEvents()
                dataSource.close()
            }
            os_unfair_lock_unlock(&currentEntryReferencesLock)
        }
    }
    
    func dataSourceErrorOccured(_ dataSource: STKHTTPDataSource) {
        if currentlyReadingEntry != dataSource {
            return
        }
        unexpectedError(.stkAudioPlayerErrorDataNotFound)
    }
    
    func dataSourceEof(_ dataSource: STKHTTPDataSource) {
        if currentlyReadingEntry == nil || currentlyReadingEntry != dataSource {
            dataSource.delegate = nil
            dataSource.unregisterForEvents()
            dataSource.close()
            return
        }
        if disposeWasRequested {
            return
        }
        closeRecordAudioFile()
        pthread_mutex_lock(&playerMutex)
        if currentlyReadingEntry == nil {
            dataSource.delegate = nil
            dataSource.unregisterForEvents()
            dataSource.close()
            return
        }
        os_unfair_lock_lock(&currentlyReadingEntry.spinLock!)
        currentlyReadingEntry.lastFrameQueued = currentlyReadingEntry.framesQueued
        os_unfair_lock_unlock(&currentlyReadingEntry.spinLock!)
        currentlyReadingEntry.delegate = nil
        currentlyReadingEntry.unregisterForEvents()
        currentlyReadingEntry.close()
        os_unfair_lock_lock(&currentEntryReferencesLock)
        currentlyReadingEntry = nil
        os_unfair_lock_unlock(&currentEntryReferencesLock)
        pthread_mutex_unlock(&playerMutex)
        _ = processRunloop()
    }
    
    func pause() {
        pthread_mutex_lock(&playerMutex)
        do {
            var error: OSStatus = 0
            if internalState != .paused && (internalState.rawValue & STKAudioPlayerInternalState.running.rawValue) != 0 {
                stateBeforePaused = internalState
                internalState = .paused
            }
            if audioGraph != nil {
                error = AUGraphStop(audioGraph!)
                if error != noErr {
                    unexpectedError(.stkAudioPlayerErrorAudioSystemError)
                    pthread_mutex_unlock(&playerMutex)
                    return
                }
            }
            wakeupPlaybackThread()
        }
        pthread_mutex_unlock(&playerMutex)
    }
    
    func resume() {
        pthread_mutex_lock(&playerMutex)
        do {
            var error: OSStatus = 0
            if internalState == .paused {
                internalState = stateBeforePaused
                if seekToTimeWasRequested {
                    resetPcmBuffers()
                }
                if audioGraph != nil {
                    error = AUGraphStart(audioGraph!)
                    if error != noErr {
                        unexpectedError(.stkAudioPlayerErrorAudioSystemError)
                        pthread_mutex_unlock(&playerMutex)
                        return
                    }
                }
                wakeupPlaybackThread()
            }
        }
        pthread_mutex_unlock(&playerMutex)
    }
    
    func resetPcmBuffers() {
        os_unfair_lock_lock(&pcmBufferSpinLock)
        pcmBufferFrameStartIndex = 0
        pcmBufferUsedFrameCount = 0
        os_unfair_lock_unlock(&pcmBufferSpinLock)
    }
    
    func stop() {
        pthread_mutex_lock(&playerMutex)
        do {       
            if self.internalState == .stopped
            {
                pthread_mutex_unlock(&playerMutex)
                return
            }
            closeRecordAudioFile()
            stopAudioUnit(with: .userAction)
            resetPcmBuffers()
            _ = invoke(onPlaybackThread: {
                pthread_mutex_lock(&self.playerMutex)
                do {
                    self.currentlyReadingEntry.delegate = nil
                    self.currentlyReadingEntry.unregisterForEvents()
                    self.currentlyReadingEntry.close()
                    if self.currentlyPlayingEntry != nil {
                        self.processFinishPlayingIfAnyAndPlayingNext(self.currentlyPlayingEntry, withNext: nil)
                    }
                    self.clearQueue()
                    os_unfair_lock_lock(&self.currentEntryReferencesLock)
                    self.currentlyPlayingEntry = nil
                    self.currentlyReadingEntry = nil
                    self.seekToTimeWasRequested = false
                    os_unfair_lock_unlock(&self.currentEntryReferencesLock)
                }
                pthread_mutex_unlock(&self.playerMutex)
            })
            wakeupPlaybackThread()
        }
        pthread_mutex_unlock(&playerMutex)
    }
    
    func stopThread() {
        var wait = false
        var runLoop = playbackThreadRunLoop
        if runLoop != nil
        {
            wait = true
            _ = invoke(onPlaybackThread: {
                self.disposeWasRequested = true
                CFRunLoopStop(CFRunLoopGetCurrent())
            })
            pthread_mutex_lock(&playerMutex)
            disposeWasRequested = true
            pthread_cond_signal(&playerThreadReadyCondition)
            pthread_mutex_unlock(&playerMutex)
            
            pthread_mutex_lock(&mainThreadSyncCallMutex)
            disposeWasRequested = true
            pthread_cond_signal(&mainThreadSyncCallReadyCondition)
            pthread_mutex_unlock(&mainThreadSyncCallMutex)
        }
        if wait {
            threadFinishedCondLock!.lock(whenCondition: 1)
            threadFinishedCondLock!.unlock(withCondition: 0)
        }
        destroyAudioResources()
        runLoop = nil
        playbackThread = nil
        playbackThreadRunLoop = nil
    }
    
    func closeRecordAudioFile() {
        if recordAudioFileId != nil {
            AudioFileClose(recordAudioFileId!)
            recordAudioFileId = nil
        }
        if recordAudioConverterRef != nil {
            AudioConverterDispose(recordAudioConverterRef!)
            recordAudioConverterRef = nil
        }
        if recordOutputBuffer != nil {
            free(recordOutputBuffer)
            recordOutputBuffer = nil
        }
        
        if recordPacketDescriptions != nil {
            recordPacketDescriptions = nil
        }
        recordFilePacketPosition = 0
    }
    
    func dispose() {
        stop()
        stopThread()
    }
    
    //    func currentlyPlayingQueueItemId() -> NSObject? {
    //        os_unfair_lock_lock(&currentEntryReferencesLock)
    //        let entry = currentlyPlayingEntry
    //        if entry == nil {
    //            os_unfair_lock_unlock(&currentEntryReferencesLock)
    //            return nil
    //        }
    //        let retval: NSObject = entry!.queueItemId
    //        os_unfair_lock_unlock(&currentEntryReferencesLock)
    //        return retval
    //    }
    
    func destroyAudioConverter() {
        if audioConverterRef != nil {
            AudioConverterDispose(audioConverterRef!)
            audioConverterRef = nil
        }
    }
    
    func createAudioConverter(_ asbd: inout AudioStreamBasicDescription) {
        var status: OSStatus = 0
        let size = MemoryLayout.size(ofValue: asbd)
        if memcmp(&asbd, &audioConverterAudioStreamBasicDescription, size) == 0, let converter = audioConverterRef
        {
            AudioConverterReset(converter)
            if recordAudioConverterRef != nil {
                AudioConverterReset(recordAudioConverterRef!)
            }
            return
        }
        destroyAudioConverter()
        let isRecording: Bool = currentlyReadingEntry.recordToFileUrl != nil
        if isRecording {
            recordAudioStreamBasicDescription = AudioStreamBasicDescription()
            recordAudioStreamBasicDescription.mFormatID = kAudioFormatMPEG4AAC
            recordAudioStreamBasicDescription.mFormatFlags = AudioFormatFlags(MPEG4ObjectID.AAC_LC.rawValue)
            recordAudioStreamBasicDescription.mChannelsPerFrame = canonicalAudioStreamBasicDescription.mChannelsPerFrame
            recordAudioStreamBasicDescription.mSampleRate = canonicalAudioStreamBasicDescription.mSampleRate
            var dataSize = UInt32(MemoryLayout.size(ofValue: recordAudioStreamBasicDescription))
            AudioFormatGetProperty(kAudioFormatProperty_FormatInfo, 0, nil, &dataSize, &recordAudioStreamBasicDescription)
        }
        if audioConverterRef == nil {
            status = AudioConverterNew(&asbd, &canonicalAudioStreamBasicDescription, &audioConverterRef)
            if status != noErr {
                unexpectedError(.stkAudioPlayerErrorAudioSystemError)
                return
            }
        }
        
        if isRecording && recordAudioConverterRef == nil {
            status = AudioConverterNew(&canonicalAudioStreamBasicDescription, &recordAudioStreamBasicDescription, &recordAudioConverterRef)
            if status != noErr {
                print("STKAudioPlayer failed to create a recording audio converter")
            }
        }
        audioConverterAudioStreamBasicDescription = asbd
        if currentlyReadingEntry.audioFileTypeHint != kAudioFileAAC_ADTSType {
            var writable: DarwinBoolean = false
            var cookieSize: UInt32 = 0
            status = AudioFileStreamGetPropertyInfo(audioFileStream!, kAudioFileStreamProperty_MagicCookieData, &cookieSize, &writable)
            guard status == noErr else { return }
            var cookiesData: [UInt8] = Array.init(repeating: 0, count: Int(cookieSize))
            status = AudioFileStreamGetProperty(audioFileStream!, kAudioConverterDecompressionMagicCookie, &cookieSize, &cookiesData)
            guard status == noErr, let _audioConverterRef = audioConverterRef else { return }
            status = AudioConverterSetProperty(_audioConverterRef, kAudioConverterDecompressionMagicCookie, cookieSize, &cookiesData)
            if status != noErr {
                unexpectedError(.stkAudioPlayerErrorAudioSystemError)
                return
            }
        }
        
        if recordAudioConverterRef != nil {
            if recordAudioFileId != nil {
                AudioFileClose(recordAudioFileId!)
                recordAudioFileId = nil
            }
            
            if recordOutputBuffer != nil {
                free(recordOutputBuffer)
                recordOutputBuffer = nil
            }
            
            if recordPacketDescriptions != nil {
                recordPacketDescriptions = nil
            }
            
            recordOutputBufferSize = 32 * 1024
            recordPacketSize = canonicalAudioStreamBasicDescription.mBytesPerPacket
            if recordPacketSize == 0 {
                var size = UInt32(MemoryLayout.size(ofValue: recordPacketSize))
                if 0 == AudioConverterGetProperty(recordAudioConverterRef!, kAudioConverterPropertyMaximumOutputPacketSize, &size, &recordPacketSize) {
                    if recordPacketSize > recordOutputBufferSize {
                        recordOutputBufferSize = recordPacketSize
                    }
                    recordPacketsPerBuffer = recordOutputBufferSize / recordPacketSize
                } else {
                    AudioConverterDispose(recordAudioConverterRef!)
                    recordAudioConverterRef = nil
                    print("STKAudioPlayer: Can't support this output format for recording")
                }
            }
            else {
                recordPacketsPerBuffer = recordOutputBufferSize / recordPacketSize
            }
            var propertySize = UInt32(MemoryLayout<UInt32>.size)
            var externallyFramed: UInt32 = 0
            var error: OSStatus = AudioFormatGetProperty(kAudioFormatProperty_FormatIsExternallyFramed, UInt32(MemoryLayout.size(ofValue: recordAudioStreamBasicDescription)), &recordAudioStreamBasicDescription, &propertySize, &externallyFramed)
            if externallyFramed != 0 {
                recordPacketDescriptions = AudioStreamPacketDescription()
            }
            recordOutputBuffer = UnsafeMutablePointer<UInt8>.allocate(capacity: MemoryLayout<UInt8>.size * Int(recordOutputBufferSize))
            error = AudioFileCreateWithURL((currentlyReadingEntry!.recordToFileUrl)! as CFURL, kAudioFileCAFType, &recordAudioStreamBasicDescription, AudioFileFlags.eraseFile, &recordAudioFileId)
            recordFilePacketPosition = 0
            if error != noErr {
                if let recordToFileUrl = currentlyReadingEntry.recordToFileUrl {
                    print("STKAudioPlayer failed to create a recording audio file at \(recordToFileUrl)")
                }
                closeRecordAudioFile()
            }
        }
    }
    
    func audioGraphIsRunning() -> DarwinBoolean {
        var status: OSStatus = 0
        var isRunning: DarwinBoolean = false
        status = AUGraphIsRunning(audioGraph!, &isRunning)
        if status != noErr {
            return false
        }
        return isRunning
    }
    
    func startAudioGraph() -> Bool {
        var status: OSStatus = 0
        resetPcmBuffers()
        if audioGraphIsRunning().boolValue {
            return false
        }
        status = AUGraphStart(audioGraph!)
        if status != noErr
        {
            unexpectedError(.stkAudioPlayerErrorAudioSystemError)
            return false
        }
        return true
    }
    
    func stopAudioUnit(with stopReasonIn: STKAudioPlayerStopReason) {
        var status: OSStatus = 0
        if audioGraph == nil {
            stopReason = stopReasonIn
            internalState = .stopped
            return
        }
        var isRunning: DarwinBoolean = false
        status = AUGraphIsRunning(audioGraph!, &isRunning)
        if status != noErr {
            unexpectedError(.stkAudioPlayerErrorAudioSystemError)
        }
        else if !isRunning.boolValue {
            stopReason = stopReasonIn
            internalState = .stopped
            return
        }
        status = AUGraphStop(audioGraph!)
        if status != noErr {
            unexpectedError(.stkAudioPlayerErrorAudioSystemError)
        }
        resetPcmBuffers()
        stopReason = stopReasonIn
        internalState = .stopped
    }
    
    func handleAudioPackets(inputData: UnsafeRawPointer, numberBytes: UInt32, numberPackets: UInt32, packetDescriptions: UnsafeMutablePointer<AudioStreamPacketDescription>?) {
        
        if currentlyReadingEntry == nil {
            return
        }
        if !currentlyReadingEntry.parsedHeader {
            return
        }
        if disposeWasRequested {
            return
        }
        if audioConverterRef == nil {
            return
        }
        if (seekToTimeWasRequested && currentlyPlayingEntry.calculatedBitRate() > 0.0) {
            wakeupPlaybackThread()
            return
        }
        discontinuous = false
        var status: OSStatus = 0
        var convertInfo = AudioConvertInfo()
        convertInfo.numberOfPackets = numberPackets
        convertInfo.packetDescriptions = packetDescriptions
        convertInfo.audioBuffer.mData = UnsafeMutableRawPointer(mutating: inputData)
        convertInfo.audioBuffer.mDataByteSize = numberBytes
        convertInfo.audioBuffer.mNumberChannels = audioConverterAudioStreamBasicDescription.mChannelsPerFrame
        var convertInfoPointer = unsafeBitCast(convertInfo, to: UnsafeMutableRawPointer.self)
        if packetDescriptions != nil && currentlyReadingEntry.processedPacketsCount < STK_MAX_COMPRESSED_PACKETS_FOR_BITRATE_CALCULATION {
            let count = Int(min(numberPackets, UInt32(STK_MAX_COMPRESSED_PACKETS_FOR_BITRATE_CALCULATION - UInt32(currentlyReadingEntry.processedPacketsCount))))
            for i in 0..<count {
                var packetSize: UInt32 = 0
                packetSize = packetDescriptions![i].mDataByteSize
                currentlyReadingEntry.processedPacketsSizeTotal += Int(packetSize )
            }
        }
        while true {
            os_unfair_lock_lock(&pcmBufferSpinLock)
            var used: UInt32 = pcmBufferUsedFrameCount
            var start: UInt32 = pcmBufferFrameStartIndex
            var end: UInt32 = (pcmBufferFrameStartIndex + pcmBufferUsedFrameCount) % pcmBufferTotalFrameCount
            var framesLeftInsideBuffer: UInt32 = pcmBufferTotalFrameCount - used
            os_unfair_lock_unlock(&pcmBufferSpinLock)
            if framesLeftInsideBuffer == 0 {
                pthread_mutex_lock(&playerMutex)
                while true {
                    os_unfair_lock_lock(&pcmBufferSpinLock)
                    used = pcmBufferUsedFrameCount
                    start = pcmBufferFrameStartIndex
                    end = (pcmBufferFrameStartIndex + pcmBufferUsedFrameCount) % pcmBufferTotalFrameCount
                    framesLeftInsideBuffer = pcmBufferTotalFrameCount - used
                    os_unfair_lock_unlock(&pcmBufferSpinLock)
                    if framesLeftInsideBuffer > 0 {
                        break
                    }
                    if disposeWasRequested || internalState == .stopped || internalState == .disposed || internalState == .pendingNext {
                        pthread_mutex_unlock(&playerMutex)
                        return
                    }
                    if seekToTimeWasRequested && currentlyPlayingEntry.calculatedBitRate() > 0.0 {
                        pthread_mutex_unlock(&playerMutex)
                        wakeupPlaybackThread()
                        return
                    }
                    waiting = true
                    pthread_cond_wait(&playerThreadReadyCondition, &playerMutex)
                    waiting = false
                }
                pthread_mutex_unlock(&playerMutex)
            }
            var localPcmBufferList: AudioBufferList = AudioBufferList(mNumberBuffers: 1, mBuffers: AudioBuffer())
            var localPcmAudioBuffer: AudioBuffer {
                get { return localPcmBufferList.mBuffers }
                set { localPcmBufferList.mBuffers = newValue }
            }
            if end >= start {
                var framesAdded: UInt32 = 0
                var framesToDecode: UInt32 = pcmBufferTotalFrameCount - end
                localPcmAudioBuffer.mData = pcmAudioBuffer.mData!.advanced(by: (Int(end * pcmBufferFrameSizeInBytes)))
                localPcmAudioBuffer.mDataByteSize = framesToDecode * pcmBufferFrameSizeInBytes
                localPcmAudioBuffer.mNumberChannels = pcmAudioBuffer.mNumberChannels
                status = AudioConverterFillComplexBuffer(audioConverterRef!, AudioConvertInfo.callbak(), &convertInfoPointer, &framesToDecode, &localPcmBufferList, nil)
                framesAdded = framesToDecode
                if (status == 100 || status == 0) && recordAudioFileId != nil && recordAudioConverterRef != nil {
                    handleRecordingOfAudioPackets(framesToDecode, audioBuffer: localPcmBufferList.mBuffers)
                }
                if status == 100 {
                    os_unfair_lock_lock(&pcmBufferSpinLock)
                    pcmBufferUsedFrameCount += framesAdded
                    os_unfair_lock_unlock(&pcmBufferSpinLock)
                    os_unfair_lock_lock(&currentlyReadingEntry.spinLock!)
                    currentlyReadingEntry.framesQueued += Int32(framesAdded)
                    os_unfair_lock_unlock(&currentlyReadingEntry.spinLock!)
                    return
                } else if status != noErr {
                    unexpectedError(.cError)
                    return
                }
                framesToDecode = start
                if framesToDecode == 0 {
                    os_unfair_lock_lock(&pcmBufferSpinLock)
                    pcmBufferUsedFrameCount += framesAdded
                    os_unfair_lock_unlock(&pcmBufferSpinLock)
                    os_unfair_lock_lock(&currentlyReadingEntry.spinLock!)
                    currentlyReadingEntry.framesQueued += Int32(framesAdded)
                    os_unfair_lock_unlock(&currentlyReadingEntry.spinLock!)
                    continue
                }
                localPcmAudioBuffer.mData = pcmAudioBuffer.mData
                localPcmAudioBuffer.mDataByteSize = framesToDecode * pcmBufferFrameSizeInBytes
                localPcmAudioBuffer.mNumberChannels = pcmAudioBuffer.mNumberChannels
                status = AudioConverterFillComplexBuffer(audioConverterRef!, AudioConvertInfo.callbak(), &convertInfoPointer, &framesToDecode, &localPcmBufferList, nil)
                framesAdded += framesToDecode
                if (status == 100 || status == 0) && recordAudioFileId != nil && recordAudioConverterRef != nil {
                    handleRecordingOfAudioPackets(framesToDecode, audioBuffer: localPcmBufferList.mBuffers)
                }
                if status == 100 {
                    os_unfair_lock_lock(&pcmBufferSpinLock)
                    pcmBufferUsedFrameCount += framesAdded
                    os_unfair_lock_unlock(&pcmBufferSpinLock)
                    os_unfair_lock_lock(&currentlyReadingEntry.spinLock!)
                    currentlyReadingEntry.framesQueued += Int32(framesAdded)
                    os_unfair_lock_unlock(&currentlyReadingEntry.spinLock!)
                    return
                } else if status == noErr {
                    os_unfair_lock_lock(&pcmBufferSpinLock)
                    pcmBufferUsedFrameCount += framesAdded
                    os_unfair_lock_unlock(&pcmBufferSpinLock)
                    os_unfair_lock_lock(&currentlyReadingEntry.spinLock!)
                    currentlyReadingEntry.framesQueued += Int32(framesAdded)
                    os_unfair_lock_unlock(&currentlyReadingEntry.spinLock!)
                    continue
                } else if status != noErr {
                    unexpectedError(.cError)
                    return
                }
            }
            else {
                var framesAdded: UInt32 = 0
                var framesToDecode: UInt32 = start - end
                localPcmAudioBuffer.mData = pcmAudioBuffer.mData!.advanced(by: (Int(end * pcmBufferFrameSizeInBytes)))
                localPcmAudioBuffer.mDataByteSize = framesToDecode * pcmBufferFrameSizeInBytes
                localPcmAudioBuffer.mNumberChannels = pcmAudioBuffer.mNumberChannels
                status = AudioConverterFillComplexBuffer(audioConverterRef!, AudioConvertInfo.callbak(), &convertInfoPointer, &framesToDecode, &localPcmBufferList, nil)
                
                framesAdded = framesToDecode
                if (status == 100 || status == 0) && recordAudioFileId != nil && recordAudioConverterRef != nil {
                    handleRecordingOfAudioPackets(framesToDecode, audioBuffer: localPcmBufferList.mBuffers)
                }
                if status == 100 {
                    os_unfair_lock_lock(&pcmBufferSpinLock)
                    pcmBufferUsedFrameCount += framesAdded
                    os_unfair_lock_unlock(&pcmBufferSpinLock)
                    
                    os_unfair_lock_lock(&currentlyReadingEntry.spinLock!)
                    currentlyReadingEntry.framesQueued += Int32(framesAdded)
                    os_unfair_lock_unlock(&currentlyReadingEntry.spinLock!)
                    
                    return
                } else if status == noErr {
                    os_unfair_lock_lock(&pcmBufferSpinLock)
                    pcmBufferUsedFrameCount += framesAdded
                    os_unfair_lock_unlock(&pcmBufferSpinLock)
                    
                    os_unfair_lock_lock(&currentlyReadingEntry.spinLock!)
                    currentlyReadingEntry.framesQueued += Int32(framesAdded)
                    os_unfair_lock_unlock(&currentlyReadingEntry.spinLock!)
                    continue
                } else if status != noErr {
                    unexpectedError(.cError)
                    return
                }
            }
            // end while
        }
    }
    
    func handleRecordingOfAudioPackets(_ numberOfPackets: UInt32, audioBuffer: AudioBuffer?) {
        if recordAudioFileId != nil && recordAudioConverterRef != nil {
            let recordConvertInfo = AudioConvertInfo()
            recordConvertInfo.done = false
            recordConvertInfo.numberOfPackets = numberOfPackets
            recordConvertInfo.packetDescriptions = nil
            recordConvertInfo.audioBuffer = audioBuffer!
            var recordConvertInfoPointer = unsafeBitCast(recordConvertInfo, to: UnsafeMutableRawPointer.self)
            var convertedData = AudioBufferList()
            convertedData.mNumberBuffers = 1
            convertedData.mBuffers.mNumberChannels = recordAudioStreamBasicDescription.mChannelsPerFrame
            convertedData.mBuffers.mDataByteSize = recordOutputBufferSize
            convertedData.mBuffers.mData! = UnsafeMutableRawPointer(recordOutputBuffer!)
            var ioOutputDataPackets: UInt32 = 0
            var status: OSStatus = 0
            while true {
                ioOutputDataPackets = recordPacketsPerBuffer
                status = AudioConverterFillComplexBuffer(recordAudioConverterRef!, AudioConvertInfo.callbak(), &recordConvertInfoPointer, &ioOutputDataPackets, &convertedData, &recordPacketDescriptions)
                if status == 100 || status == 0 {
                    if ioOutputDataPackets > 0 {
                        let writeError: OSStatus = AudioFileWritePackets(recordAudioFileId!, false, convertedData.mBuffers.mDataByteSize, &recordPacketDescriptions, Int64(recordFilePacketPosition), &ioOutputDataPackets, convertedData.mBuffers.mData!)
                        if writeError == noErr {
                            recordFilePacketPosition += ioOutputDataPackets
                        }
                    }
                } else {
                    print("STKAudioPlayer: Unexpected error during recording audio file conversion")
                }
                if status == 100
                {
                    break;
                }
            }
        }
    }
    
    func pendingQueue() -> [Any]? {
        pthread_mutex_lock(&playerMutex)
        var retval: [Any]
        var mutableArray = [AnyHashable](repeating: 0, count: upcomingQueue.count + bufferingQueue.count)
        for entry in upcomingQueue {
            mutableArray.append(entry as! STKHTTPDataSource)
        }
        for entry in bufferingQueue {
            mutableArray.append(entry as! STKHTTPDataSource)
        }
        retval = mutableArray
        pthread_mutex_unlock(&playerMutex)
        return retval
    }
    
    func pendingQueueCount() -> Int {
        pthread_mutex_lock(&playerMutex)
        let retval: Int = upcomingQueue.count + bufferingQueue.count
        pthread_mutex_unlock(&playerMutex)
        return retval
    }
    
    func mostRecentlyQueuedStillPendingItem() -> NSObject? {
        pthread_mutex_lock(&playerMutex)
        if upcomingQueue.count > 0 {
            let retval: NSObject = (upcomingQueue[0] as! STKHTTPDataSource)
            pthread_mutex_unlock(&playerMutex)
            return retval
        }
        if bufferingQueue.count > 0 {
            let retval: NSObject = (bufferingQueue[0] as! STKHTTPDataSource)
            pthread_mutex_unlock(&playerMutex)
            return retval
        }
        pthread_mutex_unlock(&playerMutex)
        return nil
    }
    
    func appendFrameFilter(withName name: String?, block: @escaping STKFrameFilter) {
        addFrameFilter(withName: name, afterFilterWithName: nil, block: block)
    }
    
    func removeFrameFilter(withName name: String?) {
        pthread_mutex_lock(&playerMutex)
        let newFrameFilters: NSMutableArray = NSMutableArray.init(capacity: frameFilters.count + 1)
        for filterEntry in frameFilters {
            if ((filterEntry as! STKFrameFilterEntry).name != name) {
                newFrameFilters.add(filterEntry)
            }
        }
        let replacement = newFrameFilters
        os_unfair_lock_lock(&pcmBufferSpinLock)
        if newFrameFilters.count > 0 {
            frameFilters = replacement as NSArray
        }
        else {
            frameFilters = []
        }
        os_unfair_lock_unlock(&pcmBufferSpinLock)
        pthread_mutex_unlock(&playerMutex)
    }
    
    func addFrameFilter(withName name: String?, afterFilterWithName: String?, block: @escaping STKFrameFilter) {
        pthread_mutex_lock(&playerMutex)
        let newFrameFilters: NSMutableArray = NSMutableArray.init(capacity: frameFilters.count + 1)
        if afterFilterWithName == nil {
            newFrameFilters.add(STKFrameFilterEntry(filter: block, andName: name))
            newFrameFilters.addObjects(from: frameFilters as! [Any])
        }
        else {
            for filterEntry in frameFilters {
                if afterFilterWithName != nil && ((filterEntry as! STKFrameFilterEntry).name == afterFilterWithName) {
                    newFrameFilters.add(STKFrameFilterEntry(filter: block, andName: name))
                }
                newFrameFilters.add(filterEntry)
            }
        }
        let replacement = newFrameFilters
        os_unfair_lock_lock(&pcmBufferSpinLock)
        frameFilters = replacement
        os_unfair_lock_unlock(&pcmBufferSpinLock)
        pthread_mutex_unlock(&playerMutex)
    }
    
    func add(_ frameFilter: @escaping STKFrameFilter, withName name: String?, afterFilterWithName: String?) {
        pthread_mutex_lock(&playerMutex)
        let newFrameFilters: NSMutableArray = NSMutableArray.init(capacity: frameFilters.count + 1)
        if afterFilterWithName == nil {
            newFrameFilters.addObjects(from: frameFilters as! [Any])
            newFrameFilters.add(STKFrameFilterEntry(filter: frameFilter, andName: name) as! AnyHashable)
        } else {
            for filterEntry in frameFilters {
                newFrameFilters.add(filterEntry as! AnyHashable)
                if afterFilterWithName != nil && ((filterEntry as! STKFrameFilterEntry).name == afterFilterWithName) {
                    newFrameFilters.add(STKFrameFilterEntry(filter: frameFilter, andName: name) as! AnyHashable)
                }
            }
        }
        let replacement = newFrameFilters
        os_unfair_lock_lock(&pcmBufferSpinLock)
        frameFilters = replacement as NSArray
        os_unfair_lock_unlock(&pcmBufferSpinLock)
        pthread_mutex_unlock(&playerMutex)
    }
    
    //
    // Phonix: added and used functions
    //
    
    func isCallbackConnected() -> Bool {
        //
        //  callBack has to be disconnected from the converterUnit
        //
        var status: OSStatus = 0
        //
        // Get number of connections of converterNode, if 2, disconnect input
        //
        var numInteractions: UInt32 = 0
        status = AUGraphCountNodeInteractions(audioGraph!, converterNode, &numInteractions)
        if status != noErr {
            print(String(format: "STK_disconnnect CallBack: AUGraphCountNodeInteractions failed: %i", status))
            return false
        }
        if numInteractions > 1 {
            return true
        }
        return false
    }
    
    func disConnectCallback() {
        //
        //  callBack has to be disconnected from the converterUnit
        //
        var status: OSStatus = 0
        //
        // Get number of connections of converterNode, if 2, disconnect input
        //
        var numInteractions: UInt32 = 0
        status = AUGraphCountNodeInteractions(audioGraph!, converterNode, &numInteractions)
        if status != noErr {
            print(String(format: "STK_disconnnect CallBack: AUGraphCountNodeInteractions failed: %i", status))
            return
        }
        if numInteractions > 1 {
            AUGraphDisconnectNodeInput(audioGraph!, converterNode, 0)
        }
        AUGraphUpdate(audioGraph!, nil)
        //        CAShow(UnsafeMutablePointer<AUGraph>(audioGraph!))
    }
    
    func connectCallback() {
        //
        //  callBack has to be connected to the converterUnit
        //
        var status: OSStatus = 0
        AUGraphNodeInfo(audioGraph!, converterNode, &converterDesc, &converterUnit)
        //
        // Get number of connections of converterNode, if < 2, connect input
        //
        var numInteractions: UInt32 = 0
        status = AUGraphCountNodeInteractions(audioGraph!, converterNode, &numInteractions)
        if status != noErr {
            print(String(format: "STK_disconnnect CallBack: AUGraphCountNodeInteractions failed: %i", status))
            return
        }
        if numInteractions < 2 {
            setOutputCallbackForFirstNodeSFB(converterNode, firstUnit: converterUnit!)
            AUGraphUpdate(audioGraph!, nil)
            return
        }
        return
    }
    
    func setOutputCallbackForFirstNodeSFB(_ firstNode: AUNode, firstUnit: AudioComponentInstance) {
        var callbackStruct = AURenderCallbackStruct()
        callbackStruct.inputProc = OutputRenderCallback
        callbackStruct.inputProcRefCon =  UnsafeMutableRawPointer(Unmanaged<STKAudioPlayer>.passUnretained(self).toOpaque())
        
        AudioUnitSetProperty(converterUnit!, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, 0, &canonicalAudioStreamBasicDescription, UInt32(MemoryLayout.size(ofValue: canonicalAudioStreamBasicDescription)))
        
        AUGraphSetNodeInputCallback(audioGraph!, firstNode, 0, &callbackStruct)
    }
    
    func streamFormat() -> AudioStreamBasicDescription {
        return canonicalAudioStreamBasicDescription
    }
    
}

class AudioConvertInfo {
    struct Constant {
        static let end: OSStatus = 100
    }
    var done = false
    var numberOfPackets: UInt32 = 0
    var audioBuffer: AudioBuffer = AudioBuffer()
    var packetDescriptions: UnsafeMutablePointer<AudioStreamPacketDescription>? = nil
    var streamBytsPerFrame: UInt32 = 0
    
    func audioConverterCallback(inAudioConverter: AudioConverterRef, ioNumberDataPackets: UnsafeMutablePointer<UInt32>, ioData: UnsafeMutablePointer<AudioBufferList>, outDataPacketDescription: UnsafeMutablePointer<UnsafeMutablePointer<AudioStreamPacketDescription>?>?) -> OSStatus {
        if done == true {
            ioNumberDataPackets.pointee = 0
            return Constant.end
        }
        ioData.pointee.mNumberBuffers = 1
        ioData.pointee.mBuffers = audioBuffer
        
        outDataPacketDescription?.pointee = packetDescriptions
        ioNumberDataPackets.pointee = numberOfPackets
        done = true
        return noErr
    }
    static func callbak() -> AudioConverterComplexInputDataProc {
        let closure: AudioConverterComplexInputDataProc = {(inAudioConverter, ioNumberDataPackets, ioData, outDataPacketDescription, inUserData) -> OSStatus in
            guard let info = inUserData else { return noErr }
            let convertInfo = info.assumingMemoryBound(to: AudioConvertInfo.self)
            return convertInfo.pointee.audioConverterCallback(inAudioConverter: inAudioConverter, ioNumberDataPackets: ioNumberDataPackets, ioData: ioData, outDataPacketDescription: outDataPacketDescription)
        }
        return closure
    }
}

func OutputRenderCallback(inRefCon: UnsafeMutableRawPointer,
                          ioActionFlags: UnsafeMutablePointer<AudioUnitRenderActionFlags>,
                          inTimeStamp:  UnsafePointer<AudioTimeStamp>,
                          inBusNumber: UInt32,
                          inNumberFrames: UInt32,
                          ioData: UnsafeMutablePointer<AudioBufferList>?) -> OSStatus {
    let audioPlayer =  unsafeBitCast(inRefCon, to: STKAudioPlayer.self)
    os_unfair_lock_lock(&audioPlayer.currentEntryReferencesLock)
    let entry  = audioPlayer.currentlyPlayingEntry
    let currentlyReadingEntry = audioPlayer.currentlyReadingEntry
    os_unfair_lock_unlock(&audioPlayer.currentEntryReferencesLock)
    os_unfair_lock_lock(&audioPlayer.pcmBufferSpinLock)
    var waitForBuffer = false
    let audioBuffer = audioPlayer.pcmAudioBuffer
    let frameSizeInBytes = audioPlayer.pcmBufferFrameSizeInBytes
    let used = audioPlayer.pcmBufferUsedFrameCount
    let start = audioPlayer.pcmBufferFrameStartIndex
    let state = audioPlayer.internalState
    let end: UInt32 = (audioPlayer.pcmBufferFrameStartIndex + audioPlayer.pcmBufferUsedFrameCount) % audioPlayer.pcmBufferTotalFrameCount
    let signal: Bool = audioPlayer.waiting && used < audioPlayer.pcmBufferTotalFrameCount / 2
    let frameFilters = audioPlayer.frameFilters
    if entry != nil {
        if state == .waitingForData {
            var framesRequiredToStartPlaying: Int64 = Int64(audioPlayer.framesRequiredToStartPlaying)
            if entry!.lastFrameQueued >= 0 {
                framesRequiredToStartPlaying = min(framesRequiredToStartPlaying, Int64(entry!.lastFrameQueued))
            }
            if (entry != nil && currentlyReadingEntry == entry && entry!.framesQueued < framesRequiredToStartPlaying) {
                waitForBuffer = true
            }
        } else if state == .rebuffering {
            var framesRequiredToStartPlaying: Int64 = Int64(audioPlayer.framesRequiredToPlayAfterRebuffering)
            if entry!.lastFrameQueued >= 0 {
                framesRequiredToStartPlaying = min(framesRequiredToStartPlaying, Int64(entry!.lastFrameQueued - entry!.framesQueued))
            }
            if used < framesRequiredToStartPlaying {
                waitForBuffer = true
            }
        } else if state == .waitingForDataAfterSeek {
            var framesRequiredToStartPlaying = Int64(1024)
            if entry!.lastFrameQueued >= 0 {
                framesRequiredToStartPlaying = min(framesRequiredToStartPlaying, Int64(entry!.lastFrameQueued - entry!.framesQueued))
            }
            if used < framesRequiredToStartPlaying {
                waitForBuffer = true
            }
        }
    }
    os_unfair_lock_unlock(&audioPlayer.pcmBufferSpinLock)
    var totalFramesCopied: UInt32 = 0
    if used > 0 && !waitForBuffer && entry != nil && state != .paused {
        if state == .waitingForData {
            // Starting
        } else if state == .rebuffering {
            // Resuming from buffering
        }
        if end > start {
            let framesToCopy: UInt32 = min(inNumberFrames, used)
            ioData!.pointee.mBuffers.mNumberChannels = 2
            ioData!.pointee.mBuffers.mDataByteSize = frameSizeInBytes * framesToCopy
            memcpy(UnsafeMutableRawPointer(ioData!.pointee.mBuffers.mData!), audioBuffer.mData!.advanced(by: (Int(start * frameSizeInBytes))), Int(ioData!.pointee.mBuffers.mDataByteSize))
            totalFramesCopied = framesToCopy
            os_unfair_lock_lock(&audioPlayer.pcmBufferSpinLock)
            audioPlayer.pcmBufferFrameStartIndex = (audioPlayer.pcmBufferFrameStartIndex + totalFramesCopied) % audioPlayer.pcmBufferTotalFrameCount
            audioPlayer.pcmBufferUsedFrameCount -= totalFramesCopied
            os_unfair_lock_unlock(&audioPlayer.pcmBufferSpinLock)
        }
        else {
            let framesToCopy: UInt32 = min(inNumberFrames, audioPlayer.pcmBufferTotalFrameCount - start)
            ioData!.pointee.mBuffers.mNumberChannels = 2
            ioData!.pointee.mBuffers.mDataByteSize = frameSizeInBytes * framesToCopy
            memcpy(UnsafeMutableRawPointer(ioData!.pointee.mBuffers.mData!), audioBuffer.mData!.advanced(by: (Int(start * frameSizeInBytes))), Int(ioData!.pointee.mBuffers.mDataByteSize))
            var moreFramesToCopy: UInt32 = 0
            let delta: UInt32 = inNumberFrames - framesToCopy
            if delta > 0 {
                moreFramesToCopy = min(delta, end)
                ioData!.pointee.mBuffers.mNumberChannels = 2
                ioData!.pointee.mBuffers.mDataByteSize += frameSizeInBytes * moreFramesToCopy
                memcpy(UnsafeMutableRawPointer(ioData!.pointee.mBuffers.mData!).advanced(by: (Int(framesToCopy * frameSizeInBytes))), audioBuffer.mData, Int(frameSizeInBytes * moreFramesToCopy))
            }
            totalFramesCopied = framesToCopy + moreFramesToCopy
            os_unfair_lock_lock(&audioPlayer.pcmBufferSpinLock)
            audioPlayer.pcmBufferFrameStartIndex = (audioPlayer.pcmBufferFrameStartIndex + totalFramesCopied) % audioPlayer.pcmBufferTotalFrameCount
            audioPlayer.pcmBufferUsedFrameCount -= totalFramesCopied
            os_unfair_lock_unlock(&audioPlayer.pcmBufferSpinLock)
        }
        audioPlayer.setInternalState(.playing)
    }
    if totalFramesCopied < inNumberFrames {
        let delta = inNumberFrames - totalFramesCopied
        memset(UnsafeMutableRawPointer(ioData!.pointee.mBuffers.mData!).advanced(by: Int(totalFramesCopied * frameSizeInBytes)), 0 ,  Int(delta * frameSizeInBytes))
        if (!(entry == nil || state == .waitingForDataAfterSeek || state == .waitingForData || state == .rebuffering)) {
            // Buffering
            audioPlayer.setInternalState(.rebuffering)
        } else if state == .waitingForDataAfterSeek {
            if totalFramesCopied == 0 {
                audioPlayer.waitingForDataAfterSeekFrameCount += Int32(inNumberFrames - totalFramesCopied)
                if audioPlayer.waitingForDataAfterSeekFrameCount > audioPlayer.framesRequiredBeforeWaitingForDataAfterSeekBecomesPlaying {
                    audioPlayer.setInternalState(.playing)
                }
            }
            else
            {
                audioPlayer.waitingForDataAfterSeekFrameCount = 0
            }
        }
        
    }
    if (frameFilters.count != 0) {
        let count: Int = frameFilters.count
        let asbd = audioPlayer.canonicalAudioStreamBasicDescription
        for i in 0..<count {
            let entry = frameFilters[i] as? STKFrameFilterEntry
            
            entry?.filter!(asbd.mChannelsPerFrame, asbd.mBytesPerFrame, inNumberFrames, ioData!.pointee.mBuffers.mData)
        }
    }
    if entry == nil {
        return 0
    }
    os_unfair_lock_lock(&entry!.spinLock!)
    var extraFramesPlayedNotAssigned = Int64(0)
    var framesPlayedForCurrent: Int64 = Int64(totalFramesCopied)
    if entry!.lastFrameQueued >= 0 {
        framesPlayedForCurrent = Int64(min(entry!.lastFrameQueued - entry!.framesPlayed, Int32(framesPlayedForCurrent)))
    }
    entry!.framesPlayed += Int32(framesPlayedForCurrent)
    extraFramesPlayedNotAssigned = Int64(totalFramesCopied - UInt32(framesPlayedForCurrent))
    let lastFramePlayed: Bool = entry!.framesPlayed == entry!.lastFrameQueued
    os_unfair_lock_unlock(&entry!.spinLock!)
    if signal || lastFramePlayed
    {
        pthread_mutex_lock(&audioPlayer.playerMutex)
        os_unfair_lock_lock(&audioPlayer.currentEntryReferencesLock)
        let currentlyPlayingEntry = audioPlayer.currentlyPlayingEntry
        os_unfair_lock_unlock(&audioPlayer.currentEntryReferencesLock)
        if lastFramePlayed && entry == currentlyPlayingEntry
        {
            audioPlayer.audioQueueFinishedPlaying(entry!)
            while extraFramesPlayedNotAssigned > 0 {
                os_unfair_lock_lock(&audioPlayer.currentEntryReferencesLock)
                let newEntry = audioPlayer.currentlyPlayingEntry
                os_unfair_lock_unlock(&audioPlayer.currentEntryReferencesLock)
                if newEntry != nil {
                    var framesPlayedForCurrent: Int64 = extraFramesPlayedNotAssigned
                    os_unfair_lock_lock(&newEntry!.spinLock!)
                    if newEntry!.lastFrameQueued > 0 {
                        framesPlayedForCurrent = Int64(min(newEntry!.lastFrameQueued - newEntry!.framesPlayed, Int32(framesPlayedForCurrent)))
                    }
                    newEntry!.framesPlayed += Int32(framesPlayedForCurrent)
                    if newEntry!.framesPlayed == newEntry!.lastFrameQueued {
                        os_unfair_lock_unlock(&newEntry!.spinLock!)
                        audioPlayer.audioQueueFinishedPlaying(newEntry!)
                    } else {
                        os_unfair_lock_unlock(&newEntry!.spinLock!)
                    }
                    extraFramesPlayedNotAssigned -= framesPlayedForCurrent
                }
                else
                {
                    break
                }
            }
        }
        pthread_cond_signal(&audioPlayer.playerThreadReadyCondition)
        pthread_mutex_unlock(&audioPlayer.playerMutex)
    }
    return 0
}

final class Atomic<A> {
    private let queue = DispatchQueue(label: "Atomic serial queue")
    private var _value: A
    init(_ value: A) {
        self._value = value
    }
    
    var value: A {
        get {
            return queue.sync { self._value }
        }
    }
    
    func mutate(_ transform: (inout A) -> ()) {
        queue.sync {
            transform(&self._value)
        }
    }
}

extension NSMutableArray {
    
    func enqueue(_ obj: Any) {
        insert(obj, at: 0)
    }
    
    func skipQueue(_ obj: Any?) {
        if let obj = obj {
            add(obj)
        }
    }
    
    func skipQueue(withQueue queue: NSMutableArray?) {
        let queue = queue
        for item in queue ?? [] {
            add(item)
        }
    }
    
    func dequeue() -> Any {
        if count == 0 {
            return ()
        }
        let retval = lastObject
        removeLastObject()
        return retval!
    }
    
    func peek() -> Any? {
        return lastObject
    }
}
