//
//  STKHTTPDataSource.swift
//  STKAudioPlayer
//
//  Created by Viktor on 14/05/2019.
//  Copyright © 2019 Thong Nguyen. All rights reserved.
//
import Foundation
import AudioToolbox

protocol STKHTTPDataSourceDelegate: NSObjectProtocol {
    func dataSourceDataAvailable(_ dataSource: STKHTTPDataSource)
    func dataSourceErrorOccured(_ dataSource: STKHTTPDataSource)
    func dataSourceEof(_ dataSource: STKHTTPDataSource)
    func dataSource(_ dataSource: STKHTTPDataSource, didUpdateMetaData metaDataDictionary: NSDictionary)
}

var radiostream: InputStream?
var eventsRunLoop: RunLoop?

let STK_BIT_RATE_ESTIMATION_MIN_PACKETS_MIN = 2
let STK_BIT_RATE_ESTIMATION_MIN_PACKETS_PREFERRED = 64

func callback(stream: CFReadStream?, eventType: CFStreamEventType, inClientInfo: UnsafeMutableRawPointer?) {
    let client = unsafeBitCast(inClientInfo, to: STKHTTPDataSource.self)
    switch eventType {
    case .errorOccurred:
        client.errorOccured()
        break
    case .endEncountered:
        client.eof()
        break
    case .hasBytesAvailable:
        client.dataAvailable()
        break
    case .openCompleted:
        client.openCompleted()
        break
    default:
        break
    }
}

class STKHTTPDataSource: NSObject, STKHTTPDataSourceDelegate {
    
    func dataSourceDataAvailable(_ dataSource: STKHTTPDataSource) {
        delegate.dataSourceDataAvailable(self)
    }
    
    func dataSourceErrorOccured(_ dataSource: STKHTTPDataSource) {
        delegate.dataSourceErrorOccured(self)
    }
    
    func dataSourceEof(_ dataSource: STKHTTPDataSource) {
        delegate.dataSourceEof(self)
    }
    
    func dataSource(_ dataSource: STKHTTPDataSource, didUpdateMetaData metaDataDictionary: NSDictionary) {
        delegate.dataSource(self, didUpdateMetaData: metaDataDictionary)
    }
    
    var spinLock: os_unfair_lock?
    var parsedHeader : Bool = false
    var sampleRate: Float64 = 0
    var packetDuration: Double = 0.0
    var audioDataOffset: UInt64 = 0
    var audioDataByteCount: UInt64 = 0
    var packetBufferSize: UInt32 = 0
    var seekTime: Float64 = 0
    var framesQueued: Int32 = 0
    var framesPlayed: Int32 = 0
    var lastFrameQueued: Int32 = 0
    var processedPacketsCount: Int = 0
    var processedPacketsSizeTotal: Int = 0
    
    var audioStreamBasicDescription: AudioStreamBasicDescription = AudioStreamBasicDescription()
    
    var durationHint: Double = 0.0
    
    var delegate: STKHTTPDataSourceDelegate!
    
    var recordToFileUrl: URL?
    
    var isInErrorState: Bool = false
    
    var supportsSeek: Bool = false
    var seekStart: Int = 0
    var relativePosition: Int = 0
    var fileLength: Int = 0
    var discontinuous: Bool = false
    var requestSerialNumber: Int32 = 0
    var prefixBytesRead: Int32 = 0
    var prefixBytes: Data?
    var iceHeaderData: NSMutableData?
    var iceHeaderSearchComplete: Bool = false
    var iceHeaderAvailable: Bool = false
    var httpHeaderNotAvailable: Bool = false
    var currentUrl: URL?
    var httpHeaders: NSDictionary = [:]
    var httpStatusCode: UInt32  = 0
    // Meta data
    var metaDataPresent: Bool = false
    var metaDataInterval: UInt = 0 // how many data bytes between meta data
    var metaDataBytesRemaining: UInt = 0 // how many bytes of metadata remain to be read
    var dataBytesRead: UInt = 0 // how many bytes of data have been read
    var foundIcyStart: Bool = false
    var foundIcyEnd: Bool = false
    var metaDataString: NSMutableString = "" //  meta data string
    var audioFileTypeHint: AudioFileTypeID? = kAudioFileMP3Type
    
    var stationIsUp: Bool = false
    
    override init() {
        super.init()
        
    }
    
    convenience init(urlIn: URL) {
        self.init()
        seekStart = 0
        relativePosition = 0
        fileLength = -1
        currentUrl = urlIn
        metaDataString = ""
        
        spinLock = os_unfair_lock.init()
        
        lastFrameQueued = -1

    }
    
    static let audioFileTypeHintFileTypesByMimeType: NSDictionary = {
        var fileTypesByMimeType: NSDictionary = [
            "audio/mp3": NSNumber(value: kAudioFileMP3Type),
            "audio/mpg": NSNumber(value: kAudioFileMP3Type),
            "audio/mpeg": NSNumber(value: kAudioFileMP3Type),
            "audio/wav": NSNumber(value: kAudioFileWAVEType),
            "audio/x-wav": NSNumber(value: kAudioFileWAVEType),
            "audio/vnd.wav": NSNumber(value: kAudioFileWAVEType),
            "audio/aifc": NSNumber(value: kAudioFileAIFCType),
            "audio/aiff": NSNumber(value: kAudioFileAIFFType),
            "audio/x-m4a": NSNumber(value: kAudioFileM4AType),
            "audio/x-mp4": NSNumber(value: kAudioFileMPEG4Type),
            "audio/aacp": NSNumber(value: kAudioFileAAC_ADTSType),
            "audio/m4a": NSNumber(value: kAudioFileM4AType),
            "audio/mp4": NSNumber(value: kAudioFileMPEG4Type),
            "video/mp4": NSNumber(value: kAudioFileMPEG4Type),
            "audio/caf": NSNumber(value: kAudioFileCAFType),
            "audio/x-caf": NSNumber(value: kAudioFileCAFType),
            "audio/aac": NSNumber(value: kAudioFileAAC_ADTSType),
            "audio/aacp": NSNumber(value: kAudioFileAAC_ADTSType),
            "audio/ac3": NSNumber(value: kAudioFileAC3Type),
            "audio/3gp": NSNumber(value: kAudioFile3GPType),
            "video/3gp": NSNumber(value: kAudioFile3GPType),
            "audio/3gpp": NSNumber(value: kAudioFile3GPType),
            "video/3gpp": NSNumber(value: kAudioFile3GPType),
            "audio/3gp2": NSNumber(value: kAudioFile3GP2Type),
            "video/3gp2": NSNumber(value: kAudioFile3GP2Type)
        ]
        return fileTypesByMimeType
    }()
    
    class func audioFileTypeHint(fromMimeType mimeType: String?) -> AudioFileTypeID {
        let number = audioFileTypeHintFileTypesByMimeType[mimeType!] as? NSNumber
        if number == nil {
            return kAudioFileMP3Type
        }
        return AudioFileTypeID(number?.intValue ?? Int(kAudioFileMP3Type))
    }
    
    func parseIceHeader(_ headerData: Data?) -> NSMutableDictionary? {
        let retval: NSMutableDictionary = [:]
        let characterSet = CharacterSet(charactersIn: "\r\n")
        var fullString: String? = nil
        if let headerData = headerData {
            fullString = String(data: headerData, encoding: .utf8)
        }
        let strings = fullString?.components(separatedBy: characterSet)
        httpHeaders = [:]
        for s in strings! {
            if s.count == 0 {
                continue
            }
            if s.hasPrefix("ICY ") {
                let parts = s.components(separatedBy: " ")
                if parts.count >= 2 {
                    httpStatusCode = UInt32(parts[1])!
                }
                continue
            }
            let range: NSRange = (s as NSString).range(of: ":")
            if range.location == NSNotFound {
                continue
            }
            let key = (s as NSString).substring(with: NSRange.init(location: 0, length: range.location))
            let value = (s as NSString).substring(from: range.location + 1)
            retval.setValue(key, forKey: value)
        }
        return retval
    }
    
    func parseHttpHeader() -> Bool {
        if !httpHeaderNotAvailable {
            let response = CFReadStreamCopyProperty(radiostream,CFStreamPropertyKey(rawValue: kCFStreamPropertyHTTPResponseHeader))
            if response != nil {
                httpHeaders = CFHTTPMessageCopyAllHeaderFields(response as! CFHTTPMessage)!.takeUnretainedValue()
                if httpHeaders.count == 0 {
                    httpHeaderNotAvailable = true
                } else {
                    httpStatusCode = UInt32(CFHTTPMessageGetResponseStatusCode(response as! CFHTTPMessage))
                }
            }
        }
        
        if httpHeaderNotAvailable {
            if iceHeaderSearchComplete && !iceHeaderAvailable {
                return true
            }
            if !iceHeaderSearchComplete {
                var byte: UInt8 = 0
                var terminal1: [Character] = ["\n", "\n"]
                var terminal2: [Character] = ["\r", "\n", "\r", "\n"]
                if iceHeaderData == nil {
                    iceHeaderData = NSMutableData.init(capacity: 1024)
                }
                while true {
                    if !hasBytesAvailable {
                        break
                    }
                    let read: Int32 = read2(intoBuffer: &byte, withSize: 1)
                    if read <= 0 {
                        break
                    }
                    iceHeaderData!.append(&byte, length: Int(read))
                    if iceHeaderData!.length >= MemoryLayout.size(ofValue: terminal1) {
                        if memcmp(&terminal1[0], iceHeaderData!.bytes.assumingMemoryBound(to: Int8.self) + iceHeaderData!.length - MemoryLayout.size(ofValue: terminal1), MemoryLayout.size(ofValue: terminal1)) == 0 {
                            iceHeaderAvailable = true
                            iceHeaderSearchComplete = true
                            break
                        }
                    }
                    if iceHeaderData!.length >=  MemoryLayout.size(ofValue: terminal2) {
                        if memcmp(&terminal2[0], iceHeaderData!.bytes.assumingMemoryBound(to: Int8.self) + iceHeaderData!.length - MemoryLayout.size(ofValue: terminal2), MemoryLayout.size(ofValue: terminal2)) == 0 {
                            iceHeaderAvailable = true
                            iceHeaderSearchComplete = true
                            break
                        }
                    }
                    if iceHeaderData!.count >= 4 {
                        if memcmp(iceHeaderData!.bytes, "ICY ", 4) != 0 && memcmp(iceHeaderData!.bytes, "HTTP", 4) != 0 {
                            iceHeaderAvailable = false
                            iceHeaderSearchComplete = true
                            prefixBytes = iceHeaderData as Data?
                            return true
                        }
                    }
                }
                if !iceHeaderSearchComplete {
                    return false
                }
            }
            httpHeaders = parseIceHeader(iceHeaderData as Data?)!
            iceHeaderData = nil
        }
        if (httpHeaders["Accept-Ranges"] ?? httpHeaders["accept-ranges"]) != nil {
            supportsSeek = true
        }
        if httpStatusCode == 200 {
            if seekStart == 0 {
                if let value = httpHeaders["Content-Length"] ?? httpHeaders["content-length"]
                {
                    fileLength = Int(value as! String)!
                }
            }
            let contentType = httpHeaders.object(forKey: "Content-Type") ?? httpHeaders.object(forKey: "content-type")
            let typeIdFromMimeType: AudioFileTypeID = STKHTTPDataSource.audioFileTypeHint(fromMimeType: contentType as? String)
            if typeIdFromMimeType != 0 {
                audioFileTypeHint = typeIdFromMimeType
            }
        }
        else if httpStatusCode == 206 {
            let contentRange = httpHeaders.object(forKey: "Content-Range") ?? httpHeaders.object(forKey: "content-range")
            let components = (contentRange as! NSString).components(separatedBy: "/")
            
            if components.count == 2 {
                fileLength = Int(components[1])!
            }
        }
        else if httpStatusCode == 416 {
            if length()! >= 0 {
                seekStart = length()!
            }
            eof()
            return false
        }
        else if httpStatusCode >= 300 {
            errorOccured()
            return false
        }
        return true
    }
    
    func dataAvailable() {
        if radiostream == nil {
            return
        }
        if httpStatusCode == 0 {
            if parseHttpHeader() {
                if hasBytesAvailable {
                    self.delegate.dataSourceDataAvailable(self)
                }
                return
            } else {
                return
            }
        } else {
            self.delegate.dataSourceDataAvailable(self)
        }
    }
    
    func position() -> Int {
        return seekStart + relativePosition
    }
    
    func length() -> Int? {
        return fileLength >= 0 ? fileLength : 0
    }
    
    func reconnect() {
        let savedEventsRunLoop = eventsRunLoop
        close()
        eventsRunLoop = savedEventsRunLoop
        seek(toOffset: supportsSeek ? position() : 0)
    }
    
    func seek(toOffset offset: Int) {
        let savedEventsRunLoop = eventsRunLoop
        close()
        eventsRunLoop = savedEventsRunLoop
        assert(RunLoop.current == eventsRunLoop, "Seek called on wrong thread")
        radiostream = nil
        relativePosition = 0
        seekStart = offset
        isInErrorState = false
        if !supportsSeek && offset != relativePosition {
            return
        }
        open(forSeek: true)
    }
    
    func read(intoBuffer buffer: UnsafeMutablePointer<UInt8>?, withSize size: Int32) -> Int32 {
        return privateRead(intoBuffer: buffer, withSize: size)
    }
    
    func privateRead(intoBuffer buffer: UnsafeMutablePointer<UInt8>?, withSize size: Int32) -> Int32 {
        if size == 0 {
            return 0
        }
        if prefixBytes != nil {
            let count = min(size, Int32(Int(prefixBytes!.count) - Int(prefixBytesRead)))
            (prefixBytes! as NSData).getBytes(buffer!, length: Int(count))
            prefixBytesRead += count
            if prefixBytesRead >= prefixBytes!.count {
                prefixBytes = nil
            }
            return count
        }
        var read = read2(intoBuffer: buffer, withSize: size)
        if read < 0 {
            return read
        }
        relativePosition = relativePosition + Int(read)
        
        read = checkForMetaDataInfo(withBuffer: buffer, andLength: read)
        return read
    }
    
    func open() {
        return open(forSeek: false)
    }
    
    func read2(intoBuffer buffer: UnsafeMutablePointer<UInt8>?, withSize size: Int32) -> Int32 {
        return Int32(CFReadStreamRead(radiostream, buffer, CFIndex(size)))
    }
    
    func open(forSeek: Bool) {
        var localRequestSerialNumber: Int32 = 0
        requestSerialNumber += 1
        localRequestSerialNumber = requestSerialNumber
        if localRequestSerialNumber != self.requestSerialNumber {
            return
        }
        if self.currentUrl == nil {
            return
        }
        
        createRemoteStream(for: self.currentUrl!)
        _ = reregisterForEvents()
        httpStatusCode = 0
        // Open
        if !CFReadStreamOpen(radiostream) {
            radiostream = nil
            errorOccured()
            return
        }
        isInErrorState = false
   
    }
    
    func eof() {
        self.delegate.dataSourceEof(self)
    }
    
    func errorOccured() {
        isInErrorState = true
        self.delegate.dataSourceErrorOccured(self)
    }
    
    func createRemoteStream(for url: URL) {
        //        let predefinedHttpHeaderValues: [String : String] = [:]
        let request = CFHTTPMessageCreateRequest(nil, "GET" as CFString, url as CFURL, kCFHTTPVersion1_1).takeRetainedValue()
        CFHTTPMessageSetHeaderFieldValue(request, "Icy-MetaData" as CFString, "1" as CFString)
        
        //        if Position.init().start > 0 {
        //            let range = "bytes=\(Position.init().start)-" as CFString
        //            CFHTTPMessageSetHeaderFieldValue(request, Keys.range.cf, range)
        //        }
        
        //        for (key, value) in predefinedHttpHeaderValues {
        //            print("Setting predefined HTTP header[\(key) : \(value)]")
        //            CFHTTPMessageSetHeaderFieldValue(request, key as CFString, value as CFString)
        //        }
        let s = CFReadStreamCreateForHTTPRequest(nil, request)
        let sstream = s.takeUnretainedValue()
        CFReadStreamSetProperty(sstream, CFStreamPropertyKey(rawValue: kCFStreamNetworkServiceType), kCFStreamNetworkServiceTypeBackground)
        CFReadStreamSetProperty(sstream, CFStreamPropertyKey(rawValue: kCFStreamPropertyHTTPShouldAutoredirect), kCFBooleanTrue)
        
        // Proxy Support
        //        if case let ProxyPolicy.custom(info) = ProxyPolicy.system {
        //            var dict: [String: Any] = [:]
        //            dict[kCFNetworkProxiesHTTPPort as String] = info.port
        //            dict[kCFNetworkProxiesHTTPProxy as String] = info.host
        //            let proxy = dict as CFDictionary
        //            if CFReadStreamSetProperty(sstream, CFStreamPropertyKey(rawValue: kCFStreamPropertyHTTPProxy), proxy) == false {
        //                //                print("setting custom proxy not success")
        //            }
        //        }
        //        else {
        //            if let proxy = CFNetworkCopySystemProxySettings()?.takeUnretainedValue() {
        //                //                let dict = proxy as NSDictionary
        //                //                print("system proxy:\(dict)")
        //                CFReadStreamSetProperty(sstream, CFStreamPropertyKey(rawValue: kCFStreamPropertyHTTPProxy), proxy)
        //            }
        //        }
        
        // SSL Support
        if url.scheme?.lowercased() == "https" {
            let sslSettings: [String: Any] = [
                kCFStreamSocketSecurityLevelNegotiatedSSL as String: false,
                kCFStreamSSLLevel as String: kCFStreamSSLValidatesCertificateChain,
                kCFStreamSSLPeerName as String: NSNull(),
            ]
            let key = CFStreamPropertyKey(rawValue: kCFStreamPropertySSLSettings)
            CFReadStreamSetProperty(sstream, key, sslSettings as CFTypeRef)
        }
        radiostream = sstream
    }
    
    // This code was mostly taken from the link below
    // https://code.google.com/p/audiostreamer-meta/
    // Returns new length: the number of bytes from buffer that contain audio data.
    // Other bytes are meta data bytes and this method "consumes" them.
    func checkForMetaDataInfo(withBuffer buffer: UnsafeMutablePointer<UInt8>?, andLength lengte: Int32) -> Int32 {
        let response = CFReadStreamCopyProperty(radiostream,CFStreamPropertyKey(rawValue: kCFStreamPropertyHTTPResponseHeader))
        if foundIcyStart == false && metaDataPresent == false {
            // check if this is a ICY 200 OK response
            let icyCheck: NSString? = NSString.init(bytes: buffer!, length: 10, encoding: String.Encoding.utf8.rawValue)
            if icyCheck != nil && icyCheck!.caseInsensitiveCompare("ICY 200 OK") == .orderedSame {
                foundIcyStart = true
            }
            else {
                if let metaIntRef = CFHTTPMessageCopyHeaderFieldValue(response as! CFHTTPMessage, "Icy-Metaint" as CFString){
                    let metaInt: NSString = (metaIntRef.takeUnretainedValue() as NSString?)!
                    if metaInt != "" {
                        metaDataPresent = true
                        metaDataInterval = UInt(metaInt.intValue)
                    }
                }
            }
        }
        var streamStart: Int = 0
        if foundIcyStart == true && foundIcyEnd == false {
            var c = [Int8](repeating:0, count:4)
            var lineStart = 0
            while streamStart + 3 < lengte {
                memcpy(&c, buffer! + streamStart, 4)
                if c[0] == stringToByteArray(string: "\r")[0] && c[1] == stringToByteArray(string: "\n")[0] {
                    let fullString: NSString = NSString.init(bytes: buffer!, length: streamStart, encoding: String.Encoding.utf8.rawValue)!
                    let line: NSString = fullString.substring(with: NSRange.init(location: lineStart, length: streamStart - lineStart)) as NSString
                    let lineItems: NSArray = line.components(separatedBy: ":") as NSArray
                    if (lineItems.count ) > 1 {
                        if (lineItems[0] as AnyObject).caseInsensitiveCompare("icy-metaint") == .orderedSame {
                            metaDataInterval = UInt((lineItems[0] as! NSString).intValue)
                        } else if (lineItems[0] as AnyObject).caseInsensitiveCompare("content-type") == .orderedSame {
                            let idFromMime: AudioFileTypeID = STKHTTPDataSource.audioFileTypeHint(fromMimeType: lineItems[1] as? String)
                            if idFromMime != 0 {
                                audioFileTypeHint = idFromMime
                            }
                        }
                    }
                    // this is the end of a line, the new line starts in 2
                    lineStart = streamStart + 2
                    if c[2] == stringToByteArray(string: "\r")[0] && c[3] == stringToByteArray(string: "\n")[0] {
                        foundIcyEnd = true
                        metaDataPresent = true
                        streamStart += 4 // skip double new line
                        break
                    }
                }
                streamStart += 1
            }
        }
        if metaDataPresent == true {
            var audioDataByteCount: Int32 = 0
            for i in streamStart..<Int(lengte) {
                // is this a metadata byte?
                if metaDataBytesRemaining > 0 {
                    metaDataString.appendFormat(NSString.init(format: "%c",buffer![i]))
                    metaDataBytesRemaining -= 1
                    if metaDataBytesRemaining == 0 {
                        dataBytesRead = 0
                        let metaDataDictionary = dictionary(fromMetaData: metaDataString as NSString)
                        delegate!.dataSource(self, didUpdateMetaData: metaDataDictionary!)
                    }
                    continue
                }
                // is this the interval byte?
                if metaDataInterval > 0 && dataBytesRead == metaDataInterval {
                    metaDataBytesRemaining = UInt(buffer![i] * 16)
                    metaDataString.setString("")
                    if metaDataBytesRemaining == 0 {
                        dataBytesRead = 0
                    }                    
                    continue
                }
                // this is a data byte
                dataBytesRead += 1
                // overwrite beginning of the buffer with the real audio data
                // we don't need those bytes any more, since we already examined them
                buffer![Int(audioDataByteCount)] = buffer![i]
                audioDataByteCount += 1
            }
            return audioDataByteCount
        } else if foundIcyStart == true {
            // still parsing icy response
            return 0
        } else {
            // no meta data in stream
            return lengte
        }
    }
    
    func unregisterForEvents() {
        if radiostream != nil {
            var ctx = CFStreamClientContext(version: 0, info: unsafeBitCast(self, to: UnsafeMutableRawPointer.self), retain: nil, release: nil, copyDescription: nil)
            let flags = CFStreamEventType.hasBytesAvailable.rawValue | CFStreamEventType.endEncountered.rawValue | CFStreamEventType.errorOccurred.rawValue
            let result = CFReadStreamSetClient(radiostream, flags, callback, &ctx)
            if result == false { return  }
            CFReadStreamUnscheduleFromRunLoop(radiostream, eventsRunLoop!.getCFRunLoop(), CFRunLoopMode.commonModes)
        }
    }
    
    func reregisterForEvents() -> Bool {
        if eventsRunLoop != nil && radiostream != nil {
            var ctx = CFStreamClientContext(version: 0, info:  unsafeBitCast(self, to: UnsafeMutableRawPointer.self), retain: nil, release: nil, copyDescription: nil)
            let flags = CFStreamEventType.hasBytesAvailable.rawValue | CFStreamEventType.endEncountered.rawValue | CFStreamEventType.errorOccurred.rawValue
            let result = CFReadStreamSetClient(radiostream, flags, callback, &ctx)
            if result == false { return false }
            CFReadStreamScheduleWithRunLoop(radiostream, eventsRunLoop!.getCFRunLoop(), CFRunLoopMode.commonModes)
            return true
        }
        return false
    }
    
    func register(forEvents runLoop: RunLoop) -> Bool {
        eventsRunLoop = runLoop
        if radiostream == nil {
            // Will register when they open or seek
            return true
        }
        var ctx = CFStreamClientContext(version: 0, info:  unsafeBitCast(self, to: UnsafeMutableRawPointer.self), retain: nil, release: nil, copyDescription: nil)
        let flags = CFStreamEventType.hasBytesAvailable.rawValue | CFStreamEventType.endEncountered.rawValue | CFStreamEventType.errorOccurred.rawValue
        let result = CFReadStreamSetClient(radiostream, flags, callback, &ctx)
        if result == false { return false }
        CFReadStreamScheduleWithRunLoop(radiostream, eventsRunLoop!.getCFRunLoop(), CFRunLoopMode.commonModes)
        return true
    }
    
    var hasBytesAvailable: Bool {
        if radiostream == nil {
            return false
        }
        return CFReadStreamHasBytesAvailable(radiostream)
    }
    
    func status() -> CFStreamStatus {
        if radiostream != nil {
            return CFReadStreamGetStatus(radiostream)
        }
        return CFStreamStatus(rawValue: 0)!
    }
    
    func openCompleted() {
    }
    
    func close() {
        if radiostream != nil {
            if eventsRunLoop != nil {
                unregisterForEvents()
            }
            CFReadStreamClose(radiostream)
            radiostream = nil
        }
    }
    func dictionary(fromMetaData metaData: NSString) -> NSDictionary? {
        let components: NSArray = metaData.components(separatedBy: ";") as NSArray
        
        let dictionary: NSMutableDictionary = [:]
        
        for entry in components {
            let equalitySignPosition: NSInteger = (entry as! NSString).range(of: "=").location
            if equalitySignPosition != NSNotFound {
                let key = (entry as? NSString)?.substring(to: equalitySignPosition)
                let value = (entry as? NSString)?.substring(from: equalitySignPosition + 1)
                let valueWithoutQuotes = (value as NSString?)?.substring(with: NSRange(location: 1, length: (value?.count ?? 0) - 2))
                dictionary[key ?? ""] = valueWithoutQuotes ?? ""
            }
        }
        return dictionary
    }
    
    func stringToByteArray(string : String)-> Array<Int8>
    {
        let array: [UInt8] = Array(string.utf8)
        var arraySigned = [Int8]()
        var convertSigned: Int8!
        for element in array
        {
            convertSigned = Int8(bitPattern: element)
            arraySigned.append(convertSigned)
        }
        return arraySigned
    }
 
    func reset() {
        os_unfair_lock_lock(&spinLock!)
        framesQueued = 0
        framesPlayed = 0
        lastFrameQueued = -1
        os_unfair_lock_unlock(&spinLock!)
    }
    
    func calculatedBitRate() -> Double {
        var retval: Double
        if packetDuration > 0 {
            if processedPacketsCount > STK_BIT_RATE_ESTIMATION_MIN_PACKETS_PREFERRED || (audioStreamBasicDescription.mBytesPerFrame == 0 && processedPacketsCount > STK_BIT_RATE_ESTIMATION_MIN_PACKETS_MIN) {
                let averagePacketByteSize = Double(processedPacketsSizeTotal) / Double(processedPacketsCount)
                retval = averagePacketByteSize / packetDuration * 8
                return retval
            }
        }
        retval = Double(Int((Float64(audioStreamBasicDescription.mBytesPerFrame) * audioStreamBasicDescription.mSampleRate)) * 8)
        return retval
    }
    
    func duration() -> Double {
        if durationHint > 0.0 {
            return durationHint
        }
        if sampleRate <= 0 {
            return 0
        }
        let audioDataLengthInBytes = self.audioDataLengthInBytes()
        let calculatedBitRate = self.calculatedBitRate()
        if calculatedBitRate < 1.0 || length()! == Int64(0) {
            return 0
        }
        return Double(audioDataLengthInBytes) / (calculatedBitRate / 8)
    }
    
    func audioDataLengthInBytes() -> UInt64 {
        if audioDataOffset != 0 {
            return audioDataByteCount
        } else {
            if length() == nil {
                return 0
            }
            return UInt64(length()!) - audioDataOffset
        }
    }
    
    func isDefinitelyCompatible(_ basicDescription: inout AudioStreamBasicDescription?) -> Bool {
        if audioStreamBasicDescription.mSampleRate == 0 {
            return false
        }
        return memcmp(&audioStreamBasicDescription, &basicDescription, MemoryLayout.size(ofValue: basicDescription)) == 0
    }
    
    func progressInFrames() -> Float64 {
        os_unfair_lock_lock(&spinLock!)
        let retval: Float64 = (seekTime + audioStreamBasicDescription.mSampleRate) + Float64(framesPlayed)
        os_unfair_lock_unlock(&spinLock!)
        return retval
    }
    
}

