import AppKit

class FileBrowserController: NSObject {
    @IBOutlet var controller  : SideViewController!
    @IBOutlet var outlineView : NSOutlineView!
    @IBOutlet var dataSource  : FileBrowserDataSource!
    
    @IBOutlet var pathControl     : NSPathControl!
    @IBOutlet var rootBarView     : NSView!
    @IBOutlet var selectorView    : NSView!
    @IBOutlet var rootSelector    : NSSegmentedControl!
    @IBOutlet var rootCountSetter : NSStepper!
    
    func menuAdd(toNewTab sender: Any, origin: URLOrigin) {
        var index: Int
        var urls = [URL]()
        var tabName: String = ""
        let selectedIndexes = outlineView.selectedRowIndexes as NSIndexSet
        index = selectedIndexes.firstIndex
        while index != NSNotFound {
            let fileItem: FileItem = outlineView.item(atRow: index)! as! FileItem
            if tabName.isEmpty && !fileItem.displayName.isEmpty {
                tabName = makeSafeTabName(for: fileItem.displayName)
            }
            urls.append(fileItem.url!)
            index = selectedIndexes.indexGreaterThanIndex(index)
        }
        controller.doAdd(toNewTab: urls, origin: origin, newTabName: tabName)
    }
    
    func menuAdd(toPlaylist sender: Any, origin: URLOrigin) {
        var index: Int
        var urls = [URL]()
        let selectedIndexes = outlineView.selectedRowIndexes as NSIndexSet
        index = selectedIndexes.firstIndex
        while index != NSNotFound {
            urls.append((outlineView.item(atRow: index)! as! FileItem).url!)
            index = selectedIndexes.indexGreaterThanIndex(index)
        }
        controller.doAdd(toPlaylist: urls, origin: origin)
    }
    
    @IBAction func doubleClickItem(_ sender: Any) {
        if NSUserDefaultsController.shared.defaults.bool(forKey: "FileBrowserDoubleClick") {
            menuAdd(toPlaylist: sender, origin: URLOrigin.URLOriginInternal)
        }
    }
    
    @IBAction func add(toNewTab sender: Any) {
        menuAdd(toNewTab: sender, origin: URLOrigin.URLOriginInternal)
    }
    
    @IBAction func setAsPlaylist(_ sender: Any) {
        controller.clear(sender)
        add(toPlaylist: sender)
    }
    
    @IBAction func add(toPlaylist sender: Any) {
        menuAdd(toPlaylist: sender, origin: URLOrigin.URLOriginInternal)
    }
    
    @IBAction func showEntry(inFinder sender: Any) {
        var index: Int
        let ws: NSWorkspace? = NSWorkspace.shared
        let selectedIndexes = outlineView.selectedRowIndexes as NSIndexSet
        index = selectedIndexes.firstIndex
        while index != NSNotFound {
            let url: URL? = (outlineView.item(atRow: index) as! FileItem).url
            ws?.selectFile(url?.path, inFileViewerRootedAtPath: (url?.path)!)
            index = selectedIndexes.indexGreaterThanIndex(index)
        }
    }
    
    func initControls() {
        pathControl.target = self
        pathControl.action = #selector(pathControlAction)
        pathControl.backgroundColor = MidGrayColor
        // pathcontrol should only highlight folders
        pathControl.allowedTypes = ["public.folder"]
        
        let numberOfRootOptions: Int = UserDefaults.standard.integer(forKey: "numberOfRootOptions")
        rootCountSetter.intValue = Int32(numberOfRootOptions)
        rootCountSetter.toolTip = rootCountSetterToolTip
        
        let selectedRootIndex: Int = UserDefaults.standard.integer(forKey: "selectedRootIndex")
        setRootSelector(segmentCount: numberOfRootOptions)
        rootSelector.setSelected(true, forSegment: selectedRootIndex)
        dataSource.selectRoot(at: selectedRootIndex)
        rootSelector.toolTip = rootSelectorToolTip
        
        dataSource.initFileWatcher()
    }
    
    @IBAction func pathControlAction(_ sender: NSPathControl) {
        if let url = pathControl.clickedPathItem?.url {
            dataSource.changeRoot(at: rootSelector.indexOfSelectedItem, to: url)
        }
    }
    
    @IBAction func rootCountSetterAction(_ sender: NSStepper) {
        setRootSelector(segmentCount: Int(sender.intValue))
    }
    
    @IBAction func setAsRoot(_ sender: Any) {
        let index : Int  = outlineView.selectedRow
        if index != -1 { // NSNotFound should be -1 but it's not //  we validate in change()
            let selectedItem = outlineView.item(atRow: index) as! FileItem
            dataSource.changeSelectedRoot(to: selectedItem.url)
        }
    }
    
    func setRootSelector(segmentCount: Int) {
        var inSegments = segmentCount
        let segments   = rootSelector.segmentCount
        //
        //  catch out of range values
        //
        if inSegments > 6 {
            inSegments = 6
        } else if inSegments < 1 {
            inSegments = 1
        }
        UserDefaults.standard.set(inSegments, forKey: "numberOfRootOptions")
        
        let itemWidth       : CGFloat = 21.0
        let itemSpacing     : CGFloat = 1.0
        let selectorWidth   : CGFloat = itemWidth * CGFloat(inSegments)
        if inSegments >= 2 {
            //
            // we have 2 or more items, show the selector and set the number of items
            //
            if rootSelector.indexOfSelectedItem >= (inSegments - 1) && (inSegments < segments) {
                //
                //  we are removing an item, ande the removed item was selected > select last item
                //
                rootSelector.setSelected(true, forSegment: inSegments - 1)
                dataSource.selectRoot(at: inSegments - 1)
            }
            rootSelector.isHidden = false
            rootSelector.segmentCount = inSegments
            //
            // set label for last segment
            //
            rootSelector.setLabel("\(inSegments)", forSegment: inSegments - 1)
            for segment in 0 ..< inSegments {
                rootSelector.setWidth(itemWidth - itemSpacing, forSegment: segment)
            }
            pathControl.frame     = CGRect(x: pathControl.frame.origin.x,
                                           y: pathControl.frame.origin.y,
                                           width: rootBarView.frame.size.width - rootCountSetter.frame.size.width - selectorWidth - 1.0,
                                           height: pathControl.frame.size.height)
            selectorView.frame    = CGRect(x: rootBarView.frame.size.width - rootCountSetter.frame.size.width - selectorWidth,
                                           y: pathControl.frame.origin.y,
                                           width: selectorWidth + rootCountSetter.frame.size.width,
                                           height: selectorView.frame.size.height)
            rootSelector.frame    = CGRect(x: 0.0,
                                           y: rootSelector.frame.origin.y,
                                           width: rootSelector.frame.size.width,
                                           height: rootSelector.frame.size.height)
            rootCountSetter.frame = CGRect(x: selectorWidth,
                                           y: rootCountSetter.frame.origin.y,
                                           width: rootCountSetter.frame.size.width,
                                           height: rootCountSetter.frame.size.height)
        } else {
            //
            //  we have only one option to select, so hide selector and adjust the view
            //
            rootSelector.setSelected(true, forSegment: 0)
            dataSource.selectRoot(at: 0)
            rootSelector.isHidden = true
            pathControl.frame     = CGRect(x: pathControl.frame.origin.x,
                                           y: pathControl.frame.origin.y,
                                           width: rootBarView.frame.size.width - rootCountSetter.frame.size.width - 1.0,
                                           height: pathControl.frame.size.height)
            selectorView.frame    = CGRect(x: rootBarView.frame.size.width - rootCountSetter.frame.size.width,
                                           y: selectorView.frame.origin.y,
                                           width: rootCountSetter.frame.size.width,
                                           height: selectorView.frame.size.height)
            rootCountSetter.frame = CGRect(x: 0.0,
                                           y: rootCountSetter.frame.origin.y,
                                           width: rootCountSetter.frame.size.width,
                                           height: rootCountSetter.frame.size.height)
        }
    }
    //
    // not used
    //
    func validateFileItem(_ afileItem: FileItem) -> Bool {
        //
        // not used
        //
        // Only let directories be Set as Root
        var isDir: ObjCBool = ObjCBool(false)
        let exists: Bool = PhonixFileManager.fileExists(atPath: afileItem.url.path, isDirectory: &isDir)
        if exists && isDir.boolValue {
            // Exists. Directory.
            return isDir.boolValue
        }
        return false
    }
    
    func validateMenuItem(_ menuItem: NSMenuItem) -> Bool {
        //
        // not used
        //
        let action: Selector? = menuItem.action
        if outlineView.numberOfSelectedRows == 0 {
            return false
        }
        if action == #selector(setAsRoot) {
            let row: Int = outlineView.selectedRow
            if outlineView.numberOfSelectedRows > 1 {
                return false
            }
            // Only let directories be Set as Root
            var isDir: ObjCBool = ObjCBool(false)
            let exists: Bool = PhonixFileManager.fileExists(atPath: (outlineView.item(atRow: row) as! FileItem).url.absoluteString, isDirectory: &isDir)            
            if exists && isDir.boolValue {
                // Exists. Directory.
                return isDir.boolValue
            }
        }
        return true
    }
}
