import AppKit

class FileBrowserDataSource: NSObject, NSOutlineViewDataSource, NSOutlineViewDelegate, NSPathControlDelegate {
   
    @IBOutlet var pathControl     : NSPathControl!
    @IBOutlet var outlineView     : NSOutlineView!
    
    var expandedNodes: NSMutableSet = []
    fileprivate var rootItem: FileItem!
    private var monitor = SwiftFSWatcher()
    
    func initFileWatcher() {
        monitor.watch { changeEvents in
            for ev in changeEvents {
                if      ev.eventFlag.intValue & (kFSEventStreamEventFlagItemIsFile | kFSEventStreamEventFlagItemModified) == kFSEventStreamEventFlagItemIsFile
                    ||  ev.eventFlag.intValue & kFSEventStreamEventFlagItemIsDir  == kFSEventStreamEventFlagItemIsDir
                {
                    self.selectRoot(at: UserDefaults.standard.integer(forKey: "selectedRootIndex"))
                }
            }
        }
    }
    
    @IBAction func rootSelectorButtonAction(_ sender: Any) {
        selectRoot(at: (sender as AnyObject).indexOfSelectedItem)
    }
    
    func selectRoot(at selection: Int) {
        var selection = selection
        selection = max(0, min(5, selection))
        
        UserDefaults.standard.set(selection, forKey: "selectedRootIndex")
        
        if let url =  UserDefaults.standard.url(forKey: "fileBrowserRoot\(selection)") {
            expandedNodes.removeAllObjects()
            if !setRoot(url: url) { return }
            if let expandedNodesArray = UserDefaults.standard.array(forKey: "fileBrowserViewExpandedNodes\(selection)") {
                if expandedNodesArray.count != 0 {
                    expandedNodes = NSMutableSet(array: expandedNodesArray)
                }
            } else {
                expandedNodes = NSMutableSet()
            }
            refreshExpNodes()
        }
    }
    
    func changeSelectedRoot(to url: URL) {
        //
        //  only called from Filebrowser rightmouse setAsRoot
        //
        if url.absoluteString != "" {
            let rootIndex: Int = UserDefaults.standard.integer(forKey: "selectedRootIndex")
            changeRoot(at: rootIndex, to: url)
        }
    }
    
    func changeRoot(at rootIndex: Int, to url: URL) {
        if setRoot(url: url) {
            UserDefaults.standard.set(url, forKey: "fileBrowserRoot\(rootIndex)")
            //
            // a new rootDir for Index -> clear all expandedNodes
            //
            expandedNodes.removeAllObjects()
            UserDefaults.standard.set(expandedNodes.allObjects, forKey: "fileBrowserViewExpandedNodes\(rootIndex)")
            refreshExpNodes()
        }
    }
    
    func setRoot(url: URL) -> Bool {
        let resourceAvailable = (try? url.checkResourceIsReachable()) ?? false
        var isDir       : ObjCBool = false
        let URLexists   = FileManager.default.fileExists(atPath: url.path, isDirectory: &isDir)
        //
        // only allow Dirs to be set as root
        //
        if !resourceAvailable || !URLexists || !isDir.boolValue { return false }
        pathControl.url = url
        rootItem = FileItem(url: url, parent: nil)
        monitor.watchingPaths = [url.path]
        return true
    }
    
    func refreshExpNodes() {
        self.outlineView.reloadData()
        //
        // we Outlineview.numberOfRows changes after expanding or collapsing, but endpoint of for loop is not updates
        //
        let nodeArray : [String] = Array(expandedNodes) as! [String]
        let ordererdNodes = nodeArray.sorted(by: <)
        for expNodePath in ordererdNodes {
            //
            // de volgorde waarin de expandedNodes worden uitgeklpat maakt uit, dir -> sub ->
            // hoe gaan we dat oplossen? ordered? Yep!
            //
            let numberOfRows = self.outlineView.numberOfRows
            for index in 0 ..< numberOfRows {
                if let pathNode = self.outlineView.item(atRow: index) as? FileItem {
                    var isDir : ObjCBool = false
                    let nodePath = pathNode.url.path
                    if expNodePath == nodePath {
                        if FileManager.default.fileExists(atPath: nodePath, isDirectory: &isDir) {
                            self.outlineView.expandItem(pathNode)
                            break
                        }
                        else {
                            expandedNodes.remove(pathNode.url.path)
                        }
                    }
                }
            }
        }
    }
    //
    // delegate functions
    //
    func outlineViewItemDidExpand(_ notification: Notification) {
        guard let node = notification.userInfo?["NSObject"] as? FileItem else { return }
        expandedNodes.add(node.url.path)
        UserDefaults.standard.set(expandedNodes.allObjects,
                                  forKey: "fileBrowserViewExpandedNodes\(UserDefaults.standard.integer(forKey: "selectedRootIndex"))")
    }
    
    func outlineViewItemDidCollapse(_ notification: Notification) {
        guard let node = notification.userInfo?["NSObject"] as? FileItem else { return }
        expandedNodes.remove(node.url.path)
        UserDefaults.standard.set(expandedNodes.allObjects,
                                  forKey: "fileBrowserViewExpandedNodes\(UserDefaults.standard.integer(forKey: "selectedRootIndex"))")
    }
    
    func outlineView(_ outlineView: NSOutlineView, viewFor tableColumn: NSTableColumn?, item: Any) -> NSView? {
        let view = outlineView.makeView(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "filecolumn"), owner: self) as? NSTableCellView
        if let it = item as? FileItem {
            // to tint or not to  tint
            view?.imageView?.image = NSWorkspace.shared.icon(forFile: it.url.path)
            view?.textField?.font = NSFont.systemFont(ofSize: CGFloat(12.0))
            view?.textField?.textColor = TextColor
            view?.textField?.stringValue = it.displayName
        }
        return view
    }
    
    func outlineView(_ outlineView: NSOutlineView, numberOfChildrenOfItem item: Any?) -> Int {
        if let it = item as? FileItem {
            return it.count
        }
        if rootItem != nil {
            return rootItem.count
        }
        return 0
    }
    
    func outlineView(_ outlineView: NSOutlineView, isItemExpandable item: Any) -> Bool {
        if let it = item as? FileItem {
            if it.count > 0 {
                return true
            }
        }
        return false
    }
    
    func outlineView(_ outlineView: NSOutlineView, child index: Int, ofItem item: Any?) -> Any {
        if let it = item as? FileItem {
            return it.childAtIndex(index)!
        }
        if rootItem?.count != nil {
            return rootItem.childAtIndex(index)!
        }
        return ()
    }
    
    func outlineView(_ outlineView: NSOutlineView, writeItems items: [Any], to pasteboard: NSPasteboard) -> Bool {
        // Get selected paths
        var urls = [AnyHashable]() /* TODO: .reserveCapacity(items.count) */
        var paths = [AnyHashable]() /* TODO: .reserveCapacity(items.count) */
        for p: Any in items {
            urls.append(((p as! FileItem).url)!)
            paths.append(((p as! FileItem).url)!.path)
        }
        //
        // Dragging multiple files using the deprecated NSFilenamesPboardType and dragging API.
        // Please update to the NSDraggingSession API.
        //
        pasteboard.declareTypes([NSPasteboard.PasteboardType(rawValue: PhonixUrlsPboardType)], owner: nil)
        // add it to pboard
        pasteboard.setData(try? NSKeyedArchiver.archivedData(withRootObject: urls, requiringSecureCoding: false),
                           forType: NSPasteboard.PasteboardType(rawValue: PhonixUrlsPboardType)
        )
        pasteboard.addTypes([NSPasteboard.PasteboardType("NSFilenamesPboardType")], owner: self)
        pasteboard.setPropertyList(paths, forType: NSPasteboard.PasteboardType("NSFilenamesPboardType"))
        return true
    }
    //
    // pathControl Delegates
    //
    func pathControl(_ pathControl: NSPathControl, willPopUp menu: NSMenu) {
        //for item in menu.items { print("FDS patchControl: tag:", item.tag, "title:", item.title) }
        var index = 0
        
        for item in pathControl.pathItems {
            if item.url?.lastPathComponent == "/"
            || item.url?.lastPathComponent == "Users" {
                if let menuItem = menu.item(withTitle: item.title) {
                    menu.removeItem(menuItem)
                }
                pathControl.pathItems.remove(at: index)
                index -= 1
            }
            index += 1
        }
        index = menu.items.count
        for item in menu.items {
            index -= 1
            if item.isSeparatorItem || item.title == "Choose…" {
                item.tag = 0
            } else {
                item.tag = index
            }
        }
    }
    
    func pathControl(_ pathControl: NSPathControl, willDisplay openPanel: NSOpenPanel) {
        //print("FDS patchControl: willDisplay openPanel:", openPanel as Any)
    }
    
    func pathControl(_ pathControl: NSPathControl, acceptDrop info: NSDraggingInfo) -> Bool {
        return true
    }
    
    func pathControl(_ pathControl: NSPathControl, validateDrop info: NSDraggingInfo) -> NSDragOperation {
        return []
    }
    
    func pathControl(_ pathControl: NSPathControl, shouldDrag pathItem: NSPathControlItem, with pasteboard: NSPasteboard) -> Bool {
        return false
    }
    
    func pathControl(_ pathControl: NSPathControl, shouldDrag pathComponentCell: NSPathComponentCell, with pasteboard: NSPasteboard) -> Bool {
        return false
    }
}

/* // for FS watcher
 print("EVNETFLAGS:",
 kFSEventStreamEventFlagNone,              // 0 = 0
 kFSEventStreamEventFlagMustScanSubDirs,   // 1 = 1
 kFSEventStreamEventFlagUserDropped,       // 2 = 2
 kFSEventStreamEventFlagKernelDropped,     // 3 = 4
 kFSEventStreamEventFlagEventIdsWrapped,   // 4 = 8
 kFSEventStreamEventFlagHistoryDone,       // 5 = 16
 kFSEventStreamEventFlagRootChanged,       // 6 = 32
 kFSEventStreamEventFlagMount,             // 7 = 64
 kFSEventStreamEventFlagUnmount,           // 8 = 128
 kFSEventStreamEventFlagItemCreated,       // 9 = 256 <
 kFSEventStreamEventFlagItemRemoved,       // 10= 512 <
 kFSEventStreamEventFlagItemInodeMetaMod,  // 11= 1024 - streamrec
 kFSEventStreamEventFlagItemRenamed,       // 12= 2048 <
 kFSEventStreamEventFlagItemModified,      // 13= 4096 <
 kFSEventStreamEventFlagItemFinderInfoMod, // 14= 8192
 kFSEventStreamEventFlagItemChangeOwner,   // 15= 16384 -
 kFSEventStreamEventFlagItemXattrMod,      // 16= 32768
 kFSEventStreamEventFlagItemIsFile,        // 17= 65536 <+
 kFSEventStreamEventFlagItemIsDir,         // 18= 131072 <+
 kFSEventStreamEventFlagItemIsSymlink,     // 19= 262144
 kFSEventStreamEventFlagOwnEvent,          // 20= 524288
 kFSEventStreamEventFlagItemIsHardlink,    // 21= 1048576
 kFSEventStreamEventFlagItemIsLastHardlink,// 22= 2097152
 kFSEventStreamEventFlagItemCloned)        // 23= 4194384 <
 */
