import AppKit

class SideViewController: NSViewController, NSSplitViewDelegate {
    @IBOutlet var sideSplitView         : NSSplitView! // filebrowser/track info
    @IBOutlet var splitview             : NSSplitView! // split main/side
    @IBOutlet var mainView              : NSView!
    @IBOutlet var mainWindow            : MainWindow!
    @IBOutlet var playlistLoader        : PlaylistLoader!
    @IBOutlet var artPanel              : NSPanel!
    @IBOutlet var artPanelController    : ArtPanelController!
    
    @objc dynamic var mediaInfoDisplay  : PlaylistEntry?
    @objc dynamic var mediaDetailsTxt   : NSAttributedString?
    
    let defDividerPosition: CGFloat = 258.0
    
    init() {
        super.init(nibName: "FileBrowser", bundle: nil)
    }
    
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
    }
    
    override func awakeFromNib() {
        ValueTransformer.setValueTransformer(missingAlbumArtTransformer,
                                             forName: NSValueTransformerName("MissingAlbumArtTransformer"))       
        splitview?.arrangesAllSubviews = true
        splitview?.dividerStyle = .thin
        //
        // set sideView visability to the stored value
        //
        showSideView(UserDefaults.standard.bool(forKey: "ShowSideView"))
        showAlbumArt(UserDefaults.standard.bool(forKey: "showAlbumArt"))
    }
    //
    // AlbumArt
    //
    @IBAction func toggleAlbumArt(_ sender: Any) {
        showAlbumArt(albumArtIsHidden())
    }
    
    func showAlbumArt(_ show: Bool = true) {
        if sideViewIsHidden() { return }
        let position : CGFloat
        if show {
            position = sideSplitView.frame.size.height - 59.0 - 256.0
        } else {
            position = sideSplitView.frame.size.height - 59.0
        }
        sideSplitView.setPosition(position, ofDividerAt: 0)
        UserDefaults.standard.set(show, forKey: "showAlbumArt")
    }
    
    func albumArtIsHidden() -> Bool {
        return !UserDefaults.standard.bool(forKey: "showAlbumArt")
    }
    //
    // artPanel
    //
    @IBAction func toggleAlbumArtPanel(_ sender: Any) {
        if !artPanel.isVisible {
            artPanel.orderFrontRegardless()
            UserDefaults.standard.set(true , forKey: "showArtPanel")
        } else {
            artPanel.orderOut(sender)
            UserDefaults.standard.set(false , forKey: "showArtPanel")
        }
    }
    
    @IBAction func hideAlbumArtPanel(_ sender: NSButton) {
        UserDefaults.standard.set(false , forKey: "showArtPanel")
        artPanel.orderOut(sender)
    }
    
    func setMediaInfoDisplay(withEntry entry: PlaylistEntry?)
    {
        if entry == nil { return }
        mediaInfoDisplay = entry        
    }
    //
    // sideView control
    //
    @IBAction func toggleSideView(_ sender: Any) {
        showSideView(sideViewIsHidden())
    }
    
    func showSideView(_ show: Bool = true) {
        if show {
            splitview.subviews = [self.view, self.mainView]
            setDividerPosition(UserDefaults.standard.float(forKey: "SideViewDividerPosition"))
        } else {
            splitview.subviews = [mainView]
            mainView.window?.makeFirstResponder(mainView)
        }
        UserDefaults.standard.set(show, forKey: "ShowSideView")
    }
    
    func sideViewIsHidden() -> Bool {
        return splitview.subviews.count == 1
    }
    
    func setDividerPosition(_ position: Float) {
        //
        // set horizontal splitDevider position
        //
        splitview.setPosition(CGFloat(position), ofDividerAt: 0)
    }
    //
    // filebrowser
    //
    func doAdd(toNewTab urls: [URL], origin: URLOrigin, newTabName name: String = "") {
        mainWindow.addTab(withName: name)
        playlistLoader.willInsertURLs(urls, origin: origin)
        playlistLoader.didInsertURLs(playlistLoader.addURLs(urls, sort: true)!, origin: origin)
    }
    
    func doAdd(toPlaylist urls: [URL], origin: URLOrigin) {
        playlistLoader.willInsertURLs(urls, origin: origin)
        playlistLoader.didInsertURLs(playlistLoader.addURLs(urls, sort: true)!, origin: origin)
    }
    
    func clear(_ sender: Any) {
        playlistLoader.clear(sender)
    }
    //
    // delegate functions
    //
    func splitView(_ splitView: NSSplitView, canCollapseSubview subView: NSView) -> Bool {
        return subView != mainView
    }
    
    func splitView(_ splitView: NSSplitView, shouldCollapseSubview subView: NSView, forDoubleClickOnDividerAt dividerIndex: Int) -> Bool {
        return subView != mainView
    }
    
    func splitView(_ splitView: NSSplitView, resizeSubviewsWithOldSize oldSize: NSSize) {
        splitview.adjustSubviews()
    }
    
    func splitView(_ splitView: NSSplitView, shouldAdjustSizeOfSubview subView: NSView) -> Bool {
        return subView == mainView
    }
    
    func splitView(_ splitView: NSSplitView, constrainMinCoordinate proposedMinimumPosition: CGFloat, ofSubviewAt dividerIndex: Int) -> CGFloat {
        //
        // proposedMinimumPosition is the Min position that is possible with the current layout dimensions : 0.00 in this case
        //
        var constrainedCoordinate: CGFloat = proposedMinimumPosition
        
        if dividerIndex == 0 {
            constrainedCoordinate = defDividerPosition
        }
        return constrainedCoordinate
    }
    
    func splitView(_ splitView: NSSplitView, constrainMaxCoordinate proposedMaximumPosition: CGFloat, ofSubviewAt dividerIndex: Int) -> CGFloat {
        //
        // proposedMinimumPosition is the Max position that is possible with the current layout dimensions : window.width in this case
        //
        var constrainedCoordinate: CGFloat = proposedMaximumPosition
        
        if dividerIndex == 0 {
            constrainedCoordinate = splitview.frame.size.width - defDividerPosition
        }
        return constrainedCoordinate
    }
    
    func splitViewDidResizeSubviews(_ notification: Notification) {
        //
        // called at all window resizing
        //
        if !sideViewIsHidden() {
            //
            // store divider position, +1 for the space between, sideSplitView is 1 pixel smaller than dividerPosition
            //
            if sideSplitView.frame.size.width + 1.0 >= defDividerPosition {
                UserDefaults.standard.set(sideSplitView.frame.size.width + 1.0, forKey: "SideViewDividerPosition")
            }
        }
    }
}
