import Accelerate
import AVFoundation
import AppKit

let semaphore = DispatchSemaphore(value: 0)

var renderDelegate : AURenderCallbackDelegate?

func renderCallback(inRefCon:UnsafeMutableRawPointer,
                    ioActionFlags:UnsafeMutablePointer<AudioUnitRenderActionFlags>,
                    inTimeStamp:UnsafePointer<AudioTimeStamp>,
                    inBusNumber:UInt32,
                    inNumberFrames:UInt32,
                    ioData:UnsafeMutablePointer<AudioBufferList>?) -> OSStatus {
    
    return renderDelegate!.setBuffer(auBuffer: ioData!.pointee.mBuffers)
}

@objc protocol AURenderCallbackDelegate {
    func setBuffer(auBuffer: AudioBuffer) -> OSStatus
}

class PlaybackController: NSObject, StatusBarDelegate, STKAudioPlayerDelegate, AURenderCallbackDelegate, MediaKeyTapDelegate {
    
    // Main outlets
    @IBOutlet var mainWindow                   : MainWindow!
    // Controllers
    @IBOutlet var playlistController           : PlaylistController!
    // Playlist outlets
    @IBOutlet var playlistView                 : PlaylistView!
    // WaveForm outlets
    @IBOutlet var waveShapeButton              : NSButton!
    @IBOutlet var httpWave                     : HTTPWaveFormViewOSX!
    @IBOutlet var fileWave                     : WaveFormViewOSX!
    // LevelMeter
    @IBOutlet var levelMeterView               : NSView!
    @IBOutlet var levelMeterEnableButton       : NSButton!
    @IBOutlet var levelMeterItem               : NSToolbarItem!
    // Spectrum Analyzer
    @IBOutlet var spectrumAnalyzerView         : NSView!
    @IBOutlet var spectrumAnalyzerEnableButton : NSButton!
    @IBOutlet var spectrumAnalyzerItem         : NSToolbarItem!
    // Miscellaneous outlets
    @IBOutlet var playbackButtons              : PlaybackButtons!
    @IBOutlet var recordButton                 : NSButton!
    @IBOutlet var fxButton                     : NSButton!
    @IBOutlet var searchField                  : NSSearchField!
    @IBOutlet var volumeSlider                 : NSSlider!
    @IBOutlet var wavepositionSlider           : WaveSlider!
    // statusbar
    var statusBar : StatusBar?
    // Timers
    var displaylink : DisplayLink!
    var timerCount  : Int = 0
    var debugCount  : Int = -1
    // Booleans
    var levelMeterEnabled       : Bool   = false
    var spectrumAnalyzerEnabled : Bool   = false
    var httpWaveEnabled         : Bool   = false
    var didEnqueueNext          : Bool   = false
    
    var playbackStatus      : Int    = 0
    // must be @objc dynamic because of KeyValueObservation in MainWindow.xib
    @objc dynamic var position      : Double = 0.0
    
    // spectrum analyzer
    var fftSetup            : FFTSetup?
    var bufLog2N            : vDSP_Length = 0
    var spectrumAnalyzer    : SpectrumAnalyzer!
    var levelMeter          : LevelMeter!
    
    var kMeterController    : K_MeterController!
    var auController        : AUController!
    
    // AU render Callback
    var buffer  : AudioBuffer?
    
    func setBuffer(auBuffer: AudioBuffer) -> OSStatus {
        if auBuffer.mData != nil {
            buffer = auBuffer
        }
        return noErr
    }
    
    var mediaKeyTap: MediaKeyTap!
    
    override init() {
        super.init()
        registerSFB()
        startObservers()
        renderDelegate = self
        initMeter(with: registerKMeter())
        if UserDefaults.standard.bool(forKey: "useAUeffects") {
           initAUcontroller()
        }
        StreamPlayer = STKAudioPlayer.init(graph: sfbAUGraph(), sfbConverterNode())
        StreamPlayer.delegate = self
        setStoredOutputDevice()
        AudioUnitAddRenderNotify(mixerUnit(), renderCallback, Unmanaged.passUnretained(self).toOpaque())
        initDisplayLink()
    }
    
    // MediaKeyTap
    func handle(mediaKey: MediaKey, event: KeyEvent) {
        switch mediaKey {
        case .playPause:
            playPauseResume()
        case .previous, .rewind:
            prev()
        case .next, .fastForward:
            next()
        }
    }
    
    func setStoredOutputDevice() {
        
        let storedOutputDevice = UserDefaults.standard.object(forKey: "outputDevice") as? [AnyHashable : Any]
        if storedOutputDevice != nil {
            if !setOutputDeviceTo( (storedOutputDevice!["UID"] as! String).cString(using: .utf8)) {
                //        if !FilePlayer.setOutputDeviceTo(storedOutputDevice) {
                //            print("PbC_restoreOutputDevice: Error:", storedOutputDevice as Any)
                //
                // the device is set to SystemOutput, we should only change our default setting from prefs.
                // we don't want our default device to change because it was not present yet.
                //
            }
        }
        else  {
            //            print("PbC_restoreOutputDevice: storedOutputDevice == nil:", storedOutputDevice as Any)
            //
            // there is no device stored in userdefaults, store our defaultSystem device in defaults
            //
            if let defDevice = systemDefaultDevice() {
                UserDefaults.standard.set(defDevice, forKey: "outputDevice")
            }
        }
    }
    
    func systemDefaultDevice() -> [AnyHashable : Any]? {
        
        var propertySize : UInt32 = 0
        var propertyAddress = AudioObjectPropertyAddress()
        
        var systemDevice : [AnyHashable: Any]
        
        propertyAddress.mScope      = kAudioObjectPropertyScopeOutput
        propertyAddress.mElement    = kAudioObjectPropertyElementMaster
        //
        // get deviceID of systemOutputDevice
        //
        propertyAddress.mSelector   = kAudioHardwarePropertyDefaultOutputDevice
        // get size
        if noErr != AudioObjectGetPropertyDataSize(AudioObjectID(kAudioObjectSystemObject), &propertyAddress, 0, nil, &propertySize)
        { return nil }
        // get data
        var deviceID : AudioDeviceID = 0
        if noErr != AudioObjectGetPropertyData(AudioObjectID(kAudioObjectSystemObject), &propertyAddress, 0, nil, &propertySize, &deviceID)
        { return nil }
        //
        // get deviceName of systemOutputDevice
        //
        propertyAddress.mSelector   = kAudioDevicePropertyDeviceName
        // get size
        if noErr != AudioObjectGetPropertyDataSize(deviceID, &propertyAddress, 0, nil,  &propertySize)
        { return nil }
        // get data
        var deviceName : [Int8] = [Int8](repeating: 0, count: Int(propertySize))
        if noErr != AudioObjectGetPropertyData(deviceID, &propertyAddress, 0, nil, &propertySize, &deviceName)
        { return nil }
        //
        // get deviceUID of systemOutputDevice
        //
        propertyAddress.mSelector   = kAudioDevicePropertyDeviceUID
        // get size
        if noErr != AudioObjectGetPropertyDataSize(deviceID, &propertyAddress, 0, nil,  &propertySize)
        { return nil }
        // get data
        var deviceUID : CFString = "" as CFString
        if noErr != AudioObjectGetPropertyData(deviceID, &propertyAddress, 0, nil, &propertySize, &deviceUID)
        { return nil }
        
        systemDevice = [ "name"      : String(utf8String: deviceName) as Any,
                         "deviceID"  : Int32(deviceID),
                         "UID"       : "\(deviceUID)" ]
        
        //print("PbC_systemDefaultDevice: systemDevice:", systemDevice)
        return systemDevice
    }
    
    func startObservers() {
        NSUserDefaultsController.shared.addObserver(self, forKeyPath: "values.BypassFX", options: [], context: nil)
        NSUserDefaultsController.shared.addObserver(self, forKeyPath: "values.ShowMenuBarButtons", options: [], context: nil)
        NSUserDefaultsController.shared.addObserver(self, forKeyPath: "values.ShowMenuBarLeveldB", options: [], context: nil)
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey: Any]?, context: UnsafeMutableRawPointer?) {
        if keyPath == "values.BypassFX" {
            fxButton.state = UserDefaults.standard.bool(forKey: "BypassFX") ? .off : .on
        } else if keyPath == "values.ShowMenuBarButtons" {
            if UserDefaults.standard.bool(forKey: "ShowMenuBarButtons") {
                statusBar = StatusBar(volumeDisplay: UserDefaults.standard.bool(forKey: "ShowMenuBarLeveldB"))
                statusBar?.delegate = self
                set(volume:nil)
            } else {
                statusBar?.removeAllItems()
                statusBar = nil
            }
        } else if keyPath == "values.ShowMenuBarLeveldB" {
            if UserDefaults.standard.bool(forKey: "ShowMenuBarButtons") {
                if UserDefaults.standard.bool(forKey: "ShowMenuBarLeveldB") {
                    statusBar = StatusBar(volumeDisplay: UserDefaults.standard.bool(forKey: "ShowMenuBarLeveldB"))
                    statusBar?.delegate = self
                    set(volume:nil)
                } else {
                    statusBar?.removeItemVolumeText()
                }
            }
        }
    }
    
    //
    // the Meters
    //
    func initLevelMeter(_ inView: NSView) { levelMeter = inView as? LevelMeter }
    
    func initSpectrumAnalyzer(_ inView: NSView) { spectrumAnalyzer = inView as? SpectrumAnalyzer }
    
    func availableDeviceSampleRates() -> [Float64] {
        var deviceSampleRates : [Float64] = [Float64](repeating: 0.0, count: 32)
        let sampleRateCount = getDeviceSampleRates(&deviceSampleRates)
        var availableSampleRates : [Float64] = [Float64](repeating: 0.0, count: Int(sampleRateCount))
        for i in 0..<sampleRateCount {
            availableSampleRates[Int(i)] = deviceSampleRates[Int(i)]
        }
        return availableSampleRates
    }
    
    func setDeviceSampleRateToEntrySampleRate(_ pe: PlaylistEntry) -> Bool {
        var newSampleRate = Float64(44100.0)
        //
        // a new file is played store it's sampleRate to defaults
        //
        if pe.url!.isFileURL && pe.sampleRate != 0.0 {
            UserDefaults.standard.set(pe.sampleRate, forKey: "mediaSampleRate")
        } else {
            //
            // for streams pe.sampleRate = 0.0, STK plays at 44100.0
            //
            UserDefaults.standard.set(44100.0, forKey: "mediaSampleRate")
        }
        //
        // set the device samplerate to fixed or media samplerate
        // should we make an exception for streams?
        //
        if NSUserDefaultsController.shared.defaults.bool(forKey: "sampleRateFixed") {
            //
            // We set DeviceSampleRate to Fixed SampleRate
            //
            newSampleRate = NSUserDefaultsController.shared.defaults.double(forKey: "fixedSampleRate")
            
        } else {
            //
            // We set DeviceSampleRate to File SampleRate
            //
            if pe.url!.isFileURL {
                newSampleRate = Float64(pe.sampleRate)
                if newSampleRate == 0.0 && pe.URLCue == nil {
                    //
                    // this works for : ogg, ...
                    //
                    let urlstring = pe.url!.path.cString(using: String.Encoding.utf8)
                    newSampleRate = Float64(sampleRateFromMetaDataForURL(urlstring))
                }
            } else {
                //
                // stream, we do nothing, newSamplerate is 44100.00
                //
            }
        }
        if !availableDeviceSampleRates().contains(newSampleRate) {
            newSampleRate = 44100.0
        }
        return setDeviceSampleRate(newSampleRate)
    }
    
    override func awakeFromNib() {
        if !UserDefaults.standard.bool(forKey: "useAUeffects") {
            mainWindow.toolbar!.removeItem(at: 7)
            mainWindow.toolbar?.validateVisibleItems()
        }
        else {
            fxButton.state   = UserDefaults.standard.bool(forKey: "BypassFX") ? .off : .on
            fxButton.toolTip = FXMenuButtonTooltip
            mainWindow.toolbar?.validateVisibleItems()
        }
        if UserDefaults.standard.bool(forKey: "MediaKeys") {
            // Init MediaKeyTap for mediakeys
            mediaKeyTap = MediaKeyTap(delegate: self)
            mediaKeyTap.start()
        }
        
        if UserDefaults.standard.bool(forKey: "ShowMenuBarButtons") {
        statusBar = StatusBar(volumeDisplay: UserDefaults.standard.bool(forKey: "ShowMenuBarLeveldB"))
        statusBar?.delegate = self
        }
        
        playbackStatus = PhonixStatus.phonixStopped.rawValue
        updatePlayerStatus()
        
        // Init Spectrum Analyzer
        initSpectrumAnalyzer(spectrumAnalyzerView)
        spectrumAnalyzerEnable(UserDefaults.standard.bool(forKey: "spectrumAnalyzerEnabled"))
        
        // Init LevelMeter
        initLevelMeter(levelMeterView)
        levelMeterEnable(UserDefaults.standard.bool(forKey: "levelMeterEnabled"))
        
        mainWindow.toolbar?.allowsUserCustomization = false
        mainWindow.toolbar?.isVisible = true
        
        wave(enable: UserDefaults.standard.bool(forKey: "waveFormPositionSlider"))
        
        waveShapeButton.state = UserDefaults.standard.value(forKey: "waveFormPositionSlider") as! NSButton.StateValue
        waveShapeButton.toolTip = waveShapeButtonTooltip
        
        //
        // set volume to userdefaults stored value
        //
        restoreVolume()
    }
    
    // ----------Transport functions
    func autoplayStart() {
        //
        //  autoPlay
        //
        let lastIndex: Int = UserDefaults.standard.integer(forKey: "lastTrackPlaying")
        if UserDefaults.standard.bool(forKey: "autoplay") && ((UserDefaults.standard.integer(forKey: "lastPlaybackStatus") & 3) == 0) && lastIndex >= 0 {
            play(entryAt: lastIndex)
            if UserDefaults.standard.integer(forKey: "lastPlaybackStatus") < 3 {
                SFBseek(UserDefaults.standard.double(forKey: "lastTrackPosition"))
            }
        }
    }
    
    func updatePlayerStatus() {
        var play : Bool = false
        switch playbackStatus {
        case PhonixStatus.phonixPlaying.rawValue    : play = true
        case PhonixStatus.phonixPaused.rawValue     : play = false
        case PhonixStatus.phonixPending.rawValue    : play = false
        case PhonixStatus.phonixStopped.rawValue    : play = false
        case PhonixStatus.phonixHTTPPlaying.rawValue: play = true
        case PhonixStatus.phonixHTTPPaused.rawValue : play = false
        case PhonixStatus.phonixHTTPStopped.rawValue: play = false
        default : play = false
        }
        
        if play {
            startTimers()
        } else {
            stopTimers()
        }
        playbackButtons.set(playStatus: play)
        statusBar?.set(playStatus: play)
    }
    
    @IBAction func playPauseResume(_ sender: Any) { playPauseResume() }
    @IBAction func pause(_ sender: Any) { pause() }
    @IBAction func resume(_ sender: Any) { resume() }
    @IBAction func play(_ sender: Any) { play() }
    @IBAction func stop(_ sender: Any) { stop() }
    @IBAction func next(_ sender: Any) { next() }
    @IBAction func prev(_ sender: Any) { prev() }
    
    @IBAction func seek(_ sender: Any) {
        let pos = (sender as AnyObject).doubleValue!
        SFBseek(pos)
        PBCsetPosition(pos)
        fileWave.updateProgress(SFBprogress())
    }
    
    @IBAction func eventSeekForward(_ sender: Any) {
        if playbackStatus <= PhonixStatus.phonixStopped.rawValue {
            SFBseekForward(1.0)
            fileWave.updateProgress(SFBprogress())
        }
    }
    
    @IBAction func eventSeekBackward(_ sender: Any) {
        if playbackStatus <= PhonixStatus.phonixStopped.rawValue {
            SFBseekBackward(1.0)
            fileWave.updateProgress(SFBprogress())
        }
    }
    
    func playPauseResume() {
        switch playbackStatus
        {
        case PhonixStatus.phonixPlaying.rawValue    : pause()
        case PhonixStatus.phonixPaused.rawValue     : resume()
        case PhonixStatus.phonixStopped.rawValue,
             PhonixStatus.phonixPending.rawValue    : play()
        case PhonixStatus.phonixHTTPPlaying.rawValue: httpPause()
        case PhonixStatus.phonixHTTPPaused.rawValue : httpPlay()
        case PhonixStatus.phonixHTTPStopped.rawValue: httpPlay()
        default: break
        }
    }
    
    func httpPause() {
        StreamPlayer.pause()
        playbackStatus = PhonixStatus.phonixHTTPPaused.rawValue
        updatePlayerStatus()
    }
    
    func httpResume() {
        StreamPlayer.resume()
        playbackStatus = PhonixStatus.phonixHTTPPlaying.rawValue
        updatePlayerStatus()
    }
    
    func httpPlay() {
        StreamPlayer.STKplay(playlistController.currentEntry!.url!)
        playbackStatus = PhonixStatus.phonixHTTPPlaying.rawValue
        updatePlayerStatus()
    }
    
    func httpStop() {
        StreamPlayer.stop()
        playbackStatus = PhonixStatus.phonixHTTPStopped.rawValue
        updatePlayerStatus()
    }
    
    func pause() {
        SFBplayPause()
        playbackStatus = PhonixStatus.phonixPaused.rawValue
        updatePlayerStatus()
    }
    
    func resume() {
        SFBplayPause()
        playbackStatus = PhonixStatus.phonixPlaying.rawValue
        updatePlayerStatus()
    }
    
    func play() {
        if playlistView.selectedRow == -1 {
            // If no entry is selected, select entry 0
            playlistView.selectRowIndexes(NSIndexSet(index: 0) as IndexSet, byExtendingSelection: false)
        }
        if playlistView.selectedRow > -1 {
            play(entryAt: playlistView.selectedRow)
        }
    }
    
    func stop() {
        
        if playbackStatus == PhonixStatus.phonixHTTPPlaying.rawValue {
            StreamPlayer.pause() // stop() causes unreliable start
            playbackStatus = PhonixStatus.phonixHTTPStopped.rawValue
            httpWave.resetHistoryBuffers()
        }
        
        if playbackStatus == PhonixStatus.phonixPlaying.rawValue || playbackStatus == PhonixStatus.phonixPaused.rawValue {
            SFBstop()
            SFBseek(0)
            playbackStatus = PhonixStatus.phonixStopped.rawValue
            fileWave.updateProgress(0)
        }
        PBCsetPosition(0)
        updatePlayerStatus()
 
    }
    
    func next() {
        mainWindow.selectPlayingTab()
        
        let next = playlistController.nextEntry()
        if play(entry: next) {
            updatePlayerStatus()
        }
    }
    
    func prev() {
        mainWindow.selectPlayingTab()
        
        if SFBcurrentTime() > 4.0 {
            SFBseek(0)
            PBCsetPosition(0)
        } else {
            let prev = playlistController.prevEntry()
            if play(entry: prev) {
                updatePlayerStatus()
            }
        }
    }
    
    // called by double-clicking on table
    func play(entryAt index: Int) { play(entryAt: index, startPaused: false) }
    
    func play(entryAt index: Int, startPaused paused: Bool) { play(entry: playlistController.entry(at: index)!, startPaused: paused) }
    
    func play(entry pe: PlaylistEntry?) -> Bool {
        if pe == nil { return false }
        play(entry: pe!, startPaused: false)
        return true
    }
    
    func play(entry pe: PlaylistEntry, startPaused paused: Bool) {
        if pe.url == nil { return }
        //
        //  clear the HTTP info, so the previous data is not shown before the current data is present
        //
        playlistController.PLCsetCurrentEntry(pe)
        
        let selectedTabId = mainWindow.selectedTabId()
        mainWindow.setPlayingTab(withId: selectedTabId)
        
        if acceptableStreamTypes().contains(pe.url!.scheme!) {
            //
            // play a stream
            //             
            let url = pe.url!
            let req = NSURLRequest.init(url: url, cachePolicy: .reloadIgnoringLocalAndRemoteCacheData, timeoutInterval: 1.0)

            let sp = StationPinger()
             
            sp.request = req as URLRequest
   
            sp.startLoading()
                
            semaphore.wait()
            
            if sp.dataTask != nil && ((sp.dataTask?.response as! HTTPURLResponse).statusCode == 200 || (sp.dataTask?.response as! HTTPURLResponse).statusCode == 400) {
//                print("Station = up")
                sp.stopLoading()
                
                stop()
                
                if StreamPlayer.state != .stopped {
                    StreamPlayer.stopThread()
                    StreamPlayer.destroyAudioResources()
                    StreamPlayer.internalState = .initialised
                    StreamPlayer.upcomingQueue = []
                    StreamPlayer.bufferingQueue = []
                    StreamPlayer.resetPcmBuffers()
                    StreamPlayer.createPlaybackThread()
                }
                
                _ = setDeviceSampleRateToEntrySampleRate(pe)
                
                if StreamPlayer.isCallbackConnected() == false {
                    //
                    // we need to settupForFormat first, and then connectCallback()
                    // if we setupForFormat for a second time, we get loud noise
                    //                    
                    _ = setupForFormat((StreamPlayer.streamFormat()))
                    StreamPlayer.connectCallback()
                }
                muteFileConverter(true)
                
                StreamPlayer.STKplay(pe.url!)
                if StreamPlayer.state == .playing { recordButton.isHidden = false }
                playbackStatus = PhonixStatus.phonixHTTPPlaying.rawValue
                
                showWaveView(pe.url!, withLength: pe.totalFrames)
            }
            else {
//                print("Station = down")
                sp.stopLoading()
            }
            
        } else {
            //
            // play a file
            //
            if !pe.isError && PhonixFileManager.fileExists(atPath: (pe.url?.path)!)    {
                let url: URL? = pe.url
                //
                // play a file wav, flac or mp3 etc
                //
                stop()
                if StreamPlayer.state != .stopped {
                    StreamPlayer.stopThread()
                    StreamPlayer.internalState = .initialised
                    StreamPlayer.upcomingQueue = []
                    StreamPlayer.bufferingQueue = []
                    StreamPlayer.resetPcmBuffers()
                    StreamPlayer.destroyAudioResources()
                    StreamPlayer.disConnectCallback()
                }
                muteFileConverter(false)
              
                recordButton.isHidden = true
                _ = setDeviceSampleRateToEntrySampleRate(pe)
                
                if  url?.pathExtension.lowercased() == "cue"
                    && pe.startCue != -1.0 && pe.endCue != -1.0 {
                    if PhonixFileManager.fileExists(atPath: pe.URLCue!.path) {
                        let frames = UInt32(pe.length.floatValue * pe.sampleRate)
                        
                        if playCueURL(pe.URLCue?.path.cString(using: .utf8), Int64(Float(pe.startCue) * pe.sampleRate), frames) {
                            showWaveView(pe.URLCue!, startFrame: Int64(Float(pe.startCue) * pe.sampleRate), withLength: frames)
                            playbackStatus = PhonixStatus.phonixPlaying.rawValue
                        }
                    }
                } else if   url?.pathExtension.lowercased() == "dsf"
                        ||  url?.pathExtension.lowercased() == "dff"
                        ||  url?.pathExtension.lowercased() == "dsd"
                {
                    if playDSDURL(pe.url?.path.cString(using: .utf8)) {
                        showWaveView(pe.url!, withLength: pe.totalFrames)
                        playbackStatus = PhonixStatus.phonixPlaying.rawValue
                    }
                } else if pe.sampleRate != 0 {
                    //
                    // play a file wav, flac or mp3 etc
                    //
                    if      pe.url!.pathExtension.lowercased() == "mod"
                        ||  pe.url!.pathExtension.lowercased() == "s3m"
                        ||  pe.url!.pathExtension.lowercased() == "it"
                        ||  pe.url!.pathExtension.lowercased() == "xm"
                    {
                        //spectrumAnalyzerEnable(false)
                    }
                    if SFBplayURL(pe.url?.path.cString(using: .utf8)) {
                        showWaveView(pe.url!, withLength: pe.totalFrames)
                        playbackStatus = PhonixStatus.phonixPlaying.rawValue
                    }
                }
            }
        }
        didEnqueueNext = false
        //
        // default actions at playback start.
        //
        updatePlayerStatus()
    }
    
    func enqueue(entry pe: PlaylistEntry) -> Bool
    {
        var succes : Bool = false
        if pe.url!.pathExtension.lowercased() == "cue" {
            succes = enqueueCueURL(pe.URLCue?.path.cString(using: .utf8),
                                   Int64(Float(pe.startCue) * pe.sampleRate),
                                   UInt32(pe.length.floatValue * pe.sampleRate))
            if succes {
                showWaveView(pe.URLCue!,
                             startFrame: Int64(Float(pe.startCue) * pe.sampleRate),
                             withLength: UInt32(pe.length.floatValue * pe.sampleRate))
            }
            
        } else if   pe.url!.pathExtension.lowercased() == "dsf"
                ||  pe.url!.pathExtension.lowercased() == "dff"
                ||  pe.url!.pathExtension.lowercased() == "dsd"
        {
            succes = playDSDURL(pe.url?.path.cString(using: .utf8))
        } else {
            succes = SFBenqueue(pe.url?.path.cString(using: .utf8))
            if succes {
                showWaveView(pe.url!, withLength: pe.totalFrames)
            }
        }
        return succes
    }
    
    @IBAction func skip(toNextAlbum sender: Any) {
        var found = false
        if playlistController.currentEntry == nil {
            return
        }
        let index: Int = playlistController.currentEntry!.peindex
        let origAlbum: String = playlistController.currentEntry!.album
        var curAlbum: String
        var pe: PlaylistEntry?
        var j: Int = 0
        for i in 1 ..< (playlistController.arrangedObjects as AnyObject).count {
            pe = playlistController.entry(at: index + i)
            if pe == nil {
                break
            }
            if (pe?.isError)! {
                break
            }
            curAlbum = (pe?.album)!
            // check for untagged files, and just play the first untagged one
            // we come across
            if curAlbum == "" {
                found = true
                j = i
                break
            }
            if curAlbum.caseInsensitiveCompare(origAlbum) != .orderedSame {
                found = true
                j = i
                break
            }
        }
        if found {
            play(entryAt: j + index)
        }
    }
    
    @IBAction func skip(toPreviousAlbum sender: Any) {
        var found = false
        var foundAlbum = false
        let index: Int = playlistController.currentEntry!.peindex
        var origAlbum: String = playlistController.currentEntry!.album
        var curAlbum: String
        var pe: PlaylistEntry?
        var j: Int = 0
        for i in 1 ..< (playlistController.arrangedObjects as AnyObject).count {
            pe = playlistController.entry(at: index - i)
            if pe == nil {
                break
            }
            if (pe?.isError)! {
                break
            }
            curAlbum = (pe?.album)!
            if curAlbum == "" {
                found = true
                j = i
                break
            }
            if curAlbum.caseInsensitiveCompare(origAlbum) != .orderedSame {
                if foundAlbum == false {
                    foundAlbum = true
                    // now we need to move up to the first song in the album, so we'll
                    // go till we either find index 0, or the first song in the album
                    origAlbum = curAlbum
                    j = i
                    continue
                } else {
                    found = true
                    j = i
                    // terminate loop
                    break
                }
            }
            j += 1
        }
        if found || foundAlbum {
            if foundAlbum == true {
                j = j - 1
            }
            play(entryAt: index - j)
        }
    }
    
    func PBCsetPosition(_ p: Double) { position = p }
    
    func prepareForNext() {
        if playbackStatus != PhonixStatus.phonixHTTPPlaying.rawValue
        {
            let currentTime = SFBcurrentTime()
            if currentTime < 0.5 {
                didEnqueueNext = false
            }
            
            if currentTime + TimeInterval(0.5) >= SFBtotalTime() && didEnqueueNext == false {
                mainWindow.selectPlayingTab()
                
                let pe = playlistController.currentEntry!
                let peNext = playlistController.nextEntry()
                
                if !playlistController.currentEntry!.isStopAfter
                    && playlistController.nextEntry() != nil
                {
                    if peNext?.url?.isFileURL == true
                    {
                        if UserDefaults.standard.bool(forKey: "gaplessplayback")
                            &&  pe.sampleRate           == peNext!.sampleRate
                            &&  pe.bitrate              == peNext!.bitrate
                            &&  pe.channels             == peNext!.channels
                            &&  pe.url!.pathExtension   == peNext!.url!.pathExtension
                        {
                            if enqueue(entry: peNext!) {
                                didEnqueueNext = true
                                playlistController.PLCsetCurrentEntry(peNext!)
                            }
                            else {
                                //
                                // backup plan for a failed enqueue
                                //
                                print("PlC_prepareForNext: ERROR: enqueue failed, lets try next()")
                                didEnqueueNext = true
                                next()
                            }
                        }
                        else {
                            didEnqueueNext = true
                            next()
                        }
                    } else if acceptableStreamTypes().contains(peNext!.url!.scheme!) {
                        //
                        // next is HTTP stream.
                        //
                        didEnqueueNext = true
                        next()
                    }
                }
                else {
                    //
                    // stopafter or last
                    //
                    stop()
                    //
                    //  we want to select the next entry
                    //
                    _ = playlistController.next()
                }
            }
        }
    }
    
    @IBAction func record(_ sender: NSButton) {
        record(active: (sender.state == .on ? true : false))
    }
    
    func record(active: Bool)
    {
        if active {
            startStreamRecorder()
        } else {
            stopStreamRecorder()
        }
        recordButton.state = active ? .on : .off
    }
    
    func startStreamRecorder() {
        var properties: [AnyHashable : Any]  = [:]
        var audioFile: AVAudioFile? = nil
        let fileFormat: AVAudioCommonFormat = .pcmFormatInt16
        let recFolder: String = PhonixRecordingFolder.path
        let filePrefix = "Phonix_StreamRecording"
        var fileName: String = ""
        var fileExt = UserDefaults.standard.string(forKey: "recordFormat")
        var recFormat: UInt32 = 0
        var fileIndex: UInt = 0
        if (fileExt == "wav") {
            recFormat = kAudioFormatLinearPCM
        } else if (fileExt == "flac") {
            recFormat = kAudioFormatFLAC
        } else if (fileExt == "m4a") {
            recFormat = kAudioFormatMPEG4AAC
        } else {
            //
            // other extensions need to be implemented here
            //
            fileExt   = "wav"
            recFormat = kAudioFormatLinearPCM
        }
        //
        // set fileName with next index, don't mind the file extension
        //
        fileIndex = getNextIndex(ofFileWith: filePrefix, at: PhonixRecordingFolder)
        fileName  = String(format: "%@/%@-%03i.%@", recFolder, filePrefix, fileIndex, fileExt ?? "")
        //
        //  we are recording
        //
        StreamPlayer.appendFrameFilter(withName: "streamRecordingFilter", block: { channelsPerFrame, bytesPerFrame, frameCount, frames in
            let sampleRate: Float = 44100.0
            if audioFile == nil {
                properties = [AVFormatIDKey: recFormat, AVSampleRateKey: sampleRate, AVNumberOfChannelsKey: channelsPerFrame]
                //
                // set fileformat
                //
                if let properties = properties as? [String : Any] {
                    audioFile = try? AVAudioFile(forWriting: URL(fileURLWithPath: fileName), settings: properties, commonFormat: fileFormat, interleaved: true)
                }
            }
            let audioFormat = AVAudioFormat(commonFormat: fileFormat, sampleRate: Double(sampleRate), channels: AVAudioChannelCount(channelsPerFrame), interleaved: true)
            
            var pcmBuffer: AVAudioPCMBuffer? = nil
            if let audioFormat = audioFormat {
                pcmBuffer = AVAudioPCMBuffer(pcmFormat: audioFormat, frameCapacity: AVAudioFrameCount(frameCount))
            }
            pcmBuffer?.frameLength = AVAudioFrameCount(frameCount)
            memmove(pcmBuffer?.int16ChannelData![0], frames, Int(frameCount * bytesPerFrame))
            if let pcmBuffer = pcmBuffer {
                if (((try? audioFile?.write(from: pcmBuffer)) as ()??)) == nil {
                    print("PbC_startStreamRecorder: write error")
                }
            }
        })
    }
    
    func stopStreamRecorder() {
        StreamPlayer.removeFrameFilter(withName: "streamRecordingFilter")
    }
    // ----------Volume functions
    @IBAction func changeVolume(_ sender: AnyObject) {
        set(volume:sender.floatValue)
    }
    
    func set(volume: Float?) {
        //print("PbC_setVolume", volume)
        var newVol : Float
        
        if volume == nil { newVol = volumeSlider.floatValue }
        else { newVol = volume! }
        
        if newVol < 0.022 {
            //
            // we don't need to go below -99dB, below is mute, we don't want 3 figure values in the volumeText
            //
            newVol = 0.0
        }
        if setMixerVolume(newVol) == true {
            volumeSlider.floatValue = newVol
            
            if UserDefaults.standard.bool(forKey: "ShowMenuBarButtons") {
                statusBar?.set(volume: newVol)
            }
        }
    }
    
    func restoreVolume() {
        set(volume:UserDefaults.standard.float(forKey: "volume"))
    }
    // ----------[end]-Volume functions
    
    @IBAction func levelMeterButtonAction(_ sender: NSButton) {
        levelMeterEnable(sender.state == .on ? true : false)
    }
    
    func levelMeterEnable(_ state: Bool) {
        let miniSize = NSMakeSize(22.0, 22.0)
        let fullSize = NSMakeSize(200.0, 22.0)
        let halfSize = NSMakeSize(100.0, 22.0)
        
        levelMeterEnabled = state
        UserDefaults.standard.set(state, forKey: "levelMeterEnabled")
        
        if levelMeterEnabled {
            levelMeterItem.minSize = halfSize
            levelMeterItem.maxSize = fullSize
            levelMeterView.isHidden = false
            levelMeterEnableButton.state = .on
        }
        else {
            levelMeterItem.minSize = miniSize
            levelMeterItem.maxSize = miniSize
            levelMeterView.isHidden = true
            levelMeterEnableButton.state = .off
        }
        levelMeterRefresh()
    }
    
    @IBAction func spectrumAnalyzerButtonAction(_ sender: NSButton) {
        if sender.state == .on {
            spectrumAnalyzerEnable(true)
        } else {
            spectrumAnalyzerEnable(false)
        }
    }
    
    func spectrumAnalyzerEnable(_ state: Bool) {
        let analyzerWidth : CGFloat = CGFloat(spectrumAnalyzer.bandCount) * spectrumAnalyzer.barWidth + spectrumAnalyzer.barSpacing
        
        let miniSize = NSMakeSize(22.0, 22.0)
        let fullSize = NSMakeSize(analyzerWidth, 22.0)
        
        spectrumAnalyzerEnabled = state
        UserDefaults.standard.set(state, forKey: "spectrumAnalyzerEnabled")
        
        if spectrumAnalyzerEnabled {
            spectrumAnalyzerItem.minSize = fullSize
            spectrumAnalyzerItem.maxSize = fullSize
            spectrumAnalyzerView.isHidden = false
            spectrumAnalyzerEnableButton.state = .on
        }
        else {
            spectrumAnalyzerItem.minSize = miniSize
            spectrumAnalyzerItem.maxSize = miniSize
            spectrumAnalyzerView.isHidden = true
            spectrumAnalyzerEnableButton.state = .off
        }
        spectrumAnalyzer?.clearSpectralView()
    }
    
    func updatePosition() {
        if self.playbackStatus == PhonixStatus.phonixPlaying.rawValue {
            let currenttime = SFBcurrentTime()
            self.PBCsetPosition(currenttime)
            self.fileWave.updateProgress(SFBprogress())
            self.prepareForNext()
        }
        if self.playbackStatus == PhonixStatus.phonixHTTPPlaying.rawValue {
            let progress = StreamPlayer.progress()
            self.PBCsetPosition(progress)
        }
    }
    
    func levelMeterRefresh() {
        var audioLevels : [Float] = [-120.0, -120.0, -120.0, -120.0]
        
        if (playbackStatus == PhonixStatus.phonixPlaying.rawValue || playbackStatus == PhonixStatus.phonixHTTPPlaying.rawValue ) && levelMeterEnabled {
            getMeterLevels(&audioLevels)
        }
        DispatchQueue.main.async {
            self.levelMeter.setLevels(audioLevels)
        }
    }
    
    func spectrumFFTData(buffer: AudioBuffer?) -> UnsafeMutablePointer<Float>?
    {
        if (buffer == nil ) { return nil }
        
        var inData      : UnsafeMutableRawPointer?
        var inDataSize  : Int = 0
        let inFrames    : Int
        var div         : Float   = 0.5
        var vectData    : [Float]
        
        if playbackStatus == PhonixStatus.phonixHTTPPlaying.rawValue  || playbackStatus == PhonixStatus.phonixPlaying.rawValue {
            typealias dataTypo = Float32
            let inBuf       = buffer!
            inData          = inBuf.mData
            inDataSize      = Int(inBuf.mDataByteSize) / MemoryLayout<dataTypo>.size
            let chanStride  : vDSP_Stride = vDSP_Stride(inBuf.mNumberChannels)
            
            if inDataSize < 256 || inData == nil  {
                let freqData = UnsafeMutablePointer<Float32>.allocate(capacity: Int(inDataSize/chanStride))
                freqData.assign(repeating: 0.0, count: Int(inDataSize/chanStride))
                
                return freqData
            }
            inFrames    = inDataSize / chanStride
            vectData    = [dataTypo].init(repeating: 0.0, count: inDataSize)
            
            let inDataPntr = inData!.bindMemory(to: dataTypo.self, capacity: inDataSize)
            
            vDSP_vsdiv(inDataPntr, chanStride, &div, &vectData, 1, vDSP_Length(inFrames))
        } else {
            return nil
        }
        let frames    : UInt32      = UInt32(inFrames)
        let dspLength : vDSP_Length = vDSP_Length(frames / 2)
        let frameslog2N = UInt(log2u(frames))
        
        if  bufLog2N != frameslog2N || bufLog2N == 0 {
            
            if (fftSetup != nil) { vDSP_destroy_fftsetup(fftSetup) }
            bufLog2N = frameslog2N
            fftSetup = vDSP_create_fftsetup(bufLog2N, FFTRadix(kFFTRadix2))
        }
        var fftNormFactor   : Float         = 1.0 / (2.0 * Float(frames));
        
        var dataReal      = [Float](repeating: 0.0, count: Int(dspLength))
        var dataImag      = [Float](repeating: 0.0, count: Int(dspLength))
        var splitComplex = DSPSplitComplex(realp: &dataReal, imagp: &dataImag)
        //
        // Adds a real vector to a complex vector; single precision. ( to convert our [Float] to complex vector
        // Single-precision complex output vector.
        //
        vDSP_zrvadd(&splitComplex, 1, vectData, 1,  &splitComplex, 1, vDSP_Length(dspLength))
        
        vDSP_fft_zrip(fftSetup!, &splitComplex, 1, bufLog2N, FFTDirection(FFT_FORWARD))
        //
        // Scale the FFT data
        //
        vDSP_vsmul(splitComplex.realp, 1, &fftNormFactor, splitComplex.realp, 1, dspLength)
        vDSP_vsmul(splitComplex.imagp, 1, &fftNormFactor, splitComplex.imagp, 1, dspLength)
        
        let freqData = UnsafeMutablePointer<Float>.allocate(capacity: Int(dspLength))
        //
        // Single-precision complex input vector > Single-precision real output vector
        //
        vDSP_zvabs(&splitComplex , 1, freqData, 1, dspLength)
        return freqData
    }
    
    func spectrumAnalyzerRefresh() {
        if spectrumAnalyzerEnabled {
            let freqData = spectrumFFTData(buffer: self.buffer)
            spectrumAnalyzer.setFrequencyData(freqData: freqData)
        } else {
            DispatchQueue.main.async {
                self.spectrumAnalyzer.clearSpectralView()
            }
        }
    }
    
    func updateHTTPwave(buffer: AudioBuffer?) {
            if buffer!.mData != nil
            {
                var samples  = [Float32]()
                let dataSize = Int(buffer!.mDataByteSize) / MemoryLayout<Float32>.size
                let data     = buffer!.mData?.assumingMemoryBound(to: Float32.self)
                samples.append(contentsOf: UnsafeBufferPointer(start: data, count: dataSize))
                DispatchQueue.main.async {
                    self.httpWave.setSampleToWaveData(theSampleData: samples, length: dataSize)
                }
            }
            else {
                //
                // this happens, a few times every minute or less
                //
                let dataSize = Int(buffer!.mDataByteSize) / MemoryLayout<Float32>.size
                let samples  = [Float32](repeating: 0.0, count: dataSize)
                self.httpWave.setSampleToWaveData(theSampleData: samples, length: dataSize)
            }
    }
    
    func startTimers() {
        if UserDefaults.standard.bool(forKey: "waveFormPositionSlider")
            && playbackStatus == PhonixStatus.phonixHTTPPlaying.rawValue
        {
            httpWave.setDisabled(false)
        }
        else {
            httpWave.setDisabled(true)
        }
        startMainTimer()
    }
    
    func stopTimers() {
        stopMainTimer()
        levelMeterRefresh()
        spectrumAnalyzerRefresh()
    }
    
    func initDisplayLink() {
        displaylink = DisplayLink.init()
        displaylink.callback = mainTimerHandler
    }
    
    func startMainTimer()
    {
        displaylink.start()
    }
    
    func stopMainTimer()
    {
        displaylink.suspend()
    }
    
    @objc func mainTimerHandler()
    {
        let countsPerSec : Int = 60
        let multiplier   : Int = 2
        // 60Hz
        if timerCount % ( 1 * multiplier) == 0
        {
            // 30Hz
            if levelMeterEnabled { levelMeterRefresh() }
            if spectrumAnalyzerEnabled { spectrumAnalyzerRefresh() }
        }
        if timerCount % ( 2 * multiplier) == 0
        {
            // 15 Hz
            if httpWaveEnabled && StreamPlayer.state.rawValue == 3
            { updateHTTPwave(buffer: buffer) }
        }
        if timerCount % (15 * multiplier) == 0
        {
            // 2Hz
            updatePosition()
        }
        timerCount += 1
        if timerCount >= countsPerSec {
            timerCount = 0
        }
    }
    
    func printGraph() {
        if debugCount < 1 {
            printAUGraph()
            debugCount = 7
        }
        debugCount -= 1
    }
    
    @IBAction func showWaveButtonAction(_ sender: NSButton) {
        let state : Bool = sender.state == .on ? true : false
        wave(enable: state)
    }
    
    func wave(enable: Bool) {
        fileWaveEnable(enable)
        httpWaveEnable(enable)
        UserDefaults.standard.set(enable, forKey: "waveFormPositionSlider")
    }
    
    func fileWaveEnable(_ state: Bool) {
        fileWave.setDisabled(!state)
    }
    
    func httpWaveEnable(_ state: Bool) {
        httpWaveEnabled = state
        httpWave.setDisabled(!state)
    }
    
    func showWaveView(_ url: URL, withLength totalFrames: Int64) {
        showWaveView(url, startFrame: 0, withLength: UInt32(totalFrames))
    }
    
    func showWaveView(_ url: URL, startFrame: Int64, withLength totalFrames: UInt32) {
        if UserDefaults.standard.bool(forKey: "waveFormPositionSlider") {
            if url.isFileURL {
                if CoreAudioTypes().contains(url.pathExtension.lowercased()) {
                    fileWave.setDisabled(false)
                    fileWave.openAudioURL(url, startFrame: startFrame, withLength: totalFrames)
                }
                else {
                    fileWave.setDisabled(true)
                }
                showWaveViewFile()
            }
            else {
                httpWave.setDisabled(false)
                showWaveViewStream()
                //
                // waveform drawing is triggered by timer
                //
            }
        }
        else {
            //
            // we need to draw the line here. no wave
            //
            fileWave.setDisabled(true)
            httpWave.setDisabled(true)
            
            if url.isFileURL {
                showWaveViewFile()
            } else {
                showWaveViewStream()
            }
        }
    }
    
    func showWaveViewFile() {
        //
        //  show flac waveView hide others
        //
        fileWave.isHidden = false
        httpWave.isHidden = true
    }
    
    func showWaveViewStream() {
        //
        //  show Stream waveView hide others
        //
        fileWave.isHidden = true
        httpWave.isHidden = false
        httpWave.resetHistoryBuffers()
    }
    //
    // --------------------------- k - meter ---------------------------------------
    //
    func registerKMeter() -> AudioComponentDescription? {
        let cacheDirectory = PhonixFileManager.urls(for: .cachesDirectory, in: .userDomainMask)[0]
        var description = AudioComponentDescription(componentType: kAudioUnitType_Effect,
                                                    componentSubType: "kmet".OSType(),
                                                    componentManufacturer: "MZut".OSType(),
                                                    componentFlags: AudioComponentFlags.sandboxSafe.rawValue,
                                                    componentFlagsMask: 0)
        
        let bundleURL = CFBundleCopyResourceURLInDirectory(cacheDirectory as CFURL, "K-Meter" as CFString, "component" as CFString, nil)
        
        if bundleURL == nil {
            print("K-Meter.component not found\n")
            return nil
        }
        
        let bundle = CFBundleCreate(kCFAllocatorDefault,bundleURL)
        let function = CFBundleGetFunctionPointerForName(bundle, "KMeterAUFactory" as CFString)
        let factFunct = unsafeBitCast(function, to: AudioComponentFactoryFunction.self)
        
        let bundleVersion = CFBundleGetVersionNumber(bundle)
        /*
         dump(bundle,            name: "PbC_registerKMeter: bundle")
         dump(function,          name: "PbC_registerKMeter: function")
         dump(factFunct,         name: "PbC_registerKMeter: fact")
         dump(description,       name: "PbC_registerKMeter: description")
         
         debugPrint("PbC_registerKMeter: bundleVersion", bundleVersion)
         */
        _ = AudioComponentRegister(&description, "Martin Zuther: K-Meter" as CFString, bundleVersion, factFunct)
        return description
    }
    
    func initMeter(with description: AudioComponentDescription?) {
        //
        // we need to addMeter and initkMeterWindow
        //
        if description == nil { return }
        
        if addMeter(description!) {
            kMeterController = K_MeterController(withOptions: kMeter(), kName: "K-Meter")
            kMeterShow(show: UserDefaults.standard.bool(forKey: "kMeterOn"))
        } else {
            //
            // if k-Meter is not regged yet, we end up here
            //
            print("PbC_initKMeter: ERROR: meter not Added")
        }
    }
    
    func kMeterShow(show: Bool) {
        kMeterController.kMeterActive(show)
    }
    
    @IBAction func kMeterButtonAction(_ sender: NSButton) {
        if sender.state == .on {
            kMeterShow(show: true)
        }
        if sender.state == .off {
            kMeterShow(show: false)
        }
    }
    
    //
    // --------------------------- AU functions -------------------------------------
    //
    func initAUcontroller() {
        auController = AUController(title: "FX-Selector")
        auController.restoreAu()
    }
    
    @IBAction func fxButtonAction(_ sender: NSButton) {
        auController.toggleFxWindow()
        if UserDefaults.standard.bool(forKey: "BypassFX") {
            fxButton.state = .off
        } else {
            fxButton.state = .on
        }
    }
    //
    // menuBarControllerDelegates
    //    
    func statusBarAppIconLeftMouse() { mainWindow.orderFrontRegardless() }
    
    func statusBarAppIconRightMouse() { mainWindow.orderOut(self) }
    
    func statusBarPausePlay() { playPauseResume() }
    
    func statusBarNext() { next() }
    
    func statusBarPrev() { prev() }
    
    func statusBarSetVolume(_ volume: Float) { set(volume:volume) }
    
    //
    // STKAudioPlayer delegate stubs
    //
    func audioPlayer(_ audioPlayer: STKAudioPlayer, stateChanged state: STKAudioPlayerState, previousState: STKAudioPlayerState) {
        

        if (state.rawValue <= 5) {
            //
            // http is playing
            //
            recordButton.isHidden = false
        } else {
            recordButton.isHidden = true
        }
        //
        // set record state .off when stream stops
        //     
        if state.rawValue >= 9 {
            record(active: false)
        }
        //
        // 3, 5, 9, 16
        //
        /*
         STKAudioPlayerStateReady,
         STKAudioPlayerStateRunning      = 1,                                       // 1
         STKAudioPlayerStatePlaying      = (1 << 1) | STKAudioPlayerStateRunning,   // 3
         STKAudioPlayerStateBuffering    = (1 << 2) | STKAudioPlayerStateRunning,   // 5
         STKAudioPlayerStatePaused       = (1 << 3) | STKAudioPlayerStateRunning,   // 9
         STKAudioPlayerStateStopped      = (1 << 4),                                // 16
         STKAudioPlayerStateError        = (1 << 5),                                // 32
         STKAudioPlayerStateDisposed     = (1 << 6)                                 // 64
         */
        /*
         switch state.rawValue {
         case 3:
         // playing
         print("PbC_audioPlayer:stateChanged: playing")
         playbackStatus = PhonixStatus.phonixHTTPPlaying.rawValue
         case 5:
         // buffering
         print("PbC_audioPlayer:stateChanged: buffering")
         playbackStatus = PhonixStatus.phonixHTTPPending.rawValue
         case 9:
         // paused
         print("PbC_audioPlayer:stateChanged: paused")
         playbackStatus = PhonixStatus.phonixHTTPPaused.rawValue
         case 16...64:
         // stopped
         print("PbC_audioPlayer:stateChanged: stopped")
         playbackStatus = PhonixStatus.phonixHTTPStopped.rawValue
         default:
         break
         }
         */
    }
    
    func audioPlayer(_ audioPlayer: STKAudioPlayer, didUpdateMetaData metaDataDictionary: NSDictionary) {
        //print("PbC_audioPlayer didUpdateMetaData: ", metaDataDictionary)
        if let title = metaDataDictionary["StreamTitle"] as? String {
            playlistController.currentEntry?.artist = title.removingPercentEncoding!
        }
        if let url = metaDataDictionary["StreamUrl"] as? String {
            playlistController.currentEntry?.album = url.removingPercentEncoding!
        }
        //
        // we need to refresh the playlistView, for the data to show, dispatchQueue to prevent crash
        //
        DispatchQueue.main.async {
            self.playlistView.reloadData()
        }
    }
    
}
