import AppKit

class PlaylistController: NSArrayController, NSTableViewDataSource, NSTableViewDelegate, NSSearchFieldDelegate {
    @IBOutlet var playlistView          : PlaylistView!
    @IBOutlet var playlistLoader        : PlaylistLoader!
    @IBOutlet var playbackController    : PlaybackController!
    @IBOutlet var sideViewController    : SideViewController!
    @IBOutlet var shuffleAllMenuItem    : NSMenuItem!
    @IBOutlet var shuffleOffMenuItem    : NSMenuItem!
    @IBOutlet var repeatAllMenuItem     : NSMenuItem!
    @IBOutlet var repeatAlbumMenuItem   : NSMenuItem!
    @IBOutlet var repeatOneMenuItem     : NSMenuItem!
    @IBOutlet var repeatOffMenuItem     : NSMenuItem!
    @IBOutlet var searchField           : NSSearchField!
    @IBOutlet var repeatButton          : NSButton!
    @IBOutlet var shuffleButton         : NSButton!
    @IBOutlet var spinner               : NSProgressIndicator!
    @IBOutlet var spinnertext           : NSTextField!
    
    @objc dynamic var shuffle: ShuffleMode {
        return (ShuffleMode)(rawValue: UserDefaults.standard.integer(forKey: "shuffle"))!
    }
    
    @objc dynamic var phonixRepeat: RepeatMode {
        return (RepeatMode)(rawValue: UserDefaults.standard.integer(forKey: "repeat"))!
    }
    
    @objc dynamic var shuffleList = NSMutableArray()
    @objc dynamic var totalTime = ""
    @objc dynamic var currentEntry: PlaylistEntry?
    
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
        shuffleList = []
    }
    
    func startSpinner() {
        spinner.isHidden = false
        spinner.startAnimation(self)
        spinnertext.stringValue = "Phonix is loading..."
    }
    
    func stopSpinner() {
        spinner.stopAnimation(self)
        spinner.isHidden = true
        spinnertext.stringValue = ""
    }
    
    override func awakeFromNib() {
        super.awakeFromNib()
        playlistView?.registerForDraggedTypes([NSPasteboard.PasteboardType(rawValue: MovedRowsType),
                                               NSPasteboard.PasteboardType("NSFilenamesPboardType"),
                                               NSPasteboard.PasteboardType("NSURLPboardType"),
                                               NSPasteboard.PasteboardType("NSStringPboardType"),
                                               NSPasteboard.PasteboardType(rawValue: PhonixUrlsPboardType),
                                               NSPasteboard.PasteboardType(rawValue: iTunesDropType),
            ])
        
        playlistView?.needsDisplay = true
        playlistView.setDraggingSourceOperationMask(NSDragOperation.every, forLocal: false)
        playlistView.setDraggingSourceOperationMask(NSDragOperation.copy, forLocal: true)
        addObserver(self, forKeyPath: "arrangedObjects", options: [.new, .old], context: nil)
        playlistView.target = self
        playlistView.action = #selector(onItemClicked)
    }
    
    @objc private func onItemClicked() {
        if shuffle != ShuffleMode.ShuffleOff {
            resetShuffleList()
        }
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey: Any]?, context: UnsafeMutableRawPointer?) {
        if keyPath == "arrangedObjects" {
            updatePlaylistIndexes()
            updateTotalTime()
        }
    }
    
    func updatePlaylistIndexes() {
        let arranged = arrangedObjects as! NSArray
        for i in 0 ..< arranged.count {
            let pe = arranged[i] as? PlaylistEntry
            if pe?.peindex != i {
                // Make sure we don't get into some kind of crazy observing loop...
                pe?.peindex = i
            }
        }
    }
    
    func updateTotalTime() {
        var sec: Double = 0
        for pe: PlaylistEntry in arrangedObjects as! [PlaylistEntry] {
            if (pe.length.doubleValue).isNaN == false {
                sec += pe.length.doubleValue
            }
        }
        self.totalTime = "Total Time: \(secondsFormatter.string(for: sec))"
        stopSpinner()
        playlistView.reloadData()
    }
    //
    // PlayList tableView delegates
    //
    func tableViewSelectionDidChange(_ notification: Notification) {
        if selectedObjects!.count == 1 {
            let entry = selectedObjects[0] as! PlaylistEntry
            sideViewController.setMediaInfoDisplay(withEntry: entry)
        }
    }
    
    func tableView(_ tableView: NSTableView, sortDescriptorsDidChange oldDescriptors: [NSSortDescriptor]) {
        let sortDescriptor = NSSortDescriptor(key: "peindex", ascending: true)
        if tableView.sortDescriptors.contains(sortDescriptor) {
            super.sortDescriptors = []
            rearrangeObjects()
        } else {
            super.sortDescriptors = sortDescriptors
            rearrangeObjects()
        }
    }
    
    func tableView(_ tableView: NSTableView, writeRowsWith rowIndexes: IndexSet, to pboard: NSPasteboard) -> Bool {
        let data = try? NSKeyedArchiver.archivedData(withRootObject: rowIndexes, requiringSecureCoding: false)
        pboard.declareTypes([NSPasteboard.PasteboardType(rawValue: MovedRowsType)], owner: self)
        pboard.setData(data, forType: NSPasteboard.PasteboardType(rawValue: MovedRowsType))
        playlistView?.needsDisplay = true
        var filenames = [AnyHashable]()
        let row: Int = -1
        while (row == rowIndexes.first!) && (row <= rowIndexes.last!) && (row == rowIndexes.integerGreaterThan(row)!) {
            let song = ((arrangedObjects as! NSArray).object(at: row) as? PlaylistEntry)
            filenames.append((song?.url)!)
        }
        pboard.addTypes([NSPasteboard.PasteboardType("NSFilenamesPboardType")], owner: self)
        pboard.setPropertyList(filenames, forType: NSPasteboard.PasteboardType("NSFilenamesPboardType"))
        return true
    }
    
    func tableView(_ aTableView: NSTableView,
                   validateDrop info: NSDraggingInfo,
                   proposedRow row: Int,
                   proposedDropOperation operation: NSTableView.DropOperation) -> NSDragOperation {
        var dragOp: NSDragOperation = .copy
        if info.draggingSource as? NSTableView == aTableView {
            dragOp = .move
        }
        // we want to put the object at, not over,
        // the current row (contrast NSTableViewDropOn)
        aTableView.setDropRow(row, dropOperation: .above)
        playlistView?.needsDisplay = true
        
        return dragOp
    }
    
    func tableView(_ tableView: NSTableView, acceptDrop info: NSDraggingInfo, row: Int, dropOperation: NSTableView.DropOperation) -> Bool {
        var Row = row
        if Row < 0 {
            Row = 0
        }
        if dropOperation == .on || dropOperation == .above {
            let supportedTypes = [PhonixUrlsPboardType, MovedRowsType, NSPasteboard.PasteboardType("NSFilenamesPboardType"), NSPasteboard.PasteboardType("NSURLPboardType"), NSPasteboard.PasteboardType("NSStringPboardType"), iTunesDropType] as NSArray
            let pboard: NSPasteboard? = info.draggingPasteboard
            let bestType: String? = (pboard?.availableType(from: supportedTypes as! [NSPasteboard.PasteboardType])).map { $0.rawValue }
            var acceptedURLs = [URL]()
            // Get files from an file drawer drop
            if bestType == PhonixUrlsPboardType {
                let urls = ((try? NSKeyedUnarchiver.unarchiveTopLevelObjectWithData(info.draggingPasteboard.data(forType: NSPasteboard.PasteboardType(rawValue: PhonixUrlsPboardType)) ?? Data()) as? [Any]) as [Any]??)
                acceptedURLs.append(contentsOf: urls! as! [URL] )
            }
            if bestType == MovedRowsType {
                let indexSet = try? NSKeyedUnarchiver.unarchiveTopLevelObjectWithData(info.draggingPasteboard.data(forType: NSPasteboard.PasteboardType(rawValue: MovedRowsType)) ?? Data()) as? IndexSet
                let selected = (arrangedObjects as? NSArray)?.objects(at: indexSet!)
                moveObjectsInArrangedObjectsFromIndexes(from: indexSet! as NSIndexSet, to: UInt(Row))
                setSelectedObjects(selected!)
                playlistView?.reloadData()
                playlistView?.needsDisplay = true
                return true
            }
            // Get files from a normal file drop (such as from Finder)
            if bestType == NSPasteboard.PasteboardType("NSFilenamesPboardType").rawValue {
                var urls = [Any]() /// [AnyHashable]()
                for file: String in (info.draggingPasteboard.propertyList(forType: NSPasteboard.PasteboardType("NSFilenamesPboardType")) as? [String])! {
                    urls.append(URL(fileURLWithPath: file))
                }
                acceptedURLs.append(contentsOf: urls as! [URL])
            }
            // enable url drag from safari adressbar
            if bestType == NSPasteboard.PasteboardType("NSStringPboardType").rawValue {
                var pboard: NSPasteboard?
                pboard = info.draggingPasteboard
                if let draggedURL = NSURL(from: pboard!) as URL? {
                    acceptedURLs.append(draggedURL)
                }
            }
            // Get files from an iTunes drop
            if bestType == iTunesDropType {
                //            let iTunesDict = pboard?.propertyList(forType: NSPasteboard.PasteboardType(rawValue: iTunesDropType)) as Dictionary
                //            //            let tracks = iTunesDict? ["Tracks"] as? Dictionary
                //            let tracks = iTunesDict?.value(forKey: "Tracks") as? Dictionary
                //
                //            // Convert the iTunes URLs to URLs....MWAHAHAH!
                //            var urls = [AnyHashable]()
                //            for trackInfo: [AnyHashable: Any] in tracks.values {
                //                urls.append((URL(string: trackInfo.value(forKey: "Location") as? String ?? ""))!)
                //            }
                //            //[playlistLoader insertURLs:urls atIndex:row sort:YES];
                //            acceptedURLs.append(contentsOf: urls)
            }
            if acceptedURLs.count != 0 {
                willInsertURLs(acceptedURLs, origin: URLOrigin.URLOriginInternal)
                if (content as! NSArray).count == 0 {
                    Row = 0
                }
                let entries = playlistLoader.insertURLs(acceptedURLs, at: Row, sort: true)!
                didInsertURLs(entries, origin: URLOrigin.URLOriginInternal)
            }
            if shuffle != ShuffleMode.ShuffleOff {
                resetShuffleList()
            }
            playlistView?.reloadData()
            playlistView?.needsDisplay = true
            return true
        }
        return false
    }
    
    func moveObjectsInArrangedObjectsFromIndexes(from indexSet: NSIndexSet, to insertIndex: UInt) {
        var insertindex = insertIndex
        let objects = arrangedObjects as? NSArray
        var dragindex: Int = indexSet.lastIndex
        var aboveInsertIndexCount: Int = 0
        var object: Any?
        var removeIndex: Int
        while NSNotFound != dragindex {
            if dragindex >= insertindex {
                removeIndex = dragindex + aboveInsertIndexCount
                aboveInsertIndexCount += 1
            } else {
                removeIndex = dragindex
                insertindex -= 1
            }
            object = objects?.object(at: removeIndex)
            remove(atArrangedObjectIndex: removeIndex)
            insert(object!, atArrangedObjectIndex: Int(insertindex))
            dragindex = indexSet.indexLessThanIndex(dragindex)
        }
        var lowerIndex = Int(insertindex)
        while NSNotFound != lowerIndex {
            lowerIndex = indexSet.indexLessThanIndex(lowerIndex)
        }
    }
    
    func tableView(_ tableView: NSTableView, toolTipFor cell: NSCell, rect: NSRectPointer, tableColumn: NSTableColumn?, row: Int, mouseLocation: NSPoint) -> String {
        return ((arrangedObjects as! NSArray).object(at: row) as! PlaylistEntry).statusMessage()
    }
    
    override func insert(contentsOf objects: [Any], atArrangedObjectIndexes indexes: IndexSet) {
        super.insert(contentsOf: objects, atArrangedObjectIndexes: indexes)
        if shuffle != ShuffleMode.ShuffleOff {
            resetShuffleList()
        }
    }
    
    override func remove(atArrangedObjectIndexes indexes: IndexSet) {
        if currentEntry!.peindex >= 0 && indexes.contains(currentEntry!.peindex) {
            currentEntry!.peindex = -currentEntry!.peindex - 1
        }
        if (currentEntry?.peindex)! < 0 {
            var i: Int = -currentEntry!.peindex - 1
            var j: Int
            j = i - 1
            while j >= 0 {
                if indexes.contains(j) {
                    i -= 1
                }
                j -= 1
            }
            currentEntry!.peindex = -i - 1
        }
        remove(atArrangedObjectIndexes: indexes as IndexSet)
        if shuffle != ShuffleMode.ShuffleOff {
            resetShuffleList()
        }
    }
    
    func setShuffle(mode: ShuffleMode) {
        UserDefaults.standard.set(mode.rawValue, forKey: "shuffle")
        
        if mode == ShuffleMode.ShuffleOff {
            shuffleButton.toolTip = ShuffleButtonTooltipOff
            if #available(OSX 10.14, *) {
                shuffleButton.image = NSImage(named: NSImage.Name("ShuffleOffTemplate"))
            }
            else {
                let image = NSImage.init(named: "ShuffleOffTemplate")
                let imageRef = image!.cgImage(forProposedRect: nil, context: nil, hints: nil)
                let beginImage = CIImage(cgImage: imageRef!)
                let filter = CIFilter(name: "CIColorInvert")
                filter!.setValue(beginImage, forKey: kCIInputImageKey)
                let rep: NSCIImageRep = NSCIImageRep(ciImage: filter!.outputImage!)
                let nsImage: NSImage = NSImage(size: rep.size)
                nsImage.addRepresentation(rep)
                shuffleButton.image = nsImage
            }
            shuffleOffMenuItem.state = .on
            shuffleAllMenuItem.state = .off
            resetShuffleList()
        } else {
            shuffleButton.toolTip = ShuffleButtonTooltipAll
            if #available(OSX 10.14, *) {
                shuffleButton.image = NSImage(named: NSImage.Name("ShuffleOnTemplate"))
            }
            else {
                let image = NSImage.init(named: "ShuffleOnTemplate")
                let imageRef = image!.cgImage(forProposedRect: nil, context: nil, hints: nil)
                let beginImage = CIImage(cgImage: imageRef!)
                let filter = CIFilter(name: "CIColorInvert")
                filter!.setValue(beginImage, forKey: kCIInputImageKey)
                let rep: NSCIImageRep = NSCIImageRep(ciImage: filter!.outputImage!)
                let nsImage: NSImage = NSImage(size: rep.size)
                nsImage.addRepresentation(rep)
                shuffleButton.image = nsImage
            }
            shuffleOffMenuItem.state = .off
            shuffleAllMenuItem.state = .on
        }
    }
    
    @IBAction func toggleShuffle(_ sender: Any) {
        let mode = shuffle
        
        if mode == ShuffleMode.ShuffleOff {
            setShuffle(mode: ShuffleMode.ShuffleAll)
        } else {
            setShuffle(mode: ShuffleMode.ShuffleOff)
        }
    }
    
    @IBAction func setShuffleAll(_ sender: Any) {
        setShuffle(mode: ShuffleMode.ShuffleAll)
    }
    
    @IBAction func setShuffleOff(_ sender: Any) {
        setShuffle(mode: ShuffleMode.ShuffleOff)
    }
    
    func addShuffledListToFront() {
        let newList = Shuffle.shuffleList(arrangedObjects as! NSArray)
        let indexSet = NSIndexSet(indexesIn: NSRange(location: 0, length: newList.count))
        shuffleList.insert(newList as! [Any], at: indexSet as IndexSet)
        for i in 0 ..< shuffleList.count {
            (shuffleList.object(at: i) as! PlaylistEntry).shuffleIndex = i
        }
    }
    
    func addShuffledListToBack() {
        let newList = Shuffle.shuffleList(arrangedObjects as! NSArray)
        let indexSet = NSIndexSet(indexesIn: NSRange(location: shuffleList.count, length: newList.count))
        shuffleList.insert(newList as! [Any], at: indexSet as IndexSet)
        for i in (shuffleList.count - newList.count) ..< shuffleList.count {
            (shuffleList.object(at: i) as! PlaylistEntry).shuffleIndex = i
        }
    }
    
    func resetShuffleList() {
        shuffleList.removeAllObjects()
        addShuffledListToFront()
        
        if (currentEntry != nil) && (currentEntry?.peindex)! >= 0 {
            shuffleList.insert(currentEntry!, at: 0)
            currentEntry!.shuffleIndex = 0
            // Need to rejigger so the current entry is at the start now...
            var found: Bool = false
            var i = 1
            while i < shuffleList.count && !found {
                if (shuffleList.object(at: i) as! PlaylistEntry) == currentEntry! {
                    found = true
                    shuffleList.removeObject(at: i)
                } else {
                    (shuffleList.object(at: i) as! PlaylistEntry).shuffleIndex = i
                }
                i += 1
            }
        }
    }
    
    func shuffledEntry(at i: Int) -> PlaylistEntry? {
        var j = i
        if j < 0 {
            if phonixRepeat == RepeatMode.RepeatAll {
                addShuffledListToFront()
                // change i appropriately
                j += (arrangedObjects as AnyObject).count
            } else {
                j += 2
            }
        }
        if j >= shuffleList.count {
            if phonixRepeat == RepeatMode.RepeatAll {
                addShuffledListToBack()
            } else {
                return shuffleList.object(at: j - 1) as? PlaylistEntry
            }
        }
        return shuffleList.object(at: j) as? PlaylistEntry
    }
    
    func setRepeat(mode: RepeatMode) {
        UserDefaults.standard.set(mode.rawValue, forKey: "repeat")
        if mode == RepeatMode.RepeatOne {
            repeatButton.toolTip = RepeatButtonTooltip
            if #available(OSX 10.14, *) {
                repeatButton.image = NSImage(named: NSImage.Name("RepeatModeOneTemplate"))
            }
            else {
                let image = NSImage.init(named: "RepeatModeOneTemplate")
                let imageRef = image!.cgImage(forProposedRect: nil, context: nil, hints: nil)
                let beginImage = CIImage(cgImage: imageRef!)
                let filter = CIFilter(name: "CIColorInvert")
                filter!.setValue(beginImage, forKey: kCIInputImageKey)
                let rep: NSCIImageRep = NSCIImageRep(ciImage: filter!.outputImage!)
                let nsImage: NSImage = NSImage(size: rep.size)
                nsImage.addRepresentation(rep)
                repeatButton.image = nsImage
            }
            repeatOffMenuItem.state = .off
            repeatOneMenuItem.state = .on
            repeatAlbumMenuItem.state = .off
            repeatAllMenuItem.state = .off
        } else if mode == RepeatMode.RepeatAlbum {
            repeatButton.toolTip = RepeatButtonTooltipAlbum
            if #available(OSX 10.14, *) {
                repeatButton.image = NSImage(named: NSImage.Name("RepeatModeAlbumTemplate"))
            }
            else {
                let image = NSImage.init(named: "RepeatModeAlbumTemplate")
                let imageRef = image!.cgImage(forProposedRect: nil, context: nil, hints: nil)
                let beginImage = CIImage(cgImage: imageRef!)
                let filter = CIFilter(name: "CIColorInvert")
                filter!.setValue(beginImage, forKey: kCIInputImageKey)
                let rep: NSCIImageRep = NSCIImageRep(ciImage: filter!.outputImage!)
                let nsImage: NSImage = NSImage(size: rep.size)
                nsImage.addRepresentation(rep)
                repeatButton.image = nsImage
            }
            repeatOffMenuItem.state = .off
            repeatOneMenuItem.state = .off
            repeatAlbumMenuItem.state = .on
            repeatAllMenuItem.state = .off
        } else if mode == RepeatMode.RepeatAll {
            repeatButton.toolTip = RepeatButtonTooltipAll
            if #available(OSX 10.14, *) {
                repeatButton.image = NSImage(named: NSImage.Name("RepeatModeAllTemplate"))
            }
            else {
                let image = NSImage.init(named: "RepeatModeAllTemplate")
                let imageRef = image!.cgImage(forProposedRect: nil, context: nil, hints: nil)
                let beginImage = CIImage(cgImage: imageRef!)
                let filter = CIFilter(name: "CIColorInvert")
                filter!.setValue(beginImage, forKey: kCIInputImageKey)
                let rep: NSCIImageRep = NSCIImageRep(ciImage: filter!.outputImage!)
                let nsImage: NSImage = NSImage(size: rep.size)
                nsImage.addRepresentation(rep)
                repeatButton.image = nsImage
            }
            repeatOffMenuItem.state = .off
            repeatOneMenuItem.state = .off
            repeatAlbumMenuItem.state = .off
            repeatAllMenuItem.state = .on
        } else if mode == RepeatMode.RepeatNone {
            repeatButton.toolTip = RepeatButtonTooltipOff
            if #available(OSX 10.14, *) {
                repeatButton.image = NSImage(named: NSImage.Name("RepeatModeOffTemplate"))
            }
            else {
                let image = NSImage.init(named: "RepeatModeOffTemplate")
                let imageRef = image!.cgImage(forProposedRect: nil, context: nil, hints: nil)
                let beginImage = CIImage(cgImage: imageRef!)
                let filter = CIFilter(name: "CIColorInvert")
                filter!.setValue(beginImage, forKey: kCIInputImageKey)
                let rep: NSCIImageRep = NSCIImageRep(ciImage: filter!.outputImage!)
                let nsImage: NSImage = NSImage(size: rep.size)
                nsImage.addRepresentation(rep)
                repeatButton.image = nsImage
            }
            repeatOffMenuItem.state = .on
            repeatOneMenuItem.state = .off
            repeatAlbumMenuItem.state = .off
            repeatAllMenuItem.state = .off
        }
    }
    
    @IBAction func toggleRepeat(_ sender: Any) {
        let mode = phonixRepeat
        
        if mode == RepeatMode.RepeatNone {
            setRepeat(mode: RepeatMode.RepeatOne)
        } else if mode == RepeatMode.RepeatOne {
            setRepeat(mode: RepeatMode.RepeatAlbum)
        } else if mode == RepeatMode.RepeatAlbum {
            setRepeat(mode: RepeatMode.RepeatAll)
        } else if mode == RepeatMode.RepeatAll {
            setRepeat(mode: RepeatMode.RepeatNone)
        }
    }
    
    @IBAction func setRepeatAll(_ sender: Any) {
        setRepeat(mode: RepeatMode.RepeatAll)
    }
    
    @IBAction func setRepeatAlbums(_ sender: Any) {
        setRepeat(mode: RepeatMode.RepeatAlbum)
    }
    
    @IBAction func setRepeatOne(_ sender: Any) {
        setRepeat(mode: RepeatMode.RepeatOne)
    }
    
    @IBAction func setRepeatOff(_ sender: Any) {
        setRepeat(mode: RepeatMode.RepeatOne)
    }
    
    func entry(at index: Int) -> PlaylistEntry? {
        if index >= (arrangedObjects as AnyObject).count || index < 0 {
            return nil
        }
        return (arrangedObjects as! NSArray).object(at: index) as? PlaylistEntry
    }
    
    func indexOfLastEntry() -> Int {
        let index = (arrangedObjects as AnyObject).count - 1
        return index
    }
    
    func filterPlaylist(onAlbum album: String) -> [Any] {
        let predicate = NSPredicate(format: "album like %@", album)
        return ((arrangedObjects as? NSArray)?.filtered(using: predicate))!
    }
    
    func prevEntry() -> PlaylistEntry? {
        let pe = currentEntry
        
        if pe == nil { return entry(at: 0) }
        
        if phonixRepeat == RepeatMode.RepeatOne {
            return pe
        }
        if shuffle != ShuffleMode.ShuffleOff {
            return shuffledEntry(at: pe!.shuffleIndex - 1)!
        }
        var i: Int = 0
        if pe!.peindex < 0 { i = -pe!.peindex - 2 } else { i = pe!.peindex - 1 }
        
        if phonixRepeat == RepeatMode.RepeatAlbum && pe!.album != "" && pe!.isSeekable {
            //
            // we are in album repeat mode, we want the prevEntry of all album tracks in the list.
            //
            let albumTracks: NSArray = filterPlaylist(onAlbum: pe!.album) as NSArray
            //
            // we need the next entry from albumTracks
            //
            var index : Int = albumTracks.index(of: pe! as PlaylistEntry)
            
            if index != NSNotFound {
                //
                // we are playing an albumTrack
                //
                if index > 0 {
                    index -= 1
                } else {
                    index = albumTracks.count - 1
                }
                
                i = (albumTracks.object(at: index) as! PlaylistEntry).peindex
            }
        } else if phonixRepeat == RepeatMode.RepeatAll {
            if entry(at: i) == nil { i = indexOfLastEntry() }
        }
        return entry(at: i)
    }
    
    @IBAction func stop(afterCurrent sender: Any) {
        currentEntry?.isStopAfter = !(currentEntry?.isStopAfter)!
        playlistView.reloadData()
    }
    
    func nextEntry() -> PlaylistEntry? {
        let pe = currentEntry
        if pe == nil { return entry(at: 0) }
        if phonixRepeat == RepeatMode.RepeatOne {
            return pe
        }
        if shuffle != ShuffleMode.ShuffleOff {
            return shuffledEntry(at: pe!.shuffleIndex + 1)!
        }
        var i: Int = 0
        if pe!.peindex < 0 { i = -pe!.peindex - 1 } else { i = pe!.peindex + 1 }
        if phonixRepeat == RepeatMode.RepeatAlbum && pe!.album != "" && pe!.isSeekable {
            //
            // we are in album repeat mode, we want the nextEntry of all album tracks in the list.
            //
            let albumTracks: NSArray = filterPlaylist(onAlbum: pe!.album) as NSArray
            //
            // we need the next entry from albumTracks
            //
            var index : Int = albumTracks.index(of: pe! as PlaylistEntry)
            
            if index != NSNotFound {
                //
                // we are playing an albumTrack
                //
                if index < albumTracks.count - 1 {
                    index += 1
                } else {
                    index = 0
                }
                
                i = (albumTracks.object(at: index) as! PlaylistEntry).peindex
            }
        } else if phonixRepeat == RepeatMode.RepeatAll {
            if entry(at: i) == nil { i = 0 }
        }
        return entry(at: i)
    }
    
    func prev() -> Bool {
        //
        // not used
        //
        var pe: PlaylistEntry?
        pe = prevEntry() //currentEntry)
        if pe == nil { return false }
        PLCsetCurrentEntry(pe!)
        return true
    }
    
    func next() -> Bool {
        var pe: PlaylistEntry?
        pe = nextEntry()
        
        if pe == nil {
            return false
        }
        PLCsetCurrentEntry(pe!)
        return true
    }
    
    func PLCsetCurrentEntry(_ pe: PlaylistEntry) {
        //
        // if metadata is not loaded or we have en error, just check again,
        // maybe your nas has mounted in the mean time ;)
        //
        if !pe.isMetadataLoaded || pe.isError {
            // Force loading metadata if it isn't loaded,
            // will hopefully prevent wrong Growl notifications,
            // non-updating progress slider and wrong remaining time
            if let metadata = playlistLoader.readEntryInfo(pe: pe) {
                pe.setMeta(data: metadata)
            }
        }
        currentEntry?.isCurrent = false
        currentEntry?.isStopAfter = false
        pe.isCurrent = true
        sideViewController.setMediaInfoDisplay(withEntry: pe)
        currentEntry = pe
        playlistView.reloadData()
        playlistView?.scrollRowToVisible(pe.peindex)
    }
    
    @IBAction func clearFilterPredicate(_ sender: Any) {
        searchField.stringValue = ""
    }
    
    @IBAction func clear(_ sender: Any) {
        searchField.stringValue = ""
        super.remove(atArrangedObjectIndexes: NSIndexSet(indexesIn: NSRange(location: 0, length: (arrangedObjects as AnyObject).count)) as IndexSet)
    }
    
    @IBAction func removeitem(_ sender: Any) {
        // It's a kind of magic.
        // Plain old NSArrayController's remove: isn't working properly for some reason.
        // The method is definitely called but (overridden) removeObjectsAtArrangedObjectIndexes: isn't called
        // and no entries are removed.
        // Putting explicit call to removeObjectsAtArrangedObjectIndexes: here for now.
        // TODO: figure it out
        let selected = selectionIndexes as NSIndexSet
        if selected.count > 0 {
            super.remove(atArrangedObjectIndexes: selected as IndexSet)
        }
    }
    
    @IBAction func showEntry(inFinder sender: Any) {
        let ws: NSWorkspace? = NSWorkspace.shared
        if NSNotFound == selectionIndex {
            return
        }
        let url: URL? = (selectedObjects[0] as! PlaylistEntry).url
        if url?.isFileURL != nil {
            ws?.selectFile(url?.path, inFileViewerRootedAtPath: (url?.path)!)
        }
    }
    
    func willInsertURLs(_ urls: [URL], origin: URLOrigin) {
        if urls.count == 0 { return }
        let event: CGEvent = CGEvent(source: nil)!
        let modifiers: CGEventFlags = event.flags
        //
        // drag             : add file to playlist
        // option  + drag   : clear and add file to playlist
        //
        let altPressed = modifiers.contains(.maskAlternate)
        //print("PlC_willInsertURLs", "\naltPressed   :", altPressed, "origin:", origin.rawValue)
        
        if altPressed {
            clear(self)
        }
    }
    
    func didInsertURLs(_ urls: [Any], origin: URLOrigin) {
        if urls.count == 0 { return }
        let event: CGEvent = CGEvent(source: nil)!
        let modifiers: CGEventFlags = event.flags
        //
        // drag             : add file to playlist
        // shift   + drag   : add file to playlist and play
        // command + drag   : add file to playlist and play
        // option  + drag   : clear and add file to playlist
        // command + option : clear add file to playlist and play
        // ctrl             : not used as modifier, because this will popup Menu.
        //
        let cmdPressed      = modifiers.contains(.maskCommand)
        let shiftPressed    = modifiers.contains(.maskShift)
        if origin == .URLOriginInternal {
            //
            // when is external used?
            //
        }
        if cmdPressed || shiftPressed {
            _ = playbackController.play(entry: urls[0] as? PlaylistEntry)
        }
    }
    //
    // metadata edit etc.
    //
    func controlTextDidChange(_ notification: Notification) {
        //print("PlC_controlTextDidChange(): Notification", (notification.object as! NSTableView).objectValue as Any )
        //
        // the Notification comes directly when the text is changed, we don't get a confirmation from Enter
        // maybe we can do this better, with a more apropriate trigger.
        //
        guard notification.object as? NSTableView == playlistView else { return }
        
//        let index = (notification.object as! NSTableView).selectedRow
//        let url: URL? = (arrangedObjects as! [PlaylistEntry])[index].url
        
        //for entry in selectedObjects as! [PlaylistEntry] { print("PlC controlTextDidChange:", entry.url as Any ) }
        
        switch playlistView.columnClicked
        {
        case playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "title")) :
            let title = (notification.object as! NSTableView).stringValue
            for entry in selectedObjects as! [PlaylistEntry] {
                let url = entry.url
                TagLibsetTitle( url!.path.cString(using: String.Encoding.utf8), title.cString(using: String.Encoding.utf8))
            }
        case playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "artist")) :
            let artist = (notification.object as! NSTableView).stringValue
            for entry in selectedObjects as! [PlaylistEntry] {
                let url = entry.url
                TagLibsetArtist(url!.path.cString(using: String.Encoding.utf8), artist.cString(using: String.Encoding.utf8))
            }
            
        case playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "album")) :
            let album = (notification.object as! NSTableView).stringValue
            for entry in selectedObjects as! [PlaylistEntry] {
                let url = entry.url
                TagLibsetAlbum( url!.path.cString(using: String.Encoding.utf8), album.cString(using: String.Encoding.utf8))
            }
        case playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "genre")) :
            let genre = (notification.object as! NSTableView).stringValue
            for entry in selectedObjects as! [PlaylistEntry] {
                let url = entry.url
                TagLibsetGenre( url!.path.cString(using: String.Encoding.utf8), genre.cString(using: String.Encoding.utf8))
            }
        case playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "year")) :
            let year = UInt32(abs((notification.object as! NSTableView).intValue))
            for entry in selectedObjects as! [PlaylistEntry] {
                let url = entry.url
                TagLibsetYear(  url!.path.cString(using: String.Encoding.utf8), year)
            }
        case playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "track")) :
            let track = UInt32(abs((notification.object as! NSTableView).intValue))
            for entry in selectedObjects as! [PlaylistEntry] {
                let url = entry.url
                TagLibsetTrack( url!.path.cString(using: String.Encoding.utf8), track)
            }
        default: return
        }
    }
    
    @IBAction func editMetaData(_ sender: Any) {
        
        if      playlistView.columnClicked == playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "title"))
            ||  playlistView.columnClicked == playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "artist"))
            ||  playlistView.columnClicked == playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "album"))
            ||  playlistView.columnClicked == playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "genre"))
            ||  playlistView.columnClicked == playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "year"))
            ||  playlistView.columnClicked == playlistView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "track"))
        {
            //print("PlC editMetaData(): edit:", playlistView.columnClicked!, playlistView.selectedRow)
            playlistView.editColumn(playlistView.columnClicked!, row: playlistView.selectedRow, with: nil, select: true)
        }
    }
}

