import AppKit

@objcMembers
class PlaylistEntry: NSObject {
    var peindex          : Int      = 0
    var shuffleIndex     : Int      = 0
    var isCurrent        : Bool     = false
    var isRemoved        : Bool     = false
    var isStopAfter      : Bool     = false
    var isQueued         : Bool     = false
    var queuePosition    : Int      = 0
    var isError          : Bool     = false
    var errorMessage     : String   = ""
    var url              : URL?             // can URL be optional? There is no entry without an URL..
    var URLCue           : URL?
    var artist           : String   = ""
    var album            : String   = ""
    var title            : String   = ""
    var genre            : String   = ""
    var year             : NSNumber?
    var track            : NSNumber?
    var albumArt         : NSImage?
    var totalFrames      : Int64    = 0
    var bitrate          : Int      = 0
    var channels         : Int      = 0
    var bitsPerSample    : Int      = 0
    var sampleRate       : Float    = 0.0
    var startCue         : Double   = -1.0
    var endCue           : Double   = -1.0
    var fileFormat       : String?
    var endian                      = ""
    var isSeekable       : Bool     = false
    var isMetadataLoaded : Bool     = false

    func display() -> String {
        if !isError {
            return String(format: "%@ - %@ | %@%@%@",
                          title,
                          artist,
                          String(format: "%ich ", channels),
                          url!.isFileURL ? (bitsPerSample > 0 ? String(format: "%ibit ", bitsPerSample) : "--bit ") : "http Stream",
                          url!.isFileURL ? String(format: "%0.1fkHz", sampleRate / 1000) : "")
        } else {
            return "Error"
        }
    }
    
    var length: NSNumber {
        if !(startCue == -1.0 && endCue == -1.0 ) { // cue file
            if endCue != 0.0 {
                if  endCue < startCue { // prevent crash in case of corrupt cue file
                    return 0.0 as NSNumber
                }
                return endCue - startCue as NSNumber
            } else {
                return NSNumber(value: (Double(totalFrames) / Double(sampleRate)) - Double(startCue))
            }
        }
        return NSNumber(value: Double(totalFrames) / Double(sampleRate))
    }
    
    var path: String {
        return url!.deletingLastPathComponent().path
    }
    
    var fileName: String {
        return url!.lastPathComponent
    }

    func lengthText() -> String {
        let secondsFormatter = SecondsFormatter()
        let time: String? = secondsFormatter.string(for: length)
        return time ?? ""
    }

    func fileinfo() -> String {
        if !isError {
            return String(format: "%@%@%@%@",
                            channels != 0 ? String(format: "%ich ", channels) : "",
                            url!.isFileURL ? (bitsPerSample > 0 ? String(format: "%ibit ", bitsPerSample) : " ") : "http Stream",
                            url!.isFileURL && sampleRate > 0 ? String(format: "%0.1fkHz ", sampleRate / 1000) : "",
                            url!.isFileURL && bitsPerSample > 0 ? (fileFormat != nil ? fileFormat! : (URL(fileURLWithPath: url!.path).pathExtension).lowercased()) : "")
        } else {
            return "file not found"
        }
    }

    func status() -> String {
        if isStopAfter {
            return "stopAfter"
        } else if isCurrent {
            return "playing"
        } else if isError && !(acceptableStreamTypes().contains(url!.scheme!)) {
            return "error"
        }
        return ""
    }

    func statusMessage() -> String {
        if isStopAfter {
            return "Stopping once finished..."
        } else if isCurrent {
            return "Playing..."
        } else if isError {
            return errorMessage 
        }
        return ""
    }

    func setMeta(data: [String : Any]) {
        let lockQueue = DispatchQueue(label: "self")
        lockQueue.sync {
            // support for reading and setting metadata from different threads
            // see readEntryInfo and loadInfoForEntires in PlaylistLoader and
            // setCurrentEntry in playlistController
            if data.isEmpty {
                self.isError = true
                self.errorMessage = "File Not Found."
            } else {
                self.setValuesForKeys(data)
                self.isError = false
                self.isMetadataLoaded = true
            }
        }
    }
}
