import AVFoundation
import AppKit

class PlaylistLoader: NSObject
{
     @IBOutlet var playlistController : PlaylistController!
     @IBOutlet var spinnertext        : NSTextField!
     @IBOutlet var menuspinnertext    : NSTextField!
     @IBOutlet var spinner            : NSProgressIndicator!
     @IBOutlet var menuspinner        : NSProgressIndicator!
     
     func saveM3u(_ filename: String) -> Bool {
          let fileHandle: FileHandle! = fileHandleForWritingAtPath(path: filename, create: true)
          if fileHandle == nil {
               return false
          }
          fileHandle?.truncateFile(atOffset: 0)
          for pe: PlaylistEntry in playlistController.arrangedObjects as! Array {
               let path: NSString = relativePath(from: filename, to: pe.url!) as NSString
               fileHandle!.write(path.appending("\n").data(using: String.Encoding.utf8)!)
          }
          fileHandle!.closeFile()
          return true
     }
     
     func insertURLs(_ urls: [URL], at insertindex: Int, sort: Bool) -> [PlaylistEntry]?
     {
          let uniqueURLs      : NSMutableSet = []
          var expandedURLs    : [URL] = []
          var containedURLs   : [URL] = []
          var fileURLS        : [URL] = []
          var validURLs       : [URL] = []
          
          if urls.isEmpty { return [] }
          
          startSpinner("loading Playlist")
          var InsertIndex = insertindex
          if InsertIndex < 0 {
               InsertIndex = 0
          }
          for unsortedURL in urls {
               if unsortedURL.isFileURL {
                    var isDir: ObjCBool = ObjCBool(false)
                    let exists: Bool = PhonixFileManager.fileExists(atPath: unsortedURL.path, isDirectory: &isDir)
                    if exists && isDir.boolValue {
                         // Exists. Directory.
                         expandedURLs.append(contentsOf: allFileURLs(atPath: unsortedURL.path)! )
                    } else if exists {
                         // Exists.
                         expandedURLs.append(unsortedURL)
                    }
               } else {
                    // Non-file URL..
                    expandedURLs.append(unsortedURL)
               }
          }
          var sortedURLs: [URL] = []
          
          if sort == true {
               sortedURLs = expandedURLs.sorted{ $0.path.caseInsensitiveCompare($1.path) == ComparisonResult.orderedAscending }
          } else {
               sortedURLs = expandedURLs
          }
          
          for (_, url) in sortedURLs.enumerated()
          {
               let sortedURL : URL = url
               // Container vs non-container url
               if acceptableContainerTypes().contains(sortedURL.pathExtension.lowercased() ) {
                    if url.pathExtension.lowercased() == "cue" {
                         containedURLs.append(contentsOf: cueURLs(forContainerURL: sortedURL))
                    } else {
                         if sortedURL.isFileURL {
                              if let entries = fileURLs(forContainerURL: sortedURL) {
                                   containedURLs.append(contentsOf: entries)
                              } else {
                                   return nil
                              }
                         } else {
                              if let entries = httpURLs(forContainerURL: sortedURL) {
                                   containedURLs.append(contentsOf: entries)
                              } else {
                                   return nil
                              }
                         }
                    }
                    // Make sure the container isn't added twice.
                    if !containedURLs.isEmpty {
                         uniqueURLs.addObjects(from: containedURLs)
                    }
               }
               else
               {
                    fileURLS.append(sortedURL)
               }
          }
          
          for fileURL in fileURLS {
               if fileURL.isFileURL && !acceptableFileTypes().contains(fileURL.pathExtension.lowercased()) {
                    continue
               }
               if !uniqueURLs.contains(fileURL) {
                    validURLs.append(fileURL)
                    uniqueURLs.add(fileURL)
               }
          }
          
          for containedURL in containedURLs {
               if !acceptableSchemes().contains(containedURL.scheme!) {
                    continue
               }
               // Need a better way to determine acceptable file types than basing it on extensions.
               if containedURL.isFileURL && !acceptableFileTypes().contains(containedURL.pathExtension.lowercased()) {
                    continue
               }
               validURLs.append(containedURL)
          }
          // Create actual entries
          var entries : [PlaylistEntry] = []
          var index = 0
          
          for url in validURLs {
               let pe = PlaylistEntry()
               pe.url = url
               pe.peindex = index
               if acceptableStreamTypes().contains(pe.url!.scheme!) {
                    pe.title = url.absoluteString
               } else {
                    pe.title = url.lastPathComponent
               }
               pe.queuePosition = -1
               entries.append(pe)
               index += 1
          }
          
          let indexSet = NSIndexSet(indexesIn: NSRange(location: InsertIndex, length: entries.count))
          playlistController.insert(contentsOf: entries, atArrangedObjectIndexes: indexSet as IndexSet)
          // Select the first entry in the group that was just added
          playlistController.setSelectionIndex(InsertIndex)
          //
          // Without global dispatch: we only see a pizza wheel untill done loading, all data appears at once, spinner is not visible
          // With global dispatch   : we see the list filling in, and spinner is visible
          //
          DispatchQueue.global(qos: .background).async {
               self.loadInfo(forEntries: entries)
               DispatchQueue.main.async {
                    // only back on the main thread, may you access UI:
                    self.playlistController.updateTotalTime()
                    self.stopSpinner()
               }
          }
          return entries
     }
     
     func loadInfo(forEntries entries: [PlaylistEntry]) {
          let queue = OperationQueue()
          for entry in entries {
               let operation = BlockOperation {
                    if let metadata = self.readEntryInfo(pe: entry) {
                         DispatchQueue.main.async {
                              entry.setMeta(data: metadata)
                         }
                    }
               }
               queue.addOperation(operation)
          }
          queue.waitUntilAllOperationsAreFinished()
     }
     
     func readEntryInfo(pe: PlaylistEntry) -> [String : Any]? {
          let lockQueue = DispatchQueue(label: "self")
          var entryInfo       = [String : Any]()
          var entryProperties = [String : Any]()
          var entryMetaData   = [String : Any]()
          lockQueue.sync {
               if pe.url?.pathExtension.lowercased() == "cue" && (pe.url?.fragment) != nil {
                    entryMetaData = cueFileInfo(for: pe.url!)!
                    let metadata  = SFBmetadataForURL(pe.url! as CFURL)
                    
                    if entryMetaData["artist"] == nil && metadata.artist != nil {
                         entryMetaData["artist"] = String(cString: metadata.artist)
                    }
                    if entryMetaData["title"] == nil && metadata.title != nil {
                         entryMetaData["title"] = String(cString: metadata.title)
                    }
                    if entryMetaData["album"] == nil && metadata.album != nil {
                         entryMetaData["album"] = String(cString: metadata.album)
                    }
                    if entryMetaData["genre"] == nil && metadata.genre != nil {
                         entryMetaData["genre"] = String(cString: metadata.genre)
                    }
                    if entryMetaData["year"] == nil && metadata.year != nil {
                         entryMetaData["year"] = String(cString: metadata.year)
                    }
                    if entryMetaData["track"] == nil && metadata.track > 0 {
                         entryMetaData["track"] = metadata.track
                    }
                    if metadata.artwork != nil {
                         let imagedata = Data(bytes: metadata.artwork, count: Int(metadata.imagelength))
                         if let image = NSImage(data: imagedata) {
                              entryMetaData["albumArt"] = image
                         }
                    } else {
                         entryMetaData["albumArt"] = getAlbumArtFromDir(for: pe.url!)
                    }
               } else {
                    if acceptableStreamTypes().contains(pe.url!.scheme!) {
                         entryMetaData = ["title": pe.title]
                    } else {
                         let data = SFBmetaAndPropsForURL(pe.url! as CFURL)
                         
                         //if !(PhonixFileManager.fileExists(atPath: pe.url!.path)) && !pe.url!.isFileURL { return }
                         //print("readEntryinfo:", data as! [String: Any])
                         
                         if data.artist != nil {
                              entryMetaData["artist"] = String(cString: data.artist)
                         }
                         if data.title != nil {
                              entryMetaData["title"] = String(cString: data.title)
                         }
                         if data.album != nil {
                              entryMetaData["album"] = String(cString: data.album)
                         }
                         if data.genre != nil {
                              entryMetaData["genre"] = String(cString: data.genre)
                         }
                         if data.year != nil {
                              entryMetaData["year"] = String(cString: data.year)
                         }
                         if data.track > 0 {
                              entryMetaData["track"] = data.track
                         }
                         if data.artwork != nil {
                              let imagedata = Data(bytes: data.artwork, count: Int(data.imagelength))
                              if let image = NSImage(data: imagedata) {
                                   entryMetaData["albumArt"] = image
                              }
                         }
                         else {
                              entryMetaData["albumArt"] = self.getAlbumArtFromDir(for: pe.url!)
                         }
                         
                         if data.channels == 0 || data.bitspersample == 0 || data.samplerate == 0 || data.totalframes == 0 {
                              if let prop = propertiesForURL(url: pe.url!) {
                                   entryProperties = prop
                              }
                         } else {
                              entryProperties["channels"]        = data.channels
                              entryProperties["bitsPerSample"]   = data.bitspersample
                              entryProperties["sampleRate"]      = data.samplerate
                              entryProperties["totalFrames"]     = data.totalframes
                              entryProperties["seekable"]        = true
                              if data.fileformat != nil {
                                   entryProperties["fileFormat"] = String(cString: data.fileformat).lowercased()
                              }
                              else {
                                   entryProperties["fileFormat"] = pe.url!.pathExtension.lowercased()
                              }
                         }
                    }
               }
               entryInfo.merge(entryMetaData)   { (current, _) in current }
               entryInfo.merge(entryProperties) { (current, _) in current }
          }
          return entryInfo
     }
     
     func getAlbumArtFromDir(for url: URL) -> NSImage? {
          //
          // Try to load image from external file
          // search for coverfile
          //
          var image: NSImage?
          let coverFileName : String? = getCoverFileName(url)
          if coverFileName != nil {
               image = NSImage(byReferencingFile: coverFileName!)
          }
          return image
     }
     
     func getCoverFileName(_ url: URL) -> String? {
          let path = url.deletingLastPathComponent()
          do {
               let fileNames = try FileManager.default.contentsOfDirectory(atPath: path.path)
               let imageFileNames = (fileNames as NSArray).pathsMatchingExtensions(imageExtensions)
               for coverFile in imageFileNames {
                    let coverURL = URL(fileURLWithPath: coverFile, relativeTo: path).deletingPathExtension().lastPathComponent.lowercased()
                    //
                    // we want fileNames that contain one of our coverNames()
                    //
                    for coverName in PlaylistLoader.coverNames() {
                         if coverURL.contains(coverName) {
                              return path.appendingPathComponent(coverFile).path
                         }
                    }
               }
          }
          catch {
          }
          return nil
     }
     
     var imageExtensions : [String] {
          return [ "bmp", "jpeg", "jpg", "pdf", "png", "gif", "tiff", "jpeg-2000"]
     }
     
     class func coverNames() -> [String] {
          return ["cover", "front", "folder", "album", ]
     }
     
     func clear(_ sender: Any) {
          playlistController.clear(sender)
     }
     
     func addURLs(_ urls: [URL], sort: Bool) -> [PlaylistEntry]? {
          return insertURLs(urls , at: (playlistController.content as! NSArray).count  , sort: sort)
     }
     
     func add(_ url: URL!) -> [PlaylistEntry]? {
          return insertURLs([url!], at:  (playlistController.content as! NSArray).count , sort: false)
     }
     
     func willInsertURLs(_ urls: [URL], origin: URLOrigin) {
          playlistController.willInsertURLs(urls  , origin: origin)
     }
     
     func didInsertURLs(_ urls: [Any], origin: URLOrigin) {
          playlistController.didInsertURLs(urls  , origin: origin)
     }
     
     func startSpinner(_ infoText: String) {
          DispatchQueue.main.async {
               self.spinner.usesThreadedAnimation = true
               self.spinner.isHidden = false
               self.spinner.startAnimation(self)
               self.spinnertext.isHidden = false
               self.spinnertext.stringValue = infoText
          }
     }
     
     func stopSpinner() {
          spinner.stopAnimation(self)
          spinner.isHidden = true
          spinnertext.isHidden = true
     }
     
     func propertiesForURL(url: URL) -> [String : Any]? {
          if !(PhonixFileManager.fileExists(atPath: url.path)) || !url.isFileURL { return [:] }
          var dict: [String : Any] = [:]
          var sourceFormat = AudioStreamBasicDescription()
          var totalFrames: Int64 = 0
          //print("PlL_propertiesForURL:", url, "extension:", url.pathExtension.lowercased() )
          if !NonCoreAudioTypes().contains(url.pathExtension.lowercased()) {
               //
               // Open the input file
               //
               do {
                    let audioFile = try AVAudioFile.init(forReading: url)
                    sourceFormat = audioFile.fileFormat.streamDescription.pointee
                    totalFrames = audioFile.length
               } catch {
                    //print("PlL_propertiesForURL: AV_Catch:", error, "url:", url)
               }
          }
          if totalFrames == 0 {
               //print("PlL_propertiesForURL: ERROR! no frames for:", url)
               //
               // some files are not read by AVAudioFile: f.i. a certain type of flac (probably an old one)
               // SFB can give this length, but then we would need to wrap it.
               // we need bitdepth, sampleRate, channels and length
               //
               //let urlstring = url.path.cString(using: String.Encoding.utf8)
               let properties = SFBpropertiesForURL(url as CFURL)
               dict["channels"]        = properties.channels
               dict["bitsPerSample"]   = properties.bitspersample
               dict["sampleRate"]      = properties.samplerate
               dict["totalFrames"]     = properties.totalframes
               if properties.fileformat != nil {
                    dict["fileFormat"] = String(cString: properties.fileformat).lowercased()
               }
               else {
                    dict["fileFormat"] = url.pathExtension.lowercased()
               }
               dict["seekable"] = true
               return dict
          }
          
          if sourceFormat.mBitsPerChannel == 0
          {
               switch sourceFormat.mFormatFlags
               {
               case kAppleLosslessFormatFlag_16BitSourceData: sourceFormat.mBitsPerChannel = 16
               case kAppleLosslessFormatFlag_20BitSourceData: sourceFormat.mBitsPerChannel = 20
               case kAppleLosslessFormatFlag_24BitSourceData: sourceFormat.mBitsPerChannel = 24
               case kAppleLosslessFormatFlag_32BitSourceData: sourceFormat.mBitsPerChannel = 32
               default: sourceFormat.mBitsPerChannel = 16
               }
               sourceFormat.mBytesPerFrame  = sourceFormat.mChannelsPerFrame * sourceFormat.mBitsPerChannel / 8
               sourceFormat.mBytesPerPacket = sourceFormat.mBytesPerFrame    * sourceFormat.mFramesPerPacket
               
               if (sourceFormat.mSampleRate        == 0.0  ) { sourceFormat.mSampleRate = 44100.0 }
               if (sourceFormat.mChannelsPerFrame  == 0    ) { sourceFormat.mChannelsPerFrame = 2 }
          }
          
          dict["fileFormat"]      = url.pathExtension.lowercased()
          dict["channels"]        = sourceFormat.mChannelsPerFrame
          dict["bitsPerSample"]   = sourceFormat.mBitsPerChannel
          dict["sampleRate"]      = Float(sourceFormat.mSampleRate)
          dict["totalFrames"]     = totalFrames
          dict["seekable"]        = true
          return dict
     }
     
     func cueURLs(forContainerURL url: URL) -> [URL] {
          // called from: containerNode updatePath
          // we need the actual cue sheet here.
          //
          if !url.isFileURL        { return [] }
          if url.fragment != nil   { return [url] }
          // input url already has fragment defined - no need to expand further
          //
          // parse cueSheet to cueTrack
          // all we need to know : track count.
          // we just need to add #01 #02 etc to the extension.
          //
          var trackURLs: [URL] = []
          let cuesheet = CueSheet.init(cueFile: url)
          
          for track in cuesheet.tracks {
               let track$ = url.absoluteString.appendingFormat("#%02i", track.track!)
               let trackURL = URL(string:track$)
               trackURLs.append(trackURL!)
          }
          return trackURLs
     }
     
     func cueFileInfo(for url : URL) -> [String : Any]? {
          if !(PhonixFileManager.fileExists(atPath: url.path)) || !url.isFileURL { return nil }
          
          var metaDict: [String : Any] = [:]
          let cuesheet = CueSheet.init(cueFile: url)
          let e: NSEnumerator = (cuesheet.tracks as NSArray).objectEnumerator()
          while let track: CueSheetTrack = e.nextObject() as? CueSheetTrack
          {
               guard let cueTrackNr = Int(url.fragment!) else { return nil }
               if (cueTrackNr == track.track)
               {
                    metaDict = [
                         "fileFormat"   : track.url.pathExtension.lowercased(),
                         "URLCue"       : track.url    as Any,
                         "track"        : track.track  as Any,
                         "startCue"     : track.begin  as Any,
                         "endCue"       : track.end    as Any,
                         "artist"       : track.artist as Any,
                         "title"        : track.title  as Any,
                         "album"        : track.album  as Any,
                         "genre"        : track.genre  as Any,
                         "year"         : track.year   as Any]
                    
                    if let dict = propertiesForURL(url: track.url) {
                         metaDict.merge(dict) { (current, _) in current }
                    }
               }
          }
          return metaDict
     }
     
     func fileURLs(forContainerURL url: URL) -> [URL]? {
          //print("PlL fileURLs(forContainerURL)", url)
          // fix for m3u files without file:// or http:// (would crash Phonix)
          if !acceptableSchemes().contains(url.scheme!) { return [] }
          var contents: String?
          var urls : [URL] = []
          var separator = "\r\n"
          
          do {
               contents = try String(contentsOf: url, encoding:.utf8)
          } catch {
               //print("PlL fileURLs: catch#1", error)
               do {
                    contents = try String(contentsOf: url, encoding: .windowsCP1251)
               } catch {
                    //print("PlL fileURLs: catch#2", error)
                    do {
                         contents = try String(contentsOf: url, encoding: .isoLatin1)
                    } catch {
                         //print("PlL fileURLs: catch#3", error)
                    }
               }
          }
          if contents != nil { if !contents!.contains(separator) { separator = "\n" } }
          
          if let entries = contents?.components(separatedBy: separator) as [String]? {
               for entry in entries {
                    if entry.hasPrefix("#") || entry == "" { continue }
                    
                    if let urlLocal = urlForPath(for: entry, relativeTo: url) { urls.append(urlLocal) }
               }
          }
          return urls
     }
     
     func httpURLs(forContainerURL url: URL) -> [URL]? {
          // fix for m3u files without file:// or http:// (would crash Phonix)
          //          if !acceptableStreamTypes().contains(url.scheme!) { return [] }
          var contents  : String?
          var urls      : [URL] = []
          var separator = "\r\n"
          
          do {
               contents = try String(contentsOf: url, encoding:.utf8)
               //print("PlL httpURLs: try #1: url:", url, "response:\n", contents! )
          } catch {
               //print("PlL httpURLs:", error)
               do {
                    contents = try String(contentsOf: url, encoding: .windowsCP1251)
                    //print("PlL httpURLs: try #1: url:", url, "response:\n", contents! )
               } catch {
                    print("PlL httpURLs:", error)
                    return []
               }
          }
          
          if contents != nil {
               //
               // some severs respond with a 404 page...
               //
               if contents!.lowercased().hasPrefix("<!doctype") { return [] }
               if !contents!.contains(separator) { separator = "\n" }
               
               if url.pathExtension.lowercased().contains("m3u") {
                    if contents!.contains("#EXTM3U") {
                         if let entries = MediaItem.parseM3U(contents: contents!) {
                              for entry in entries {
                                   guard let url$ = entry.urlString else { continue }
                                   if let entryURL = URL(string: url$) { urls.append(entryURL) }
                              }
                         }
                    } else {
                         if let entries = contents?.components(separatedBy: separator) as [String]? {
                              for entry in entries {
                                   if entry.hasPrefix("#") || (entry == "") { continue }
                                   if let url = URL(string: entry) { urls.append(url) }
                              }
                         }
                    }
               } else if url.pathExtension.contains("pls") {
                    if contents!.lowercased().hasPrefix("[playlist]") {
                         if let entries = MediaItem.parsePLS(contents: contents!) {
                              for entry in entries {
                                   guard let url$ = entry.urlString else { continue }
                                   if let entryURL = URL(string: url$) { urls.append(entryURL) }
                              }
                         }
                    }
               } else if url.pathExtension.contains("asx") {
                    if contents!.lowercased().hasPrefix("<asx") {
                         if let entries = MediaItem.parseASX(contents: contents!) {
                              for entry in entries {
                                   guard let url$ = entry.urlString else { continue }
                                   if let entryURL = URL(string: url$) { urls.append(entryURL) }
                              }
                         }
                    }
               } else {
                    //
                    // should not be triggered. But as a last resort.
                    //
                    print("PlL THIS SHOULD NOT BE PRINTED!, httpContainer did not Parse, last resort!")
                    let detector = try! NSDataDetector(types: NSTextCheckingResult.CheckingType.link.rawValue)
                    let matches = detector.matches(in: contents!, options: [], range: NSRange(location: 0, length: contents!.utf8.count))
                    
                    for match in matches {
                         guard let range = Range(match.range, in: contents!) else { continue }
                         let url$ = String(contents![range])
                         if url$.contains("http") {
                              if let entryURL = URL(string: url$) { urls.append(entryURL) }
                         }
                    }
               }
          }
          return urls
     }
     
     func urlForPath(for path: String, relativeTo baseFilename: URL) -> URL? {
          if path.contains("://") {
               let returURL = URL(string: path)
               return returURL
          }
          let basePath : URL = baseFilename.deletingLastPathComponent()
          let url = URL(fileURLWithPath:basePath.relativePath.appending("/").appending(path))
          return url
     }
}
