import AppKit

class PlaylistView: NSTableView {
    
    @IBOutlet var playbackController: PlaybackController!
    @IBOutlet var playlistController: PlaylistController!
    @IBOutlet var mainWindow: MainWindow!
    
    var headerContextMenu : NSMenu?
    var columnClicked     : Int?
    
    override func awakeFromNib() {
        menu?.autoenablesItems = false
        autosaveName = NSTableView.AutosaveName("PlaylistHeaderView")
        autosaveTableColumns = true
        // Configure bindings to scale font size and row height
        let ctlSize: NSControl.ControlSize = ControlSize(rawValue: UInt(UserDefaults.standard.float(forKey: "fontSize")))!
        let listFont: NSFont = NSFont.systemFont(ofSize: NSFont.systemFontSize(for: ctlSize))
        for column: NSTableColumn in tableColumns {
            (column.dataCell as! NSCell).controlSize = ctlSize
            (column.dataCell as! NSCell).font = listFont
        }
        // Set up formatters
        let secondsFormatter = SecondsFormatter()
        let lengthcolumn = (tableColumn(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "length")))
        let lengthCell: NSCell = lengthcolumn!.dataCell as! NSCell
        lengthCell.formatter = secondsFormatter
        
        let indexFormatter = IndexFormatter()
        let indexcolumn = (tableColumn(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "index")))
        let indexCell: NSCell = indexcolumn!.dataCell as! NSCell
        indexCell.formatter = indexFormatter
        // end setting up formatters
        verticalMotionCanBeginDrag = true
        headerContextMenu = NSMenu(title: "Playlist Header Context Menu")
        var visibleTableColumns: Int = 0
        var menuIndex: Int = 0
        
        let sortDescriptor = NSSortDescriptor(key: "identifier", ascending: true)
        let sortDescriptors = [sortDescriptor]
        
        let aParagraphStyle  = NSMutableParagraphStyle()
        aParagraphStyle.lineBreakMode = .byTruncatingTail
        aParagraphStyle.alignment     = .left
        
        let titleAttributes = [ NSAttributedString.Key.font: listFont,
                                NSAttributedString.Key.paragraphStyle: aParagraphStyle,
                                NSAttributedString.Key.foregroundColor: TextColor.withAlphaComponent(0.7)]
        
        for column: NSTableColumn in ((tableColumns as NSArray).sortedArray(using: sortDescriptors) as? [NSTableColumn])!
        {
            var title = ""
            var attributedTitle : NSMutableAttributedString
            
            if column.identifier.rawValue == "status" {
                title = ">"
            } else if column.identifier.rawValue == "index" {
                //
                // index column text can be a bit darker, this is not part of track info
                //
                title = "#"
                ((column.dataCell) as! NSTextFieldCell).textColor = TextColor.withAlphaComponent(0.5)
            } else {
                title = column.headerCell.title
                ((column.dataCell) as! NSTextFieldCell).textColor = TextColor
            }
            attributedTitle = NSMutableAttributedString(string: title)
            attributedTitle.addAttributes(titleAttributes, range: NSRange(location: 0, length: attributedTitle.length))
            column.headerCell.attributedStringValue = attributedTitle
            
            let contextMenuItem: NSMenuItem? = headerContextMenu?.insertItem(withTitle: title,
                                                                             action: #selector(toggleColumn(_:)),
                                                                             keyEquivalent: "",
                                                                             at: menuIndex)
            contextMenuItem?.target = self
            contextMenuItem?.representedObject = column
            contextMenuItem?.state = column.isHidden ? .off : .on
            visibleTableColumns += Int(truncating: NSNumber(value: !column.isHidden))
            menuIndex += 1
        }
        if visibleTableColumns == 0 {
            for column: NSTableColumn in tableColumns {
                column.isHidden = false
            }
        }
        headerView?.menu = headerContextMenu
    }
    
    override func drawRow(_ row: Int, clipRect: NSRect) {
        super.drawRow(row, clipRect: clipRect)
        
        if UserDefaults.standard.bool(forKey: "drawAlbumSeparator") {
            var drawSeparator: Bool = false
            let listCount = (playlistController.arrangedObjects as! NSArray).count
            if listCount > 0 && row < listCount - 1 {
                let item = (playlistController.arrangedObjects as! NSArray).object(at: row) as? PlaylistEntry
                let nextRow = row + 1
                guard let nextitem = (playlistController?.arrangedObjects as! NSArray).object(at: nextRow) as? PlaylistEntry else { return }
                if item!.album != nextitem.album
                    && item!.isSeekable
                    || item!.fileinfo() != nextitem.fileinfo() {
                    drawSeparator = true
                }
            }
            if drawSeparator {
                let rowRect: NSRect = rect(ofRow: row)
                let gridPath = NSBezierPath()
                gridPath.move(to: rowRect.origin)
                gridPath.move(to: NSMakePoint(rowRect.origin.x, rowRect.origin.y + rowRect.size.height))
                gridPath.line(to: NSMakePoint(rowRect.origin.x + rowRect.size.width, rowRect.origin.y + rowRect.size.height))
                MeterScaleColor.blended(withFraction: 0.25, of: NSColor.clear)!.set()
                let linePattern: [CGFloat] = [10.0]
                gridPath.setLineDash(linePattern, count: 0, phase: 0.0)
                gridPath.stroke()
            }
        }
    }
    
    override var acceptsFirstResponder: Bool {
        return true
    }
    
    override func resignFirstResponder() -> Bool {
        return true
    }
    
    override func acceptsFirstMouse(for event: NSEvent?) -> Bool {
        return false
    }
    
    override func mouseDown(with event: NSEvent) {
        super.mouseDown(with: event)
        if event.type == .leftMouseDown && event.clickCount == 2 && selectedRowIndexes.count == 1 {
            playbackController.play(self)
        }
    }
    
    @IBAction func toggleColumn(_ sender: NSButton) {
        let tc = (sender as AnyObject).representedObject
        if sender.state == .off {
            sender.state = .on
            (tc as! NSTableColumn).isHidden = false
        } else {
            sender.state = .off
            (tc as! NSTableColumn).isHidden = true
        }
    }
    
    override func menu(for event: NSEvent) -> NSMenu? {
        // Find which row is under the cursor
        //print("PLV_menu(for event:", event as Any)
        window?.makeFirstResponder(self)
        let menuPoint = (convert(event.locationInWindow, from: nil))
        let iRow: Int = row(at: menuPoint)
        let tableViewMenu: NSMenu? = menu
        let localLocation = (convert(event.locationInWindow, from: nil))
        columnClicked = column(at: localLocation)
        //
        /* Update the table selection before showing menu
         Preserves the selection if the row under the mouse is selected (to allow for
         multiple items to be selected), otherwise selects the row under the mouse */
        let currentRowIsSelected: Bool = selectedRowIndexes.contains(iRow)
        if !currentRowIsSelected {
            if iRow == -1 {
                deselectAll(self)
            } else {
                selectRowIndexes(NSIndexSet(index: iRow) as IndexSet, byExtendingSelection: false)
            }
        }
        if numberOfSelectedRows <= 0 {
            //
            // No rows are selected, so the table should be displayed with all items disabled
            //
            for i in 0 ..< (tableViewMenu!.numberOfItems) {
                tableViewMenu?.item(at: i)?.isEnabled = false
            }
        }
        let editmetadata    : NSMenuItem? = tableViewMenu?.item(withTitle: "Edit metadata")
        let showinfinder    : NSMenuItem? = tableViewMenu?.item(withTitle: "Show in Finder")
        
        if  playlistController.entry(at: selectedRow) == nil { return nil}
        
        var enabled: Bool = false
        
        if      playlistController.entry(at: selectedRow)!.isSeekable   == false
            ||  playlistController.entry(at: selectedRow)!.isError      == true
            ||  playlistController.entry(at: selectedRow)!.URLCue       != nil
        {
            enabled = false
        }
        else {
            enabled = true
        }
        //print("PLV_menu:() metaOptions:", enabled)
        editmetadata?   .isEnabled = enabled
        showinfinder?   .isEnabled = enabled
        
        return tableViewMenu
    }

    
    override func keyDown(with event: NSEvent) {
        let commandkey = NSEvent.ModifierFlags.command.rawValue
        let shiftkey = NSEvent.ModifierFlags.shift.rawValue
        let controlkey = NSEvent.ModifierFlags.control.rawValue
        let optionkey = NSEvent.ModifierFlags.option.rawValue
        let modifiers = UInt(event.modifierFlags.rawValue & (commandkey | shiftkey | controlkey | optionkey))
        let characters = event.characters! as NSString
        var c: unichar
        if characters.length != 1 {
            super.keyDown(with: event)
            return
        }
        c = (characters.character(at: 0))
        if modifiers == 0 && (c == NSDeleteCharacter || c == NSBackspaceCharacter || c == NSDeleteFunctionKey) {
            playlistController.removeitem(self)
        } else if modifiers == 0 && UnicodeScalar(c) == " " {
            playbackController.playPauseResume(self)
        } else if modifiers == 0 && (c == NSEnterCharacter || c == NSCarriageReturnCharacter) {
            playbackController.play(self)
        } else if modifiers == 0 && c == 0x1B {
            playlistController.clearFilterPredicate(self)
        } else {
            super.keyDown(with: event)
        }
    }
    
    @IBAction func scroll(toCurrentEntry sender: Any) {
        mainWindow.selectPlayingTab()
        scrollRowToVisible((playlistController.currentEntry?.peindex)!)
        selectRowIndexes(NSIndexSet(index: (playlistController.currentEntry?.peindex)!) as IndexSet, byExtendingSelection: false)
    }
    
    override func validateUserInterfaceItem(_ item: NSValidatedUserInterfaceItem) -> Bool {
        let action: Selector? = item.action
        //
        // scroll to currentEntry and play, TOFO: we need to check if we ar still in the same tab,
        //
        if      action == #selector(scroll(toCurrentEntry:))
            && (playbackController.playbackStatus == PhonixStatus.phonixStopped.rawValue)
        {
            return false
        }
        return super.validateUserInterfaceItem(item)
    }
    
    override func drawBackground(inClipRect clipRect: NSRect) {
        super.drawBackground(inClipRect: clipRect)
        guard usesAlternatingRowBackgroundColors else { return }
        let rectHeight = rowHeight + intercellSpacing.height
        let row = rows(in: clipRect).location
        drawAlternatingBackground(inClipRect: clipRect, height: rectHeight, rows: row)
    }
}

class PlaylistHeaderView: NSTableHeaderView {
    override func draw(_ dirtyRect: NSRect) {
        layer!.backgroundColor = BackgroundColorA.cgColor
        super.draw(dirtyRect)
    }
    
    override func mouseDown(with event: NSEvent) {
        //
        // left Mouse on PlayListHeader
        //
        let event_location = event.locationInWindow
        let local_point = (convert(event_location, from: nil))
        var column: Int = self.column(at: local_point)
        if event.clickCount == 2 && column != -1 {
            // compute whether the clickpoint is a column separator or not
            var clickedSeperator = false
            // handle a click one pixel away at right
            let rect: NSRect = headerRect(ofColumn: column)
            if abs(rect.origin.x - local_point.x) <= 1 && column > 0 {
                column = column - column
                clickedSeperator = true
            } else if abs(rect.origin.x + rect.size.width - local_point.x) <= 3 {
                clickedSeperator = true
            }
            if clickedSeperator {
                let col: NSTableColumn = (tableView?.tableColumns[column])!
                // Info about the font and such
                let cell = col.dataCell
                // Binding info...reaching deep!
                let bindingInfo = col.infoForBinding(NSBindingName("value"))
                let boundArray = (bindingInfo![.observedObject] as AnyObject).value(forKeyPath: bindingInfo![.observedKeyPath] as? String ?? "") as? [Any]
                // find the longest string display length in that column
                var max_width: Float = -1
                for row: Any in boundArray! {
                    (cell as AnyObject).setObjectValue(row as? NSCopying)
                    let width: Float = Float((cell as AnyObject).cellSize.width)
                    if width > max_width {
                        max_width = width
                    }
                }
                // set the new width (plus a 5 pixel extra to avoid "..." string substitution) //
                col.width = CGFloat(max_width)
            } else {
                super.mouseDown(with: event)
            }
        } else {
            super.mouseDown(with: event)
        }
    }
}
