protocol CuteRadioSearchDelegate {
    func searchInProgress(_ state: Bool)
    func genresLoaded(_ genres: [String])
    func countriesLoaded(_ countries: [String])
    func reloadData()
}

class CuteradioSearch {
    private let apiURL                  : String = "https://marxoft.co.uk/api/cuteradio"
    private let searchAllUrl            : String = "https://marxoft.co.uk/api/cuteradio/stations?limit=50&sort=title&sortDescending=false&approved=1"
    private let searchGenreUrl          : String = "https://marxoft.co.uk/api/cuteradio/stations?limit=50&sort=title&sortDescending=false&genre=%@&approved=1"
    private let searchCountryUrl        : String = "https://marxoft.co.uk/api/cuteradio/stations?limit=50&sort=title&sortDescending=false&country=%@&approved=1"
    private let searchCountryAndGenreUrl: String = "https://marxoft.co.uk/api/cuteradio/stations?limit=50&sort=title&sortDescending=false&country=%@&genre=%@&approved=1"
    private let nextSearchUrl           : String = "https://marxoft.co.uk/api/cuteradio%@"
    private let genreRequest            : String = "https://marxoft.co.uk/api/cuteradio/genres?limit=50&sort=name&sortDescending=false"
    private let countryRequest          : String = "https://marxoft.co.uk/api/cuteradio/countries?limit=50&sort=name&sortDescending=false"
    
    private var stations    : [CuteradioRadioStation]
    private var genres      : [String]
    private var countries   : [String]
    private var currentUrl  : String
    private var nextUrl     : String
    private var appendResult: Bool
    
    var delegate : CuteRadioSearchDelegate?
    
    init() {
        stations  = []
        genres    = []
        countries = []
        currentUrl = String()
        nextUrl  = String()
        appendResult = false
    }
    
    func stationsCount() -> Int {
        return stations.count
    }
    
    subscript(index: Int) -> CuteradioRadioStation? {
        if index >= stations.count { return nil }
        return stations[index]
    }
    
    func getGenreList() {
        appendResult = false
        let url = URL(string: genreRequest)
        currentUrl = url!.absoluteString.replacingOccurrences(of: apiURL, with: "")
        getCuteGenres(url!)
    }
    
    func getCountryList() {
        appendResult = false
        let url = URL(string: countryRequest)
        currentUrl = url!.absoluteString.replacingOccurrences(of: apiURL, with: "")
        getCuteCountries(url!)
    }
    
    func searchBy(genre: String, country: String) {
        appendResult = false
        nextUrl = String()
        var url: URL?
        if genre.isEmpty && country.isEmpty {
            url = URL(string: String(format: searchAllUrl))
        } else if genre.isEmpty {
            let encodedCountry: String = country.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
            url = URL(string: String(format: searchCountryUrl, encodedCountry))
        } else if country.isEmpty {
            let encodedGenre: String = genre.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
            url = URL(string: String(format: searchGenreUrl, encodedGenre))
        } else {
            let encodedGenre  : String = genre.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
            let encodedCountry: String = country.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
            url = URL(string: String(format: searchCountryAndGenreUrl, encodedCountry, encodedGenre))
        }
        currentUrl = url!.absoluteString.replacingOccurrences(of: apiURL, with: "")
        getCuteStations(url!)
        delegate?.searchInProgress(true)
    }
    
    func stopSearch() -> Bool {
        var isSearching = false
        if nextUrl != "" {
            currentUrl = String()
            isSearching = true
        }
        delegate?.searchInProgress(false)
        return isSearching
    }
    
    private func nextGenreSearch() {
        if nextUrl.isEmpty { return }
        
        appendResult = true
        let url = URL(string: String(format: nextSearchUrl, nextUrl))
        getCuteGenres(url!)
    }
    
    private func nextCountrySearch() {
        if nextUrl.isEmpty { return }
        
        appendResult = true
        let url = URL(string: String(format: nextSearchUrl, nextUrl))
        getCuteCountries(url!)
    }
    
    private func nextStationSearch() {
        if nextUrl.isEmpty { return }
        
        appendResult = true
        let url = URL(string: String(format: nextSearchUrl, nextUrl))
        getCuteStations(url!)
    }
    
    private func getCuteStations(_ url: URL) {
        URLSession.shared.dataTask(with: url, completionHandler:
            { (data, response, error) -> Void in self.stationsCompleted(data: data, response: response, error: error) }
            ).resume()
    }
    
    private func getCuteGenres(_ url: URL) {
        URLSession.shared.dataTask(with: url, completionHandler:
            { (data, response, error) -> Void in self.genresCompleted(data: data, response: response, error: error) }
            ).resume()
    }
    
    private func getCuteCountries(_ url: URL) {
        URLSession.shared.dataTask(with: url, completionHandler:
            { (data, response, error) -> Void in self.countriesCompleted(data: data, response: response, error: error) }
            ).resume()
    }
    
    private func stationsCompleted(data: Data?, response: URLResponse?, error: Error?) {
        if data == nil || error != nil { return }
        if let jsonObj = try? JSONSerialization.jsonObject(with: data!, options: .allowFragments) as? [String: AnyHashable]
        {
            var temp_stations: [CuteradioRadioStation] = []
            
            if let items = jsonObj["items"] as? [[String : AnyHashable]] {
                
                for item in items {
                    //print("CRS stationsCompleted:", item as Any)
                    let station: CuteradioRadioStation = CuteradioRadioStation()
                    if let url      = item["source"]    as? String {
                        station.url = url.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)! }
                    if let title    = item["title"]     as? String { station.title   = title    }
                    if let genre    = item["genre"]     as? String { station.genre   = genre    }
                    if let country  = item["country"]   as? String { station.country = country  }
                    let prefixes = ["http://", "mms://", "mmsh://", "rtsp://", "https://"]
                    //print("CRS stationsCompleted: id:", station.id as Any )
                    if station.url!.hasSuffix(".asf") || station.url!.hasSuffix(".asp") { continue }
                    
                    for prefix in prefixes {
                        if station.url!.lowercased().hasPrefix(prefix) {
                            //print("CRS stationsCompleted: url:", station.url as Any )
                            temp_stations.append(station)
                            break
                        }
                    }
                }
            }
            if !appendResult {
                stations.removeAll()
            }
            for item in temp_stations {
                stations.append(item)
            }
            
            if let nextURL = jsonObj["next"] as? String {
                nextUrl = nextURL.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
                
                let componentsNext = nextUrl.components(separatedBy: "&offset")
                let componentsCurr = currentUrl.components(separatedBy: "&offset")
                if componentsNext[0] == componentsCurr[0] {
                    nextStationSearch()
                } else {
                    if !currentUrl.isEmpty { stations.removeAll() }
                    for item in temp_stations {
                        stations.append(item)
                    }
                    nextUrl = String()
                }
            } else {
                DispatchQueue.main.async {
                    self.delegate?.searchInProgress(false)
                }
                nextUrl = String()
            }
        } else {
            stations.removeAll()
        }
        
        DispatchQueue.main.async {
            self.delegate?.reloadData()
        }
    }
    
    private func genresCompleted(data: Data?, response: URLResponse?, error: Error?) {
        if data == nil || error != nil { return }
        if let jsonObj = try? JSONSerialization.jsonObject(with: data!, options: .allowFragments) as? [String: AnyHashable]
        {
            var temp_genres: [String] = []
            if let items = jsonObj["items"] as? [[String : AnyHashable]] {
                
                for item in items {
                    
                    if  let genre = item["name"]  as? String,
                        let count = item["count"] as? Int   {
                        
                        if genre.lengthOfBytes(using: .utf8) > 1
                            && count > 1 {
                            
                            if genre.lengthOfBytes(using: .utf8) > 2 {
                                if      !genre.contains("'")
                                    &&   genre != "All"
                                    &&  !genre.contains("http:") {
                                    temp_genres.append(genre)
                                }
                            }
                        }
                    }
                }
            }
            if !appendResult {
                genres.removeAll()
            }
            for item in temp_genres {
                genres.append(item)
            }
            
            if let nextURL = jsonObj["next"] as? String {
                nextUrl = nextURL.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
                let componentsNext = nextUrl.components(separatedBy: "&offset")
                let componentsCurr = currentUrl.components(separatedBy: "&offset")
                
                if componentsNext[0] == componentsCurr[0] {
                    nextGenreSearch()
                } else {
                    if !currentUrl.isEmpty { genres.removeAll() }
                    for item in temp_genres {
                        genres.append(item)
                    }
                    nextUrl = String()
                }
            } else {
                DispatchQueue.main.async {
                    self.delegate?.searchInProgress(false)
                    self.delegate?.genresLoaded(self.genres)
                }
                nextUrl = String()
            }
        } else {
            genres.removeAll()
        }
    }
    
    private func countriesCompleted(data: Data?, response: URLResponse?, error: Error?) {
        if data == nil || error != nil { return }
        if let jsonObj = try? JSONSerialization.jsonObject(with: data!, options: .allowFragments) as? [String: AnyHashable]
        {
            var temp_countries: [String] = []
            
            if let items = jsonObj["items"] as? [[String : AnyHashable]] {
                
                for item in items {
                    
                    if  let country = item["name"] as? String,
                        let count   = item["count"] as? Int {
                        
                        if count > 1 && country.lengthOfBytes(using: .utf8) > 1  {
                            
                            if      !country.contains("'")
                                &&  country != "No"
                                &&  country != "De"
                                &&  !country.hasSuffix(".")
                                &&  !country.hasPrefix(" ")
                            {
                                temp_countries.append(country)
                            }
                        }
                    }
                }
            }
            if !appendResult {
                countries.removeAll()
            }
            for item in temp_countries {
                countries.append(item)
            }
            
            if let nextURL = jsonObj["next"] as? String {
                nextUrl = nextURL.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
                let componentsNext = nextUrl.components(separatedBy: "&offset")
                let componentsCurr = currentUrl.components(separatedBy: "&offset")
                
                if componentsNext[0] == componentsCurr[0] {
                    nextCountrySearch()
                } else {
                    if !currentUrl.isEmpty { countries.removeAll() }
                    for item in temp_countries {
                        countries.append(item)
                    }
                    nextUrl = String()
                }
            } else {
                DispatchQueue.main.async {
                    self.delegate?.searchInProgress(false)
                    self.delegate?.countriesLoaded(self.countries)
                }
                nextUrl = String()
            }
        } else {
            countries.removeAll()
        }
    }
}
