import AppKit

class ShoutCastController: NSController, NSTableViewDataSource, NSTableViewDelegate, ShoutcastSearchDelegate {
    
    @IBOutlet var playlistLoader               : PlaylistLoader!
    @IBOutlet var playbackController           : PlaybackController!
    @IBOutlet var shoutCastSearchPanel         : NSPanel!
    @IBOutlet var shoutCastSearchField         : NSTextField!
    @IBOutlet var bitRateSelector              : NSPopUpButton!
    @IBOutlet var shoutCastGenreSelector       : NSPopUpButton!
    @IBOutlet var shoutCastSearchTable         : NSTableView!
    @IBOutlet var shoutCastButton              : NSButton?
    
    private let getGenreListUrl : String = "https://api.shoutcast.com/legacy/genrelist?k=%@"
    
    var shoutCastSearch     : ShoutcastSearch!
    var shoutCastBitRates   = [String]()
    
    override func awakeFromNib() {
        initShoutCastSearch()
        shoutCastButton?.isEnabled = false
        if isConnectedToNetwork() {
        populateGenreSelector()
        }
        populateBitrateSelector()
    }
    
    func initShoutCastSearch() {
        guard let SHOUTcastAPIkey = NSDataAsset(name: "SHOUTcastAPIkey")?.data else { return }
        let key = String(data:SHOUTcastAPIkey, encoding: String.Encoding.utf8)!
        shoutCastSearch = ShoutcastSearch(apiKey: String(key.reversed()))
        shoutCastSearch.delegate = self
    }
    
    func populateGenreSelector() {
        let genreListurl = String(format: getGenreListUrl, shoutCastSearch.SHOUTcastAPIkey)
        var genres : [String] = ["All"]
        let url = URL(string: genreListurl)
        URLSession.shared.dataTask(with: url!, completionHandler:
            { (data, response, error) -> Void in
                if error != nil {
                }
                if let data = data {
                    let xml = XML.parse(data)
                    let genreList = xml["genrelist"]["genre"]
                    if genreList.all != nil {
                        for (_, node) in (genreList.all?.enumerated())! {
                            if let name = node.attributes["name"] {
                                genres.append(name)
                            }
                        }
                        DispatchQueue.main.async {
                            self.shoutCastGenreSelector.addItems(withTitles: genres)
                            self.shoutCastButton?.isEnabled = true
                        }
                    } else {
                        //                        print("SCC_Error: No genres found")
                    }
                }
        }).resume()
    }
    
    func populateBitrateSelector() {
        shoutCastBitRates = ["All", "320", "256", "192", "160", "128", "112", "96", "80", "64"]//, "56", "48", "32", "24"]
        bitRateSelector.addItems(withTitles: shoutCastBitRates)
    }
    
    @IBAction func openCloseSearchPanel(_ sender: AnyObject) {
        if shoutCastSearchPanel.isVisible {
            shoutCastSearchPanel.close()
        } else {
            shoutCastSearchPanel.orderFront(sender)
            shoutCastGenreAndBitRateFilter(sender)
        }
    }
    
    @IBAction func shoutCastGenreAndBitRateFilter(_ sender: AnyObject) {
        let bitrate = bitRateSelector.titleOfSelectedItem!
        let genre   = shoutCastGenreSelector.titleOfSelectedItem!.replacingOccurrences(of: " ", with: "+")
        shoutCastSearch.searchBy(genre: genre, bitrate: bitrate)
    }
    
    @IBAction func shoutCastSearch(_ sender: AnyObject) {
        let search: String = shoutCastSearchField.stringValue
        if !search.isEmpty {
            shoutCastSearch.searchStation(searchFor: search)
        }
    }
    
    @IBAction func addShoutcastStationToPlaylist(_ sender: AnyObject) {
        let selectedRow: Int = shoutCastSearchTable.selectedRow
        if selectedRow > -1 {
            var stationURLs = [URL]()
            if let radioStation = shoutCastSearch[selectedRow] {
                if radioStation.id != nil {
                    DispatchQueue.main.async {
                        stationURLs.append(URL(string: String(format: "https://yp.shoutcast.com/sbin/tunein-station.m3u?id=%@", radioStation.id!))!)
                        guard let insertedEntries = self.playlistLoader.addURLs(stationURLs, sort: false) else { return }
                        if !insertedEntries.isEmpty {
                            if let logo$ = radioStation.logo {
                                if let logoURL = URL(string: logo$) {
                                    if let logo = NSImage(contentsOf: logoURL) {
                                        for entry in insertedEntries {
                                            entry.albumArt = logo
                                        }
                                    }
                                }
                            }
                            _ = self.playbackController.play(entry: insertedEntries[0])
                        }
                    }
                }
            }
        }
    }
    //
    // TableView Delegate functions
    //
    func numberOfRows(in tableView: NSTableView) -> Int {
        return shoutCastSearch.stationsCount()
    }
    
    func tableView(_ tableView: NSTableView, objectValueFor tableColumn: NSTableColumn?, row: Int) -> Any? {
        
        if shoutCastSearch.stationsCount() > row {
            if let data = shoutCastSearch[row] {
                switch tableColumn?.identifier.rawValue {
                case "index"            : return row
                case "title"            : return data.title
                case "currentplaying"   : return data.currentlyPlaying
                case "genre"            : return data.genre
                case "bitrate"          : return data.bitrate
                default                 : return nil
                }
            }
        }
        return nil
    }
    //
    // ReloadDataAfterFetchDelegate Delegate functions
    //
    func reloadData() {
        DispatchQueue.main.async {
            self.shoutCastSearchTable.reloadData()
        }
    }
    func searchInProgress(_ state: Bool) {
        
    }
}
