import AppKit

class SpectrumAnalyzer: NSView {
    private var frequencyBands: Int = 31
    var frequencyData: UnsafeMutablePointer<Float>?
    
    let refLevel        : Float = 0.5
    let rangedB         : Float = -60.0
    let minIndactorSize : CGFloat = 1.0
    var barWidth        : CGFloat = 3.0
    let barSpacing      : CGFloat = 1.0
    let releaseSpeed    : CGFloat = 0.6
    var meterMuted      : Bool = false
    var bandHeights     = UnsafeMutablePointer<CGFloat>.allocate(capacity: 32)
    
    var bandCount : Int {
        set {
            if     newValue >= 1
                && newValue <= 32
                && newValue != frequencyBands
            {
                frequencyBands = newValue
                draw(frame)
            }
        }
        get {
            return frequencyBands
        }
    }
    
    override func draw(_ meterScale: CGRect) {
        drawBands(muted: meterMuted)
    }
    
    func setFrequencyData(freqData: UnsafeMutablePointer<Float>?) {
        if freqData != nil {
            frequencyData = freqData
            updateSpectralView()
        } else {
            clearSpectralView()
        }
    }
    
    func updateSpectralView() {
        let scaleHeight: CGFloat = frame.size.height
        var bandLvlSize: CGFloat
        var leveldB: Float = -140.0
        for band in 0 ..< frequencyBands {
            //
            //  frequencyData = (0.0 - 0.5)
            //  level range : -140dB - 0dB
            //  we want a visible range of -60dB
            //  scale should be dB linear.  22.0 pixels, 1 pixel is 3dB
            //
            //  we need frequencyData[x] where x = { 0 - 127 }
            //
            let freqF = exp2( 7.0 * Float(band) / Float(frequencyBands - 1) ) - 1.0
            let freq  = Int(round(freqF))
            
            if frequencyData != nil {
                leveldB = 20.0 * log10(frequencyData![freq] / refLevel)
            } else {
                leveldB = -140.0
            }
            //
            // limit de dB values to our desired range
            //
            leveldB = leveldB > rangedB ? leveldB : rangedB
            //
            // convert dB's to pixels
            //
            bandLvlSize = scaleHeight - scaleHeight * CGFloat(leveldB) / CGFloat(rangedB)
            //
            //  we want: - instant rise
            //           - constant dB/s level fall
            //
            if bandLvlSize > bandHeights[band] {
                //
                // rise
                //
                bandHeights[band] = bandLvlSize
            } else {
                //
                // release
                //
                bandHeights[band] = (bandHeights[band] - releaseSpeed <= bandLvlSize) ? bandLvlSize : (bandHeights[band] - releaseSpeed)
            }
        }
        needsDisplay = true
    }
    
    func drawBands(muted: Bool) {
        let muted = meterMuted
        for band in 0..<frequencyBands
        {
            if (    muted
                ||  bandHeights[band].isNaN
                ||  bandHeights[band] < minIndactorSize )
            {
                bandHeights[band] = minIndactorSize
            }
            let bandRect: CGRect = CGRect(x: CGFloat(band) * barWidth + barSpacing,
                                          y: 0.0,
                                          width: barWidth - barSpacing,
                                          height: bandHeights[band])
            MeterBarColor.setFill()
            bandRect.fill()
        }
        meterMuted = false
    }
    
    func clearSpectralView() {
        meterMuted = true
        layer?.backgroundColor = MeterBackgroundColor.cgColor
        needsDisplay = true
    }
}
