import AppKit

class OutputsArrayController: NSArrayController, NSTableViewDelegate, NSTableViewDataSource {
    
    @IBOutlet var sampleRateFixOrMediaBtn   : NSButton!
    @IBOutlet var sampleRateDisplay         : NSTextField!
    @IBOutlet var sampleRatesView           : NSView!
    @IBOutlet var sampleRatesTableView      : NSTableView!

    var deviceSampleRates: [Float64]?

    override func awakeFromNib() {
        getOutputDevices()
        
        // init fixed/media samplerate button
        setFixedOrMediaSampleRate(fixed: NSUserDefaultsController.shared.defaults.bool(forKey: "sampleRateFixed"))
    }

    func getOutputDevices() {
        if let anObjects = arrangedObjects as? [Any] {
            remove(contentsOf: anObjects)
        }
        selectsInsertedObjects = false
        //
        var propertyAddress: AudioObjectPropertyAddress = AudioObjectPropertyAddress()
        var deviceIDs = [AudioObjectID]()
        var propertySize: UInt32 = UInt32(MemoryLayout<AudioObjectID>.size)
        var numDevices: Int!
        //
        // get stored default device,
        //
        let selectedOutputDevice = NSUserDefaultsController.shared.defaults.object(forKey: "outputDevice") as? NSDictionary
        
        propertyAddress.mScope      = kAudioObjectPropertyScopeOutput
        propertyAddress.mElement    = kAudioObjectPropertyElementMaster
        propertyAddress.mSelector   = kAudioHardwarePropertyDevices
        
        if noErr != AudioObjectGetPropertyDataSize(AudioObjectID(kAudioObjectSystemObject), &propertyAddress, 0, nil, &propertySize)
        { return }
        
        numDevices = Int(propertySize) / MemoryLayout<AudioDeviceID>.size
        deviceIDs  = [AudioObjectID](repeating: 0, count: numDevices)
        
        if noErr != AudioObjectGetPropertyData(AudioObjectID(kAudioObjectSystemObject), &propertyAddress, 0, nil, &propertySize, &deviceIDs)
        { return }
        
        for idx in 0 ..< numDevices {
            //
            // DeviceName
            //
            propertyAddress.mSelector = kAudioDevicePropertyDeviceName
            
            if noErr != AudioObjectGetPropertyDataSize(deviceIDs[idx], &propertyAddress, 0, nil, &propertySize)
            { break }
            var deviceName = [Int8](repeating: Int8(), count: Int(propertySize))
            
            if noErr != AudioObjectGetPropertyData(deviceIDs[idx], &propertyAddress, 0, nil, &propertySize, &deviceName)
            { break }
            //
            // DeviceUID
            //
            var deviceUID: CFString = "" as CFString
            
            propertyAddress.mSelector = kAudioDevicePropertyDeviceUID
            
            if noErr != AudioObjectGetPropertyDataSize(deviceIDs[idx], &propertyAddress, 0, nil, &propertySize)
            { break }
            
            if noErr != AudioObjectGetPropertyData(deviceIDs[idx], &propertyAddress, 0, nil, &propertySize, &deviceUID)
            { break }
            
            let devUID = String(format: "%@", deviceUID as! CVarArg)
            
            if !devUID.lowercased().contains("input") && !devUID.lowercased().contains("mic") {
                let device : [AnyHashable : Any] = [
                    "name"      : String(utf8String: deviceName) as Any,
                    "deviceID"  : Int32(deviceIDs[idx]),
                    "UID"       : "\(deviceUID)"]
                
                addObject(device)
                
                if selectedOutputDevice != nil {
                    if (selectedOutputDevice?.object(forKey: "UID") as! String).isEqual((device["UID"] as! String)) {
                        setSelectedObjects([device])
                        createSampleRateTableView()
                    }
                } else {
                    print("OAC.getOutputDevices: Error! No Default Output Device was found in UserDefaults")
                    //
                    // !! this will, or should not happen!!, Pbc will have created a default at startup if there was none
                    //
                }
            }
        }
    }

    func getSelectedDeviceSampleRates() -> Bool {
        
        var deviceRates : [Float64] = [Float64](repeating: 0.0, count: 32)
        let sampleRateCount = Int(getDeviceSampleRates(&deviceRates))
        deviceSampleRates = []
        
        for i in 0..<sampleRateCount {
            deviceSampleRates?.append(deviceRates[i])
        }
        //print("OAC.getSelectedDeviceSampleRates", sampleRateCount, deviceSampleRates as Any )
        if sampleRateCount > 0 {
            return true
        } else {
            return false
        }
    }
    
    func createSampleRateTableView() {
        let scrollView = NSScrollView(frame: sampleRatesView.bounds)
        sampleRatesTableView = NSTableView(frame: sampleRatesView.bounds)

        let srColumn = NSTableColumn(identifier: NSUserInterfaceItemIdentifier("SampleRate"))
        srColumn.width = sampleRatesTableView.frame.size.width
        srColumn.headerCell.isBordered = false
        srColumn.headerCell.isBezeled = false
        srColumn.headerCell.stringValue = SampleRateTableHeader
        srColumn.headerCell.alignment = .center
        (srColumn.dataCell as! NSCell).alignment = .right
        
        sampleRatesTableView.addTableColumn(srColumn)
        sampleRatesTableView.usesAlternatingRowBackgroundColors = false
        sampleRatesTableView.selectionHighlightStyle = .regular
        sampleRatesTableView.focusRingType = .none
        sampleRatesTableView.rowHeight = 18.0
        sampleRatesTableView.delegate = self
        sampleRatesTableView.dataSource = self

        scrollView.borderType = .noBorder
        scrollView.autoresizesSubviews = true
        scrollView.autoresizingMask = [.width, .height]
        scrollView.documentView = sampleRatesTableView
        
        sampleRatesView.addSubview(scrollView)
    }

    func tableView(_ aTableView: NSTableView, objectValueFor aTableColumn: NSTableColumn?, row rowIndex: Int) -> Any? {
        guard deviceSampleRates != nil else { return nil }
        
        let cellContent = String(format: "%.3f", deviceSampleRates![rowIndex] / 1000.0)
        return cellContent
    }

    func tableView(_ aTableView: NSTableView, shouldEdit aTableColumn: NSTableColumn?, row rowIndex: Int) -> Bool {
        // make cell non editable
        return false
    }

    func tableView(_ tableView: NSTableView, shouldSelectRow row: Int) -> Bool {
        if NSUserDefaultsController.shared.defaults.bool(forKey: "sampleRateFixed") {
            let selectedSampleRate = deviceSampleRates![row]
            UserDefaults.standard.set(selectedSampleRate, forKey: "fixedSampleRate")
            setDeviceSampleRate(selectedSampleRate)
            return true
        }
        return false
    }
    
    func numberOfRows(in tableView: NSTableView) -> Int {
        if deviceSampleRates == nil {
            if getSelectedDeviceSampleRates() == false {
                return 0
            }
        }
        return deviceSampleRates!.count
    }

    @IBAction func sampleRateFixedButton(_ sender: NSButton) {
        let state : Bool = sender.state.rawValue != 0
        setFixedOrMediaSampleRate(fixed: state)
    }
    
    func setFixedOrMediaSampleRate(fixed: Bool) {
        UserDefaults.standard.set(fixed, forKey: "sampleRateFixed")
        var sampleRate: Float64
        
        if fixed {
            sampleRateFixOrMediaBtn.state = .on
            sampleRate = NSUserDefaultsController.shared.defaults.double(forKey: "fixedSampleRate")
            if let index = deviceSampleRates?.firstIndex(of: sampleRate) {
                sampleRatesTableView.selectRowIndexes(IndexSet(integer: index), byExtendingSelection: false)
            }
        } else {
            sampleRateFixOrMediaBtn.state = .off
            sampleRate = NSUserDefaultsController.shared.defaults.double(forKey: "mediaSampleRate")
        }
        sampleRate = getAvailableDeviceSampleRate(sampleRate)
        setDeviceSampleRate(sampleRate)
    }
    
    func getAvailableDeviceSampleRate(_ sampleRate: Float64) -> Float64 {
        var newSampleRate = Float64(44100.0)
        var deviceSampleRates : [Float64] = [Float64](repeating: 0.0, count: 32)
        _ = getDeviceSampleRates(&deviceSampleRates)
        
        if deviceSampleRates.contains(sampleRate) {
            newSampleRate = sampleRate
        }
        return newSampleRate
    }
    
}
